/*
 * Flashbook GUI
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <string.h>
#include <gtk/gtk.h>

#include "fbgui.h"

#if !defined(G_OS_WIN32) && !defined(GDK_WINDOWING_QUARTZ)

GtkWidget* make_file_button(const char* title,
			    const char* old_path,
			    const char* suggest_dir,
			    const char* filters)
{
  GtkWidget* fcb = gtk_file_chooser_button_new
    (title, GTK_FILE_CHOOSER_ACTION_OPEN);

  setup_file_filters(GTK_FILE_CHOOSER(fcb), filters);

  if (old_path)
    gtk_file_chooser_select_filename(GTK_FILE_CHOOSER(fcb), old_path);
  else if (suggest_dir)
    gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(fcb), suggest_dir);

  return fcb;
}

char* file_button_get_selection(GtkWidget* filebtn)
{
  return gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(filebtn));
}

#else

/* Custom file chooser button using native file selector */

#define FB_TYPE_FILE_CHOOSER_BUTTON            (fb_file_chooser_button_get_type())
#define FB_FILE_CHOOSER_BUTTON(obj)            (G_TYPE_CHECK_INSTANCE_CAST((obj), FB_TYPE_FILE_CHOOSER_BUTTON, FBFileChooserButton))
#define FB_FILE_CHOOSER_BUTTON_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST((klass), FB_TYPE_FILE_CHOOSER_BUTTON, FBFileChooserButtonClass))
#define FB_IS_FILE_CHOOSER_BUTTON(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), FB_TYPE_FILE_CHOOSER_BUTTON))
#define FB_IS_FILE_CHOOSER_BUTTON_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), FB_TYPE_FILE_CHOOSER_BUTTON))
#define FB_FILE_CHOOSER_BUTTON_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), FB_TYPE_FILE_CHOOSER_BUTTON, FBFileChooserButtonClass))

typedef struct _FBFileChooserButton FBFileChooserButton;
typedef struct _FBFileChooserButtonClass FBFileChooserButtonClass;

struct _FBFileChooserButton {
  GtkHBox parent;

  GtkWidget* entry;
  GtkWidget* button;
  char* title;
  char* suggest_dir;
  char* filters;
  char* result;
};

struct _FBFileChooserButtonClass {
  GtkHBoxClass parent;
};

static void fb_file_chooser_button_finalize(GObject *obj);

static guint selection_changed_signal = 0;

G_DEFINE_TYPE(FBFileChooserButton, fb_file_chooser_button, GTK_TYPE_HBOX);

static void
fb_file_chooser_button_class_init(FBFileChooserButtonClass *class)
{
  GObjectClass *o_class;

  fb_file_chooser_button_parent_class = g_type_class_peek_parent(class);

  o_class = G_OBJECT_CLASS(class);
  o_class->finalize = fb_file_chooser_button_finalize;

  selection_changed_signal =
    g_signal_new("selection-changed",
		 G_OBJECT_CLASS_TYPE(o_class),
		 G_SIGNAL_RUN_LAST,
		 0, NULL, NULL,
		 g_cclosure_marshal_VOID__VOID,
		 G_TYPE_NONE,
		 0);
}

static void
entry_changed(GtkEntry *entry,
	      gpointer data)
{
  FBFileChooserButton* fcb = FB_FILE_CHOOSER_BUTTON(data);
  const char* text_utf8;
  char* text;
  char* result;

  text_utf8 = gtk_entry_get_text(entry);
  text = g_filename_from_utf8(text_utf8, -1, 0, 0, 0);

  if (!text || !text[0]) {
    result = NULL;
  }
  else if (g_path_is_absolute(text)) {
    if (g_file_test(text, G_FILE_TEST_EXISTS))
      result = g_strdup(text);
    else
      result = NULL;
  }
  else if (!fcb->suggest_dir) {
    result = NULL;
  }
  else {
    result = g_build_filename(fcb->suggest_dir, text, NULL);
    if (!g_file_test(result, G_FILE_TEST_EXISTS)) {
      g_free(result);
      result = NULL;
    }
  }
  g_free(text);

  if (!result && !fcb->result)
    return;

  if (result && fcb->result && !strcmp(result, fcb->result)) {
    g_free(result);
    return;
  }

  g_free(fcb->result);
  fcb->result = result;

  g_signal_emit(fcb, selection_changed_signal, 0, NULL);
}

static void
button_clicked(GtkButton *button G_GNUC_UNUSED,
	       gpointer data)
{
  FBFileChooserButton* fcb = FB_FILE_CHOOSER_BUTTON(data);
  GtkWidget* toplevel = gtk_widget_get_toplevel(GTK_WIDGET(fcb));
  char* result;
  char* text;

  if (!GTK_IS_WINDOW(toplevel))
    toplevel = NULL;

  result = run_file_chooser(fcb->title,
			    GTK_WINDOW(toplevel),
			    GTK_FILE_CHOOSER_ACTION_OPEN,
			    fcb->result,
			    NULL,
			    fcb->suggest_dir,
			    fcb->filters);

  if (result) {
    text = g_filename_to_utf8(result, -1, NULL, NULL, NULL);
    gtk_entry_set_text(GTK_ENTRY(fcb->entry), text);
    g_free(text);
    g_free(fcb->result);
    fcb->result = result;
  }
}

static void
fb_file_chooser_button_init(FBFileChooserButton *fcb)
{
  fcb->entry = gtk_entry_new();
  gtk_entry_set_activates_default(GTK_ENTRY(fcb->entry), TRUE);
  gtk_widget_show(fcb->entry);
  gtk_box_pack_start(GTK_BOX(fcb), fcb->entry, TRUE, TRUE, 0);
  g_signal_connect(fcb->entry, "changed", G_CALLBACK(entry_changed), fcb);

  fcb->button = gtk_button_new_with_mnemonic("_Browse...");
  gtk_widget_show(fcb->button);
  gtk_box_pack_start(GTK_BOX(fcb), fcb->button, FALSE, FALSE, 0);
  g_signal_connect(fcb->button, "clicked", G_CALLBACK(button_clicked), fcb);
}

static void
fb_file_chooser_button_finalize(GObject *obj)
{
  FBFileChooserButton* fcb = FB_FILE_CHOOSER_BUTTON(obj);

  g_free(fcb->title);
  g_free(fcb->suggest_dir);
  g_free(fcb->filters);
  g_free(fcb->result);

  if (G_OBJECT_CLASS(fb_file_chooser_button_parent_class)->finalize)
    (*G_OBJECT_CLASS(fb_file_chooser_button_parent_class)->finalize)(obj);
}

GtkWidget* make_file_button(const char* title,
			    const char* old_path,
			    const char* suggest_dir,
			    const char* filters)
{
  FBFileChooserButton* fcb = g_object_new(FB_TYPE_FILE_CHOOSER_BUTTON,
					  "homogeneous", FALSE,
					  "spacing", 4,
					  NULL);
  gchar* old_path_utf8;

  if (old_path) {
    old_path_utf8 = g_filename_to_utf8(old_path, -1, 0, 0, 0);
    gtk_entry_set_text(GTK_ENTRY(fcb->entry), old_path_utf8);
    g_free(old_path_utf8);

    g_free(fcb->result);
    fcb->result = g_strdup(old_path);
  }

  if (title)
    fcb->title = g_strdup(title);
  if (suggest_dir)
    fcb->suggest_dir = g_strdup(suggest_dir);
  if (filters)
    fcb->filters = g_strdup(filters);

  entry_changed(GTK_ENTRY(fcb->entry), fcb);

  return GTK_WIDGET(fcb);
}

char* file_button_get_selection(GtkWidget* filebtn)
{
  FBFileChooserButton* fcb = FB_FILE_CHOOSER_BUTTON(filebtn);

  g_return_val_if_fail(FB_IS_FILE_CHOOSER_BUTTON(filebtn), NULL);

  if (fcb->result)
    return g_strdup(fcb->result);
  else
    return NULL;
}

#endif
