/*
 * Flashbook GUI
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <string.h>
#include <gtk/gtk.h>

#include "fbgui.h"

#ifdef G_OS_WIN32
# define WIN32_LEAN_AND_MEAN
# include <windows.h>
# include <commdlg.h>
# include <gdk/gdkwin32.h>
# include <wchar.h>

# ifndef OPENFILENAME_SIZE_VERSION_400
#  define OPENFILENAME_SIZE_VERSION_400 sizeof(OPENFILENAMEA)
# endif

struct fcinfo {
  const char* title;
  GtkWindow* parent;
  GtkFileChooserAction action;
  const char* old_path;
  const char* suggest_name;
  const char* suggest_dir;
  const char* filters;
};

static char* file_chooser_main(const struct fcinfo* fci)
{
  if (G_WIN32_HAVE_WIDECHAR_API()) {
    OPENFILENAMEW ofnw;
    wchar_t *titlew, *filterw, *initdirw, *defextw;
    wchar_t filenamew[MAXPATHLEN + 1];
    wchar_t *p;
    char *q;
    int result;
    int i;

    titlew = g_utf8_to_utf16(fci->title, -1, 0, 0, 0);

    filterw = g_utf8_to_utf16(fci->filters, -1, 0, 0, 0);
    for (i = 0; filterw[i]; i++)
      if (filterw[i] == '|') filterw[i] = 0;

    memset(&ofnw, 0, sizeof(ofnw));
    ofnw.lStructSize = OPENFILENAME_SIZE_VERSION_400;

    if (fci->parent && GTK_WIDGET(fci->parent)->window)
      ofnw.hwndOwner = GDK_WINDOW_HWND(GTK_WIDGET(fci->parent)->window);

    ofnw.lpstrTitle = titlew;
    ofnw.lpstrFilter = filterw;
    ofnw.nFilterIndex = 1;
    ofnw.lpstrFile = filenamew;
    ofnw.nMaxFile = MAXPATHLEN;
    filenamew[0] = 0;

    initdirw = NULL;

    if (fci->old_path) {
      p = g_utf8_to_utf16(fci->old_path, -1, 0, 0, 0);
      if (p) {
	wcsncpy(filenamew, p, MAXPATHLEN);
	filenamew[MAXPATHLEN] = 0;
	g_free(p);
      }
    }
    else if (fci->suggest_name && fci->suggest_dir) {
      q = g_build_filename(fci->suggest_dir, fci->suggest_name, NULL);
      p = g_utf8_to_utf16(q, -1, 0, 0, 0);
      if (p) {
	wcsncpy(filenamew, p, MAXPATHLEN);
	filenamew[MAXPATHLEN] = 0;
	g_free(p);
      }
      g_free(q);
    }
    else if (fci->suggest_name) {
      p = g_utf8_to_utf16(fci->suggest_name, -1, 0, 0, 0);
      if (p) {
	wcsncpy(filenamew, p, MAXPATHLEN);
	filenamew[MAXPATHLEN] = 0;
	g_free(p);
      }
    }
    else if (fci->suggest_dir) {
      initdirw = g_utf8_to_utf16(fci->suggest_dir, -1, 0, 0, 0);
      ofnw.lpstrInitialDir = initdirw;
    }

    if (fci->suggest_name && (q = strrchr(fci->suggest_name, '.')))
      defextw = g_utf8_to_utf16(q + 1, -1, 0, 0, 0);
    else
      defextw = NULL;

    ofnw.lpstrDefExt = defextw;

    ofnw.Flags = (fci->action == GTK_FILE_CHOOSER_ACTION_SAVE
		  ? OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY
		  : OFN_FILEMUSTEXIST | OFN_HIDEREADONLY);

    result = (fci->action == GTK_FILE_CHOOSER_ACTION_SAVE
	      ? GetSaveFileNameW(&ofnw)
	      : GetOpenFileNameW(&ofnw));

    g_free(titlew);
    g_free(filterw);
    g_free(initdirw);
    g_free(defextw);

    if (result)
      return g_utf16_to_utf8(filenamew, -1, 0, 0, 0);
    else
      return NULL;
  }
  else {
    OPENFILENAMEA ofna;
    char *titlel, *filterl, *initdirl, *defextl;
    char filenamel[MAXPATHLEN + 1];
    char *p, *q;
    int result;
    int i;

    titlel = g_locale_from_utf8(fci->title, -1, 0, 0, 0);

    filterl = g_locale_from_utf8(fci->filters, -1, 0, 0, 0);
    for (i = 0; filterl[i]; i++)
      if (filterl[i] == '|') filterl[i] = 0;

    memset(&ofna, 0, sizeof(ofna));
    ofna.lStructSize = OPENFILENAME_SIZE_VERSION_400;

    if (fci->parent && GTK_WIDGET(fci->parent)->window)
      ofna.hwndOwner = GDK_WINDOW_HWND(GTK_WIDGET(fci->parent)->window);

    ofna.lpstrTitle = titlel;
    ofna.lpstrFilter = filterl;
    ofna.nFilterIndex = 1;
    ofna.lpstrFile = filenamel;
    ofna.nMaxFile = MAXPATHLEN;
    filenamel[0] = 0;

    initdirl = NULL;

    if (fci->old_path) {
      p = g_locale_from_utf8(fci->old_path, -1, 0, 0, 0);
      if (p) {
	strncpy(filenamel, p, MAXPATHLEN);
	filenamel[MAXPATHLEN] = 0;
	g_free(p);
      }
    }
    else if (fci->suggest_name && fci->suggest_dir) {
      q = g_build_filename(fci->suggest_dir, fci->suggest_name, NULL);
      p = g_locale_from_utf8(q, -1, 0, 0, 0);
      if (p) {
	strncpy(filenamel, p, MAXPATHLEN);
	filenamel[MAXPATHLEN] = 0;
	g_free(p);
      }
      g_free(q);
    }
    else if (fci->suggest_name) {
      p = g_locale_from_utf8(fci->suggest_name, -1, 0, 0, 0);
      if (p) {
	strncpy(filenamel, p, MAXPATHLEN);
	filenamel[MAXPATHLEN] = 0;
	g_free(p);
      }
    }
    else if (fci->suggest_dir) {
      initdirl = g_locale_from_utf8(fci->suggest_dir, -1, 0, 0, 0);
      ofna.lpstrInitialDir = initdirl;
    }

    if (fci->suggest_name && (q = strrchr(fci->suggest_name, '.')))
      defextl = g_locale_from_utf8(q + 1, -1, 0, 0, 0);
    else
      defextl = NULL;

    ofna.lpstrDefExt = defextl;

    ofna.Flags = (fci->action == GTK_FILE_CHOOSER_ACTION_SAVE
		  ? OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY
		  : OFN_FILEMUSTEXIST | OFN_HIDEREADONLY);

    result = (fci->action == GTK_FILE_CHOOSER_ACTION_SAVE
	      ? GetSaveFileNameA(&ofna)
	      : GetOpenFileNameA(&ofna));

    g_free(titlel);
    g_free(filterl);
    g_free(initdirl);
    g_free(defextl);

    if (result)
      return g_locale_to_utf8(filenamel, -1, 0, 0, 0);
    else
      return NULL;
  }
}

static gboolean wakeup(gpointer data G_GNUC_UNUSED)
{
  gtk_main_quit();
  return FALSE;
}

static gpointer file_chooser_thread(gpointer data)
{
  struct fcinfo* fci = data;
  gpointer res = file_chooser_main(fci);
  g_idle_add(wakeup, NULL);
  g_main_context_wakeup(g_main_context_default());
  return res;
}

char* run_file_chooser(const char* title,
		       GtkWindow* parent,
		       GtkFileChooserAction action,
		       const char* old_path,
		       const char* suggest_name,
		       const char* suggest_dir,
		       const char* filters)
{
  struct fcinfo fci;
  GThread* thread;
  gpointer result;
  GtkWidget* dummy;

  if (!g_thread_supported())
    g_thread_init(NULL);

  fci.title = title;
  fci.parent = parent;
  fci.action = action;
  fci.old_path = old_path;
  fci.suggest_name = suggest_name;
  fci.suggest_dir = suggest_dir;
  fci.filters = filters;

  if ((thread = g_thread_create(file_chooser_thread, &fci, TRUE, NULL))) {
    dummy = gtk_invisible_new();
    gtk_grab_add(dummy);
    gtk_main();
    result = g_thread_join(thread);
    gtk_widget_destroy(dummy);
    return result;
  }
  else
    return file_chooser_main(&fci);
}

#elif defined(HAVE_COCOA_COCOA_H) && defined(GDK_WINDOWING_QUARTZ) && defined(USE_COCOA_FILE_CHOOSER)
# include <gdk/gdkquartz.h>
# import <Cocoa/Cocoa.h>

// The following code is currently disabled, because it doesn't work.

static gboolean wakeup(gpointer data G_GNUC_UNUSED)
{
  gtk_main_quit();
  return FALSE;
}

@interface FileChooserDelegate : NSObject { }

- (void) dialogEnded: (NSSavePanel *) sheet
	  returnCode: (int) returnCode
	 contextInfo: (void *) contextInfo;

@end

@implementation FileChooserDelegate

- (void) dialogEnded: (NSSavePanel *) sheet
	  returnCode: (int) returnCode
	 contextInfo: (void *) contextInfo
{
  NSString* fnamestr;
  const char* utf8fname;
  char** fnamep = contextInfo;

  g_free(*fnamep);
  *fnamep = NULL;

  if (returnCode == NSOKButton) {
    fnamestr = [sheet filename];
    if (fnamestr) {
      utf8fname = [fnamestr UTF8String];
      (*fnamep) = g_filename_from_utf8(utf8fname, -1, NULL, NULL, NULL);
    }
  }

  g_idle_add(wakeup, NULL);
  g_main_context_wakeup(g_main_context_default());
}

@end

char* run_file_chooser(const char* title,
		       GtkWindow* parent,
		       GtkFileChooserAction action,
		       const char* old_path,
		       const char* suggest_name,
		       const char* suggest_dir,
		       const char* filters)
{
  char *p, *q;
  NSString* name;
  NSString* dir;
  NSWindow* parentwin;
  FileChooserDelegate* delegate;
  char *result = NULL;
  GtkWidget* dummy;

  if (old_path) {
    p = g_path_get_basename(old_path);
    q = g_filename_to_utf8(p, -1, NULL, NULL, NULL);
    name = [NSString stringWithUTF8String: q];
    g_free(p);
    g_free(q);

    p = g_path_get_dirname(old_path);
    q = g_filename_to_utf8(p, -1, NULL, NULL, NULL);
    dir = [NSString stringWithUTF8String: q];
    g_free(p);
    g_free(q);
  }
  else {
    if (suggest_name) {
      p = g_filename_to_utf8(suggest_name, -1, NULL, NULL, NULL);
      name = [NSString stringWithUTF8String: p];
      g_free(p);
    }
    else
      name = nil;

    if (suggest_dir) {
      p = g_filename_to_utf8(suggest_dir, -1, NULL, NULL, NULL);
      dir = [NSString stringWithUTF8String: p];
      g_free(p);
    }
    else
      dir = nil;
  }

  if (parent && GTK_WIDGET(parent)->window)
    parentwin = gdk_quartz_window_get_nswindow(GTK_WIDGET(parent)->window);
  else
    parentwin = nil;

  delegate = [FileChooserDelegate alloc];

  if (action == GTK_FILE_CHOOSER_ACTION_SAVE) {
    NSSavePanel* savepanel;
    NSString* ftype;

    p = strstr(filters, "|*.");
    p += 3;
    q = strchr(p, '|');
    p = g_strndup(p, q - p);
    ftype = [NSString stringWithUTF8String: p];
    g_free(p);

    savepanel = [NSSavePanel savePanel];
    [savepanel setTitle: [NSString stringWithUTF8String: title]];
    [savepanel setRequiredFileType: ftype];

    [savepanel beginSheetForDirectory: dir
	       file: name
	       modalForWindow: parentwin
	       modalDelegate: delegate
	       didEndSelector: @selector(dialogEnded:returnCode:contextInfo:)
	       contextInfo: &result];
  }
  else {
    NSOpenPanel* openpanel;
    NSArray* types = [NSArray array];
    char *p, *q, *patterns;
    char buf[256];

    strcpy(buf, filters);
    p = buf;

    while (*p) {
      patterns = strchr(p, '|');
      if (!patterns)
	break;
      patterns++;

      p = strchr(patterns, '|');
      if (!p)
	break;
      *p = 0;
      p++;

      while (patterns) {
	q = strchr(patterns, ';');
	if (q) {
	  *q = 0;
	  q++;
	}

	if (*patterns == '*' && patterns[1] == '.') {
	  patterns += 2;
	  types = [types arrayByAddingObject:
			   [NSString stringWithUTF8String: patterns]];
	}

	patterns = q;
      }
    }

    openpanel = [NSOpenPanel openPanel];
    [openpanel setTitle: [NSString stringWithUTF8String: title]];

    if (parent)
      [openpanel beginSheetForDirectory: dir
		 file: name
		 types: types
		 modalForWindow: parentwin
		 modalDelegate: delegate
		 didEndSelector: @selector(dialogEnded:returnCode:contextInfo:)
		 contextInfo: &result];
    else
      [openpanel beginForDirectory: dir
		 file: name
		 types: types
		 modelessDelegate: delegate
		 didEndSelector: @selector(dialogEnded:returnCode:contextInfo:)
		 contextInfo: &result];
  }

  dummy = gtk_invisible_new();
  gtk_grab_add(dummy);
  gtk_main();
  gtk_widget_destroy(dummy);

  [delegate release];

  return result;
}

#else  /* ! G_OS_WIN32, ! HAVE_COCOA_COCOA_H */

/* Case insensitive filter function */
static gboolean filter_suffix_case(const GtkFileFilterInfo* info,
				   gpointer data)
{
  const char* suffix = data;
  int n, m;

  n = strlen(info->filename);
  m = strlen(suffix);

  if (n < m)
    return FALSE;
  else if (g_ascii_strcasecmp(info->filename + n - m, suffix))
    return FALSE;
  else
    return TRUE;
}

void setup_file_filters(GtkFileChooser* chooser,
			const char* filters)
{
  GtkFileFilter* ffilt;
  char buf[256];
  char *p, *q, *desc, *patterns;

  strcpy(buf, filters);
  p = buf;

  while (*p) {
    desc = p;
    patterns = strchr(desc, '|');
    if (!patterns)
      break;
    *patterns = 0;
    patterns++;

    p = strchr(patterns, '|');
    if (!p)
      break;
    *p = 0;
    p++;

    ffilt = gtk_file_filter_new();
    gtk_file_filter_set_name(ffilt, desc);

    while (patterns) {
      q = strchr(patterns, ';');
      if (q) {
	*q = 0;
	q++;
      }

      if (*patterns == '*'
	  && !strchr(patterns + 1, '*')
	  && !strchr(patterns + 1, '?'))
	gtk_file_filter_add_custom(ffilt, GTK_FILE_FILTER_FILENAME,
				   filter_suffix_case,
				   g_strdup(patterns + 1),
				   g_free);
      else
	gtk_file_filter_add_pattern(ffilt, patterns);

      patterns = q;
    }

    gtk_file_chooser_add_filter(chooser, ffilt);
  }
}

char* run_file_chooser(const char* title,
		       GtkWindow* parent,
		       GtkFileChooserAction action,
		       const char* old_path,
		       const char* suggest_name,
		       const char* suggest_dir,
		       const char* filters)
{
  GtkWidget* filesel;
  GtkWidget* dlg;
  GtkWidget* button;
  static char* previous_dir = NULL;
  char *fname, *p, *q, *r;

  filesel = gtk_file_chooser_dialog_new(title, parent, action,
					GTK_STOCK_CANCEL,
					GTK_RESPONSE_CANCEL,
					(action == GTK_FILE_CHOOSER_ACTION_SAVE
					 ? GTK_STOCK_SAVE
					 : GTK_STOCK_OPEN),
					GTK_RESPONSE_ACCEPT,
					NULL);
  gtk_dialog_set_alternative_button_order(GTK_DIALOG(filesel),
					  GTK_RESPONSE_ACCEPT,
					  GTK_RESPONSE_CANCEL,
					  -1);
  gtk_dialog_set_default_response(GTK_DIALOG(filesel),
				  GTK_RESPONSE_ACCEPT);

  if (old_path)
    gtk_file_chooser_select_filename(GTK_FILE_CHOOSER(filesel), old_path);
  else {
    if (suggest_name)
      gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(filesel),
					suggest_name);

    if (suggest_dir)
      gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(filesel),
					  suggest_dir);
    else if (previous_dir)
      gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(filesel),
					  previous_dir);
  }

  setup_file_filters(GTK_FILE_CHOOSER(filesel), filters);

  while (1) {
    if (gtk_dialog_run(GTK_DIALOG(filesel)) != GTK_RESPONSE_ACCEPT) {
      fname = NULL;
      break;
    }

    fname = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(filesel));
    if (!fname)
      continue;

    if (action == GTK_FILE_CHOOSER_ACTION_OPEN) {
      if (g_file_test(fname, G_FILE_TEST_EXISTS))
	break;
      else {
	g_free(fname);
	continue;
      }
    }

    if (!g_file_test(fname, G_FILE_TEST_EXISTS))
      break;

    p = g_path_get_basename(fname);
    q = g_filename_to_utf8(p, -1, NULL, NULL, NULL);
    dlg = gtk_message_dialog_new(GTK_WINDOW(filesel),
				 GTK_DIALOG_MODAL,
				 GTK_MESSAGE_QUESTION,
				 GTK_BUTTONS_NONE,
				 "A file named \"%s\" already exists.  "
				 "Do you want to replace it?",
				 q);
    g_free(p);
    g_free(q);
    p = g_path_get_dirname(fname);
    q = g_path_get_basename(p);
    r = g_filename_to_utf8(q, -1, NULL, NULL, NULL);
    gtk_message_dialog_format_secondary_markup
      (GTK_MESSAGE_DIALOG(dlg),
       "The file already exists in \"%s\".  Replacing it will "
       "overwrite its contents.", r);
    g_free(p);
    g_free(q);
    g_free(r);

    gtk_dialog_add_button(GTK_DIALOG(dlg),
			  GTK_STOCK_CANCEL,
			  GTK_RESPONSE_CANCEL);

    button = gtk_button_new_with_mnemonic("_Replace");
    GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
    gtk_button_set_image(GTK_BUTTON(button),
			 gtk_image_new_from_stock(GTK_STOCK_SAVE,
						  GTK_ICON_SIZE_BUTTON));
    gtk_widget_show(button);
    gtk_dialog_add_action_widget(GTK_DIALOG(dlg), button,
				 GTK_RESPONSE_ACCEPT);

    gtk_dialog_set_alternative_button_order(GTK_DIALOG(dlg),
					    GTK_RESPONSE_ACCEPT,
					    GTK_RESPONSE_CANCEL,
					    -1);

    if (gtk_dialog_run(GTK_DIALOG(dlg)) == GTK_RESPONSE_ACCEPT) {
      gtk_widget_destroy(dlg);
      break;
    }

    g_free(fname);
    gtk_widget_destroy(dlg);
  }

  if (fname) {
    p = gtk_file_chooser_get_current_folder(GTK_FILE_CHOOSER(filesel));
    if (p) {
      g_free(previous_dir);
      previous_dir = p;
    }
  }

  gtk_widget_destroy(filesel);
  return fname;
}

#endif
