/*
 * Flashbook Font Editor
 * Copyright (c) 2008-2010 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gucharmap/gucharmap.h>
#include <glib/gstdio.h>
#include <errno.h>
#include <fb.h>
#include "fbgui.h"
#include "fontedit.h"
#include "msgbox.h"

#ifdef HAVE_LIBIGE_MAC_INTEGRATION
# include <ige-mac-menu.h>
#endif


static void run_sample_text_dialog(appinfo* ai);


/**************** Character map ****************/

#if (GUCHARMAP_VERSION_MAJOR > 2) || (GUCHARMAP_VERSION_MINOR >= 23)

/* new gucharmap-2 api */

static GtkWidget* charmap_new()
{
  GtkWidget* charmap = gucharmap_charmap_new();
  GucharmapChaptersModel* model = gucharmap_block_chapters_model_new();
  gucharmap_charmap_set_chapters_model(GUCHARMAP_CHARMAP(charmap), model);
  g_object_unref(model);
  gucharmap_charmap_set_active_chapter(GUCHARMAP_CHARMAP(charmap), "All");
  gucharmap_charmap_set_active_character(GUCHARMAP_CHARMAP(charmap), 0x20);
  return charmap;
}

static gunichar charmap_get_char(GtkWidget* charmap)
{
  return gucharmap_charmap_get_active_character(GUCHARMAP_CHARMAP(charmap));
}

static void charmap_set_char(GtkWidget* charmap, gunichar c)
{
  gucharmap_charmap_set_active_character(GUCHARMAP_CHARMAP(charmap), c);
}

static char* charmap_get_font(GtkWidget* charmap)
{
  PangoFontDescription* fontdesc;
  fontdesc = gucharmap_charmap_get_font_desc(GUCHARMAP_CHARMAP(charmap));
  return pango_font_description_to_string(fontdesc);
}

static void charmap_set_font(GtkWidget* charmap, const char* name)
{
  PangoFontDescription* fontdesc;
  fontdesc = pango_font_description_from_string(name);
  gucharmap_charmap_set_font_desc(GUCHARMAP_CHARMAP(charmap), fontdesc);
  pango_font_description_free(fontdesc);
}

static void charmap_connect_active_char(GtkWidget* charmap,
					void (*callback)(void*),
					void *data)
{
  g_signal_connect_swapped(charmap, "notify::active-character",
			   G_CALLBACK(callback), data);
}

#else

/* original gucharmap api */

static GtkWidget* charmap_new()
{
  GtkWidget* chapters = gucharmap_block_chapters_new();
  GtkWidget* charmap = gucharmap_charmap_new(GUCHARMAP_CHAPTERS(chapters));
  gucharmap_charmap_go_to_character(GUCHARMAP_CHARMAP(charmap), 0x20);
  return charmap;
}

static gunichar charmap_get_char(GtkWidget* charmap)
{
  GucharmapTable* chartbl;
  chartbl = gucharmap_charmap_get_chartable(GUCHARMAP_CHARMAP(charmap));
  return gucharmap_table_get_active_character(chartbl);
}

static void charmap_set_char(GtkWidget* charmap, gunichar c)
{
  gucharmap_charmap_go_to_character(GUCHARMAP_CHARMAP(charmap), c);
}

static char* charmap_get_font(GtkWidget* charmap)
{
  GucharmapTable* chartbl;
  chartbl = gucharmap_charmap_get_chartable(GUCHARMAP_CHARMAP(charmap));
  if (chartbl->font_name)
    return g_strdup(chartbl->font_name);
  else
    return NULL;
}

static void charmap_set_font(GtkWidget* charmap, const char* name)
{
  gucharmap_charmap_set_font(GUCHARMAP_CHARMAP(charmap), name);
}

static void charmap_connect_active_char(GtkWidget* charmap,
					void (*callback)(void*),
					void *data)
{
  GucharmapTable* chartbl;
  chartbl = gucharmap_charmap_get_chartable(GUCHARMAP_CHARMAP(charmap));
  g_signal_connect_swapped(chartbl, "set-active-char",
			   G_CALLBACK(callback), data);
}

#endif

/**************** Font bitmap editor ****************/

static void clear_bitmap(fbfontchar* bm)
{
  memset(bm, 0, sizeof(fbfontchar));
  bm->custom = 1;
}

static void select_editor(editinfo* ei)
{
  gdouble xmin, xmax;
  int boxxmin, boxxmax;

  if (ei->app && ei->app->active_edit != ei) {
    if (ei->app->active_edit)
      gtk_frame_set_shadow_type(GTK_FRAME(gtk_bin_get_child(GTK_BIN(ei->app->active_edit->box))),
				GTK_SHADOW_NONE);
    gtk_frame_set_shadow_type(GTK_FRAME(gtk_bin_get_child(GTK_BIN(ei->box))),
			      GTK_SHADOW_IN);
    ei->app->active_edit = ei;
  }

  if (ei->box && ei->app && ei->app->edithadj) {
    /* scroll viewport so that the editor box is visible */
    xmin = ei->app->edithadj->value;
    xmax = xmin + ei->app->edithadj->page_size;
    boxxmin = ei->box->allocation.x;
    boxxmax = boxxmin + ei->box->allocation.width;

    if (boxxmin <= xmin)
      gtk_adjustment_set_value(ei->app->edithadj, boxxmin);
    else if (boxxmax >= xmax)
      gtk_adjustment_set_value(ei->app->edithadj,
			       boxxmax - ei->app->edithadj->page_size);
  }
}

static int preview_append_bitmap(guint64* rows, const fbfontchar* bm)
{
  int i;

  if (rows) {
    for (i = 0; i < FB_MAX_FONT_HEIGHT; i++) {
      rows[i + 1] <<= bm->width;
      rows[i + 1] |= bm->bitmap[i];
    }
  }

  return bm->width;
}

static int preview_append_string(editinfo* ei, guint64* rows, const gchar* s)
{
  gunichar c;
  fbfontchar* bm;
  int w = 0;

  while (*s) {
    c = g_utf8_get_char(s);
    bm = g_tree_lookup(ei->bitmap_tree, GINT_TO_POINTER(c));
    if (bm)
      w += preview_append_bitmap(rows, bm);
    s = g_utf8_next_char(s);
  }

  return w;
}

static void editor_update_preview(gpointer eip, gpointer data G_GNUC_UNUSED)
{
  editinfo* ei = eip;
  int w = 1;

  if (ei->app && ei->app->preview_left_text)
    w += preview_append_string(ei, NULL, ei->app->preview_left_text);
  if (ei->current_bitmap)
    w += ei->current_bitmap->width;
  if (ei->app && ei->app->preview_right_text)
    w += preview_append_string(ei, NULL, ei->app->preview_right_text);
  if (w > 64)
    w = 64;

  gtk_widget_set_size_request(ei->da_preview, 2 * w, 2 * (ei->cell_height + 2));
  gtk_widget_queue_draw(ei->da_preview);
}

static void editor_set_current_modified(editinfo* ei, gboolean mod)
{
  ei->current_modified = mod;

  if (mod)
    ei->modified = TRUE;

  if (ei->current_bitmap->custom
      && (mod || ei->current_bitmap != &ei->tmp_bitmap))
    gtk_widget_set_sensitive(ei->btn_delete, TRUE);
  else
    gtk_widget_set_sensitive(ei->btn_delete, FALSE);

  editor_update_preview(ei, NULL);
}

static gboolean da_expose(GtkWidget* w,
			  GdkEventExpose* ev G_GNUC_UNUSED,
			  gpointer data)
{
  editinfo* ei = data;
  GdkDrawable* d = GDK_DRAWABLE(w->window);
  gint width, height;
  int i, j;
  int minx, maxx, miny, maxy;
  int bwidth, bcwidth, bcheight, bline;
  guint64 rows[FB_MAX_FONT_HEIGHT+2];

  if (w == ei->da_preview) {
    bcheight = ei->cell_height + 2;
    bcwidth = 1;

    for (i = 0; i < FB_MAX_FONT_HEIGHT + 2; i++)
      rows[i] = 0;

    if (ei->app && ei->app->preview_left_text)
      bcwidth += preview_append_string(ei, rows, ei->app->preview_left_text);
    if (ei->current_bitmap)
      bcwidth += preview_append_bitmap(rows, ei->current_bitmap);
    if (ei->app && ei->app->preview_right_text)
      bcwidth += preview_append_string(ei, rows, ei->app->preview_right_text);
    if (bcwidth > 64)
      bcwidth = 64;

    bwidth = bcwidth;
    bline = -1;
  }
  else {
    bcwidth = ei->cell_width;
    bcheight = ei->cell_height;
    bwidth = ei->current_bitmap->width;
    for (i = 0; i < bcheight; i++)
      rows[i] = ei->current_bitmap->bitmap[i];
    bline = ei->baseline + 1;
  }

  gdk_drawable_get_size(d, &width, &height);
  if (width * bcheight > height * bcwidth)
    width = (height * bcwidth) / bcheight;
  else
    height = (width * bcheight) / bcwidth;


  if (!ei->active) {
    gdk_window_set_background(w->window, &w->style->bg[w->state]);
    gdk_window_clear(w->window);
    return TRUE;
  }

  gdk_window_set_background(w->window, &w->style->base[w->state]);
  gdk_window_clear(w->window);

  maxy = 0;
  for (i = 0; i < bcheight; i++) {
    miny = maxy;
    maxy = (((i + 1) * height + bcheight / 2) / bcheight);

    maxx = 0;
    for (j = 0; j < bcwidth; j++) {
      minx = maxx;
      maxx = (((j + 1) * width + bcwidth / 2) / bcwidth);

      if ((rows[i] >> (bcwidth - j - 1)) & 1)
	gdk_draw_rectangle(d, w->style->text_gc[w->state],
			   TRUE, minx, miny, maxx-minx, maxy-miny);
      else if (j < (bcwidth - bwidth))
	gdk_draw_rectangle(d, w->style->mid_gc[w->state],
			   TRUE, minx, miny, maxx-minx, maxy-miny);
    }
  }

  if (bline > 0) {
    miny = ((bline * height + bcheight / 2) / bcheight);
    gdk_draw_line(d, w->style->text_aa_gc[w->state],
		  0, miny, width, miny);
  }

  if (GTK_WIDGET_HAS_FOCUS(w)) {
    minx = (ei->kbeditx * width + bcwidth / 2) / bcwidth;
    maxx = ((ei->kbeditx + 1) * width + bcwidth / 2) / bcwidth;
    miny = (ei->kbedity * height + bcheight / 2) / bcheight;
    maxy = ((ei->kbedity + 1) * height + bcheight / 2) / bcheight;

    if (rows[ei->kbedity] & (1 << (bcwidth - ei->kbeditx - 1)))
      gtk_paint_focus(w->style, w->window, GTK_STATE_SELECTED,
		      NULL, w, "fb-fontedit-pixel",
		      minx, miny, maxx - minx, maxy - miny);
    else
      gtk_paint_focus(w->style, w->window, GTK_STATE_NORMAL,
		      NULL, w, "fb-fontedit-pixel",
		      minx, miny, maxx - minx, maxy - miny);
  }

  return TRUE;
}

static void paint_pixel(editinfo* ei, int x, int y, int button)
{
  GtkAdjustment* adj;
  gint wid, i;
  guchar mask, col;

  mask = (1 << (ei->cell_width - x - 1));

  if (button == 2) {
    if (ei->current_bitmap->bitmap[y] & mask)
      button = 3;
    else
      button = 1;
  }

  if (ei->autometric && button == 1) {
    for (i = col = 0; i < ei->cell_height; i++)
      col |= ei->current_bitmap->bitmap[i];

    if (!(col & mask)) {
      wid = ei->cell_width - x;
      if (wid > ei->current_bitmap->width) {
	adj = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(ei->sb_width));
	gtk_adjustment_set_value(adj, wid);
      }
    }
    ei->autometric = TRUE;
  }

  if (button == 1)
    ei->current_bitmap->bitmap[y] |= mask;
  else
    ei->current_bitmap->bitmap[y] &= ~mask;

  if (ei->autometric && button != 1) {
    for (i = col = 0; i < ei->cell_height; i++)
      col |= ei->current_bitmap->bitmap[i];

    if (!(col & mask)) {
      wid = ei->cell_width - x;
      if (wid == ei->current_bitmap->width) {
	while (wid > ei->default_width
	       && !(col & mask)) {
	  wid--;
	  mask >>= 1;
	}
	adj = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(ei->sb_width));
	gtk_adjustment_set_value(adj, wid);
      }
    }
    ei->autometric = TRUE;
  }

  editor_set_current_modified(ei, TRUE);
  select_editor(ei);
  gtk_widget_queue_draw(ei->da_edit);
  gtk_widget_queue_draw(ei->da_preview);
}

static void mouse_draw(GtkWidget* w, gdouble mousex, gdouble mousey,
		       editinfo* ei)
{
  GdkDrawable* d = GDK_DRAWABLE(w->window);
  gint width, height;
  gint x, y;

  gdk_drawable_get_size(d, &width, &height);
  if (width * ei->cell_height > height * ei->cell_width)
    width = (height * ei->cell_width) / ei->cell_height;
  else
    height = (width * ei->cell_height) / ei->cell_width;

  x = (mousex * ei->cell_width) / width;
  y = (mousey * ei->cell_height) / height;

  if (x < 0 || x >= ei->cell_width || y < 0 || y >= ei->cell_height)
    return;

  ei->kbeditx = x;
  ei->kbedity = y;

  if (!ei->mousebutton)
    return;

  paint_pixel(ei, x, y, ei->mousebutton);
}

static gboolean da_button_press(GtkWidget* w, GdkEventButton* ev,
				gpointer data)
{
  editinfo* ei = data;

  if (ev->button == 1 && ev->state & GDK_CONTROL_MASK)
    ei->mousebutton = 3;
  else
    ei->mousebutton = ev->button;

  mouse_draw(w, ev->x, ev->y, ei);
  gtk_widget_grab_focus(w);
  return TRUE;
}

static gboolean da_preview_button_press(GtkWidget* w G_GNUC_UNUSED,
					GdkEventButton* ev, gpointer data)
{
  editinfo* ei = data;

  if (ev->button == 1)
    run_sample_text_dialog(ei->app);
  return TRUE;
}

static gboolean da_button_release(GtkWidget* w, GdkEventButton* ev,
				  gpointer data)
{
  editinfo* ei = data;
  ei->mousebutton = 0;
  mouse_draw(w, ev->x, ev->y, ei);
  gtk_widget_queue_draw(ei->da_edit);
  return TRUE;
}

static gboolean da_motion_notify(GtkWidget* w, GdkEventMotion* ev,
				 gpointer data)
{
  editinfo* ei = data;
  mouse_draw(w, ev->x, ev->y, ei);
  return TRUE;
}

static gboolean da_key_press(GtkWidget* w, GdkEventKey* ev,
			     gpointer data)
{
  editinfo* ei = data;
  GtkAdjustment* adj;

  switch (ev->keyval) {
  case GDK_KP_End:
  case GDK_KP_1:
    ei->kbeditx--;
    ei->kbedity++;
    break;

  case GDK_Down:
  case GDK_KP_Down:
  case GDK_KP_2:
    ei->kbedity++;
    break;

  case GDK_KP_Next:
  case GDK_KP_3:
    ei->kbeditx++;
    ei->kbedity++;
    break;

  case GDK_Left:
  case GDK_KP_Left:
  case GDK_KP_4:
    if (ev->state & GDK_SHIFT_MASK) {
      adj = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(ei->sb_width));
      gtk_adjustment_set_value(adj, adj->value + 1);
    }
    else {
      ei->kbeditx--;
    }
    break;

  case GDK_Right:
  case GDK_KP_Right:
  case GDK_KP_6:
    if (ev->state & GDK_SHIFT_MASK) {
      adj = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(ei->sb_width));
      gtk_adjustment_set_value(adj, adj->value - 1);
    }
    else {
      ei->kbeditx++;
    }
    break;

  case GDK_KP_Home:
  case GDK_KP_7:
    ei->kbeditx--;
    ei->kbedity--;
    break;

  case GDK_Up:
  case GDK_KP_Up:
  case GDK_KP_8:
    ei->kbedity--;
    break;

  case GDK_KP_Prior:
  case GDK_KP_9:
    ei->kbeditx++;
    ei->kbedity--;
    break;

  case GDK_space:
  case GDK_KP_Space:
  case GDK_Return:
  case GDK_KP_Enter:
  case GDK_KP_Begin:
  case GDK_KP_5:
    paint_pixel(ei, ei->kbeditx, ei->kbedity, 2);
    break;

  default:
    return FALSE;
  }

  ei->kbeditx = CLAMP(ei->kbeditx, 0, ei->cell_width - 1);
  ei->kbedity = CLAMP(ei->kbedity, 0, ei->cell_height - 1);
  gtk_widget_queue_draw(w);
  return TRUE;
}

static void builtin_changed(GtkToggleButton* btn, gpointer data)
{
  editinfo* ei = data;
  gboolean value = gtk_toggle_button_get_active(btn);
  GtkAdjustment* adj;

  if (value && ei->current_bitmap->custom) {
    ei->current_bitmap->custom = 0;
    load_builtin_bitmap(ei, ei->current_bitmap);

    gtk_widget_set_sensitive(ei->da_edit, FALSE);
    gtk_widget_set_sensitive(ei->sb_width, FALSE);

    adj = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(ei->sb_width));
    gtk_adjustment_set_value(adj, ei->current_bitmap->width);

    editor_set_current_modified(ei, TRUE);
    gtk_widget_queue_draw(ei->da_edit);
    gtk_widget_queue_draw(ei->da_preview);
    select_editor(ei);
  }
  else if (!value && !ei->current_bitmap->custom) {
    ei->current_bitmap->custom = 1;

    gtk_widget_set_sensitive(ei->da_edit, TRUE);
    gtk_widget_set_sensitive(ei->sb_width, TRUE);

    editor_set_current_modified(ei, TRUE);
    gtk_widget_queue_draw(ei->da_edit);
    select_editor(ei);
    ei->autometric = TRUE;
  }
}

static void width_changed(GtkAdjustment* adj, gpointer data)
{
  editinfo* ei = data;
  int w = gtk_adjustment_get_value(adj);

  if (ei->active && w != ei->current_bitmap->width) {
    ei->current_bitmap->width = w;
    editor_set_current_modified(ei, TRUE);
    gtk_widget_queue_draw(ei->da_edit);
    gtk_widget_queue_draw(ei->da_preview);

    select_editor(ei);
    ei->autometric = FALSE;
  }
}

static void delete_bitmap(GtkButton* button G_GNUC_UNUSED, gpointer data)
{
  editinfo* ei = data;
  int i;
  GtkAdjustment* adj;

  select_editor(ei);

  if (find_builtin_bitmap(ei, ei->current_bitmap)) {
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ei->cb_builtin), TRUE);
    return;
  }

  if (ei->current_bitmap != &ei->tmp_bitmap)
    g_tree_remove(ei->bitmap_tree, GINT_TO_POINTER(ei->current_bitmap->value));

  ei->current_bitmap = &ei->tmp_bitmap;
  for (i = 0; i < FB_MAX_FONT_HEIGHT; i++)
    ei->current_bitmap->bitmap[i] = 0;

  adj = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(ei->sb_width));
  gtk_adjustment_set_value(adj, ei->default_width);

  ei->current_bitmap->custom = 1;
  ei->autometric = TRUE;

  editor_set_current_modified(ei, FALSE);
  gtk_widget_queue_draw(ei->da_edit);
  gtk_widget_queue_draw(ei->da_preview);
}

static int compare_gunichars(gconstpointer a, gconstpointer b,
			     gpointer data G_GNUC_UNUSED)
{
  gunichar ca = GPOINTER_TO_INT(a);
  gunichar cb = GPOINTER_TO_INT(b);

  if (ca < cb) return -1;
  else if (ca > cb) return 1;
  else return 0;
}

static gboolean widget_focus(GtkWidget* w G_GNUC_UNUSED,
			     GdkEventFocus* ev G_GNUC_UNUSED,
			     gpointer data)
{
  editinfo* ei = data;
  select_editor(ei);
  return FALSE;
}

static gboolean ebox_button_press(GtkWidget* w G_GNUC_UNUSED,
				  GdkEventButton* ev G_GNUC_UNUSED,
				  gpointer data)
{
  editinfo* ei = data;
  select_editor(ei);
  gtk_widget_grab_focus(ei->da_edit);
  return TRUE;
}

static editinfo* new_editor(appinfo* app, int charset, int cell_width,
			    int cell_height, int min_row_height,
			    int baseline, int default_width)
{
  editinfo* ei = g_new0(editinfo, 1);
  GtkWidget *frame, *af, *vbox, *hbox, *hbox2, *lbl;
  GtkAdjustment *adj;

  ei->app = app;
  ei->charset = charset;
  ei->cell_width = cell_width;
  ei->cell_height = cell_height;
  ei->min_row_height = min_row_height;
  ei->baseline = baseline;
  ei->default_width = default_width;
  ei->bitmap_tree = g_tree_new_full(compare_gunichars, NULL, NULL, g_free);
  ei->current_bitmap = &ei->tmp_bitmap;
  ei->active = FALSE;

  ei->box = gtk_event_box_new();
  gtk_widget_add_events(ei->box, GDK_BUTTON_PRESS_MASK);
  g_signal_connect(ei->box, "button-press-event",
		   G_CALLBACK(ebox_button_press), ei);

  frame = gtk_frame_new("Untitled");
  gtk_container_set_border_width(GTK_CONTAINER(frame), 5);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_NONE);
  {
    hbox = gtk_hbox_new(FALSE, 10);
    gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);
    {
      vbox = gtk_vbox_new(FALSE, 10);
      {
	af = gtk_aspect_frame_new(NULL, 0.5, 0.5, 0.0, TRUE);
	gtk_frame_set_shadow_type(GTK_FRAME(af), GTK_SHADOW_IN);
	{
	  ei->da_edit = gtk_drawing_area_new();
	  gtk_widget_set_size_request(ei->da_edit,
				      16 * cell_width, 16 * cell_height);
	  GTK_WIDGET_SET_FLAGS(ei->da_edit, GTK_CAN_FOCUS);
	  gtk_widget_add_events(ei->da_edit,
				(GDK_BUTTON_PRESS_MASK
				 | GDK_BUTTON_RELEASE_MASK
				 | GDK_BUTTON1_MOTION_MASK
				 | GDK_BUTTON3_MOTION_MASK
				 | GDK_KEY_PRESS_MASK));

	  g_signal_connect(ei->da_edit, "expose-event",
			   G_CALLBACK(da_expose), ei);
	  g_signal_connect(ei->da_edit, "button-press-event",
			   G_CALLBACK(da_button_press), ei);
	  g_signal_connect(ei->da_edit, "button-release-event",
			   G_CALLBACK(da_button_release), ei);
	  g_signal_connect(ei->da_edit, "motion-notify-event",
			   G_CALLBACK(da_motion_notify), ei);
	  g_signal_connect(ei->da_edit, "key-press-event",
			   G_CALLBACK(da_key_press), ei);
	  g_signal_connect(ei->da_edit, "focus-in-event",
			   G_CALLBACK(widget_focus), ei);

	  gtk_container_add(GTK_CONTAINER(af), ei->da_edit);
	  gtk_widget_show(ei->da_edit);
	}
	gtk_box_pack_start(GTK_BOX(vbox), af, TRUE, TRUE, 0);
	gtk_widget_show(af);

	hbox2 = gtk_hbox_new(FALSE, 10);
	{
	  lbl = gtk_label_new("Preview:");
	  gtk_box_pack_start(GTK_BOX(hbox2), lbl, FALSE, FALSE, 0);
	  gtk_widget_show(lbl);

	  af = gtk_aspect_frame_new(NULL, 0.5, 0.5, 0.0, TRUE);
	  gtk_frame_set_shadow_type(GTK_FRAME(af), GTK_SHADOW_IN);
	  {
	    ei->da_preview = gtk_drawing_area_new();
	    gtk_widget_add_events(ei->da_preview, GDK_BUTTON_PRESS_MASK);
	    editor_update_preview(ei, NULL);
	    g_signal_connect(ei->da_preview, "expose-event",
			     G_CALLBACK(da_expose), ei);
	    g_signal_connect(ei->da_preview, "button-press-event",
			     G_CALLBACK(da_preview_button_press), ei);

	    gtk_container_add(GTK_CONTAINER(af), ei->da_preview);
	    gtk_widget_show(ei->da_preview);
	  }
	  gtk_box_pack_end(GTK_BOX(hbox2), af, TRUE, TRUE, 0);
	  gtk_widget_show(af);
	}
	gtk_box_pack_start(GTK_BOX(vbox), hbox2, FALSE, FALSE, 0);
	gtk_widget_show(hbox2);
      }
      gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);
      gtk_widget_show(vbox);

      vbox = gtk_vbox_new(FALSE, 10);
      {
	ei->cb_builtin = gtk_check_button_new_with_mnemonic("Use _Builtin");
	g_signal_connect(ei->cb_builtin, "focus-in-event",
			 G_CALLBACK(widget_focus), ei);
	g_signal_connect(ei->cb_builtin, "toggled",
			 G_CALLBACK(builtin_changed), ei);
	gtk_box_pack_start(GTK_BOX(vbox), ei->cb_builtin, FALSE, FALSE, 0);
	gtk_widget_show(ei->cb_builtin);

	hbox2 = gtk_hbox_new(FALSE, 6);
	{
	  lbl = gtk_label_new_with_mnemonic("_Width:");
	  gtk_misc_set_alignment(GTK_MISC(lbl), 1.0, 0.5);
	  gtk_box_pack_start(GTK_BOX(hbox2), lbl, TRUE, TRUE, 0);
	  gtk_widget_show(lbl);

	  ei->sb_width = gtk_spin_button_new_with_range(0, 99, 1);
	  g_signal_connect(ei->sb_width, "focus-in-event",
			   G_CALLBACK(widget_focus), ei);
	  adj = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(ei->sb_width));
	  g_signal_connect(adj, "value-changed", G_CALLBACK(width_changed), ei);

	  gtk_box_pack_start(GTK_BOX(hbox2), ei->sb_width, FALSE, FALSE, 0);
	  gtk_label_set_mnemonic_widget(GTK_LABEL(lbl), ei->sb_width);
	  gtk_widget_show(ei->sb_width);
	}
	gtk_box_pack_start(GTK_BOX(vbox), hbox2, FALSE, FALSE, 0);
	gtk_widget_show(hbox2);

	ei->btn_delete = gtk_button_new_from_stock(GTK_STOCK_CLEAR);
	g_signal_connect(ei->btn_delete, "clicked",
			 G_CALLBACK(delete_bitmap), ei);
	g_signal_connect(ei->btn_delete, "focus-in-event",
			 G_CALLBACK(widget_focus), ei);
	gtk_box_pack_end(GTK_BOX(vbox), ei->btn_delete, FALSE, FALSE, 0);
	gtk_widget_show(ei->btn_delete);
      }
      gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);
      gtk_widget_show(vbox);
    }
    gtk_container_add(GTK_CONTAINER(frame), hbox);
    gtk_widget_show(hbox);
  }
  gtk_container_add(GTK_CONTAINER(ei->box), frame);
  gtk_widget_show(frame);

  gtk_widget_show(ei->box);

  return ei;
}

static void free_editor(editinfo* ei)
{
  appinfo* ai = ei->app;

  if (ai) {
    ai->edits = g_list_remove(ai->edits, ei);
    if (ai->active_edit == ei) {
      ai->active_edit = NULL;
      if (ai->edits)
	select_editor(ai->edits->data);
    }
  }

  g_free(ei->filename);
  g_free(ei->headers);

  g_tree_destroy(ei->bitmap_tree);

  if (ei->box)
    gtk_widget_destroy(ei->box);

  g_free(ei);
}

fbfontchar* editor_add_character(editinfo* ei, const fbfontchar* bm)
{
  fbfontchar* nbm = g_new0(fbfontchar, 1);
  *nbm = *bm;
  g_tree_insert(ei->bitmap_tree, GINT_TO_POINTER(bm->value), nbm);
  return nbm;
}

static void editor_commit(editinfo* ei)
{
  if (ei->current_bitmap == &ei->tmp_bitmap
      && ei->current_bitmap->value
      && ei->current_modified) {
    ei->current_bitmap = editor_add_character(ei, ei->current_bitmap);
    editor_set_current_modified(ei, FALSE);
  }
}

static void editor_select_character(gpointer eip, gpointer cp)
{
  editinfo* ei = eip;
  GtkAdjustment *adj;
  fbfontchar* bm;
  gunichar c = GPOINTER_TO_INT(cp);

  editor_commit(ei);

  if (c && c != '\n' && gucharmap_unichar_validate(c)) {
    bm = g_tree_lookup(ei->bitmap_tree, GINT_TO_POINTER(c));
    if (bm) {
      ei->current_bitmap = bm;
      if (!ei->current_bitmap->custom
	  || find_builtin_bitmap(ei, ei->current_bitmap))
	gtk_widget_set_sensitive(ei->cb_builtin, TRUE);
      else
	gtk_widget_set_sensitive(ei->cb_builtin, FALSE);
    }
    else {
      ei->current_bitmap = &ei->tmp_bitmap;
      clear_bitmap(&ei->tmp_bitmap);
      ei->current_bitmap->width = ei->default_width;
      ei->current_bitmap->value = c;

      if (load_builtin_bitmap(ei, ei->current_bitmap))
	gtk_widget_set_sensitive(ei->cb_builtin, TRUE);
      else
	gtk_widget_set_sensitive(ei->cb_builtin, FALSE);
    }
    ei->active = TRUE;

    if (!ei->current_bitmap->custom) {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ei->cb_builtin), TRUE);
      gtk_widget_set_sensitive(ei->da_edit, FALSE);
      gtk_widget_set_sensitive(ei->sb_width, FALSE);
    }
    else {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ei->cb_builtin), FALSE);
      gtk_widget_set_sensitive(ei->da_edit, TRUE);
      gtk_widget_set_sensitive(ei->sb_width, TRUE);
    }
  }
  else {
    ei->current_bitmap = &ei->tmp_bitmap;
    clear_bitmap(&ei->tmp_bitmap);
    ei->current_bitmap->value = c;
    ei->active = FALSE;
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ei->cb_builtin), FALSE);
    gtk_widget_set_sensitive(ei->da_edit, FALSE);
    gtk_widget_set_sensitive(ei->cb_builtin, FALSE);
    gtk_widget_set_sensitive(ei->sb_width, FALSE);
  }

  adj = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(ei->sb_width));
  gtk_adjustment_set_value(adj, ei->current_bitmap->width);

  ei->autometric = TRUE;
  editor_set_current_modified(ei, FALSE);

  gtk_widget_queue_draw(ei->da_edit);
  gtk_widget_queue_draw(ei->da_preview);
}


static GtkWidget* new_gnome_style_frame(const gchar* label, GtkWidget* contents)
{
  GtkWidget *frame, *align;
  char *str;

  str = g_strconcat("<b>", label, "</b>", NULL);
  frame = gtk_frame_new(str);
  g_free(str);

  g_object_set(gtk_frame_get_label_widget(GTK_FRAME(frame)),
	       "use-markup", TRUE, NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_NONE);

  align = gtk_alignment_new(0.5, 0.5, 1.0, 1.0);
  gtk_alignment_set_padding(GTK_ALIGNMENT(align), 0, 0, 12, 0);
  gtk_widget_show(align);
  gtk_container_add(GTK_CONTAINER(frame), align);
  gtk_container_add(GTK_CONTAINER(align), contents);
  gtk_widget_show(frame);

  return frame;
}

/**************** Main UI actions ****************/

static void set_active_char(gpointer data)
{
  appinfo* ai = data;
  gunichar c;
  c = charmap_get_char(ai->charmap);
  g_list_foreach(ai->edits, editor_select_character, GINT_TO_POINTER(c));
}

static void initial_allocate(GtkWidget* w G_GNUC_UNUSED,
			     GtkAllocation* alloc G_GNUC_UNUSED,
			     gpointer data)
{
  editinfo* ei = data;
  g_signal_handlers_disconnect_by_func(w, G_CALLBACK(initial_allocate), data);
  select_editor(ei);
}

static void new_font(appinfo* ai, int charset, int cell_width,
		     int cell_height, int min_row_height,
		     int baseline, int default_width)
{
  editinfo* ei;
  gunichar c;

  c = charmap_get_char(ai->charmap);

  ei = new_editor(ai, charset, cell_width, cell_height, min_row_height,
		  baseline, default_width);
  load_builtin_bitmaps(ei);
  editor_select_character(ei, GINT_TO_POINTER(c));
  ai->edits = g_list_append(ai->edits, ei);
  gtk_box_pack_start(GTK_BOX(ai->editbox), ei->box, FALSE, FALSE, 0);
  g_signal_connect(ei->box, "size-allocate", G_CALLBACK(initial_allocate), ei);
  gtk_widget_grab_focus(ei->da_edit);
}

typedef struct _newfontwidgets {
  GtkWidget *sb_ascent;
  GtkWidget *sb_descent;
  GtkWidget *sb_rowspacing;
} newfontwidgets;

static void custom_mode_changed(GtkToggleButton* tb, gpointer data)
{
  newfontwidgets* nfw = data;
  gboolean on = gtk_toggle_button_get_active(tb);
  gtk_widget_set_sensitive(nfw->sb_ascent, on);
  gtk_widget_set_sensitive(nfw->sb_descent, on);
  gtk_widget_set_sensitive(nfw->sb_rowspacing, on);
}

static void ascent_descent_changed(GtkAdjustment* adj,
				   gpointer data)
{
  newfontwidgets* nfw = data;
  GtkAdjustment *adj_a, *adj_d;
  gdouble val_a, val_d;

  adj_a = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(nfw->sb_ascent));
  adj_d = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(nfw->sb_descent));
  val_a = gtk_adjustment_get_value(adj_a);
  val_d = gtk_adjustment_get_value(adj_d);

  if (val_a + val_d > FB_MAX_FONT_HEIGHT) {
    if (adj == adj_a)
      gtk_adjustment_set_value(adj_d, FB_MAX_FONT_HEIGHT - val_a);
    else
      gtk_adjustment_set_value(adj_a, FB_MAX_FONT_HEIGHT - val_d);
  }
}

static void action_new(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  appinfo* ai = data;
  GtkWidget *dlg, *frame, *vbox, *rb_small, *rb_large, *rb_custom,
    *align, *tbl, *lbl;
  newfontwidgets nfw;
  GtkAdjustment *adj_a, *adj_d;
  gdouble va, vd, vr;

  dlg = gtk_dialog_new_with_buttons("New Font", GTK_WINDOW(ai->mainwin),
				    GTK_DIALOG_MODAL,
				    GTK_STOCK_CANCEL,
				    GTK_RESPONSE_CANCEL,
				    GTK_STOCK_NEW,
				    GTK_RESPONSE_ACCEPT,
				    NULL);
  gtk_dialog_set_alternative_button_order(GTK_DIALOG(dlg),
					  GTK_RESPONSE_ACCEPT,
					  GTK_RESPONSE_CANCEL,
					  -1);
  gtk_dialog_set_default_response(GTK_DIALOG(dlg),
				  GTK_RESPONSE_ACCEPT);

  vbox = gtk_vbox_new(FALSE, 10);
  {
    rb_small = gtk_radio_button_new_with_mnemonic
      (NULL, "_Small (5 pixels)");
    gtk_box_pack_start(GTK_BOX(vbox), rb_small, TRUE, TRUE, 0);
    gtk_widget_show(rb_small);

    rb_large = gtk_radio_button_new_with_mnemonic_from_widget
      (GTK_RADIO_BUTTON(rb_small), "_Large (7 pixels)");
    gtk_box_pack_start(GTK_BOX(vbox), rb_large, TRUE, TRUE, 0);
    gtk_widget_show(rb_large);

    rb_custom = gtk_radio_button_new_with_mnemonic_from_widget
      (GTK_RADIO_BUTTON(rb_small), "C_ustom");
    gtk_box_pack_start(GTK_BOX(vbox), rb_custom, TRUE, TRUE, 0);
    gtk_widget_show(rb_custom);

    align = gtk_alignment_new(0.5, 0.5, 1.0, 1.0);
    gtk_alignment_set_padding(GTK_ALIGNMENT(align), 0, 0, 12, 0);
    {
      tbl = gtk_table_new(3, 2, FALSE);
      gtk_table_set_row_spacings(GTK_TABLE(tbl), 6);
      gtk_table_set_col_spacings(GTK_TABLE(tbl), 12);
      {
	lbl = gtk_label_new_with_mnemonic("_Ascent:");
	gtk_misc_set_alignment(GTK_MISC(lbl), LABEL_X_ALIGN, 0.5);
	gtk_table_attach(GTK_TABLE(tbl), lbl, 0, 1, 0, 1,
			 GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(lbl);

	nfw.sb_ascent = gtk_spin_button_new_with_range
	  (0.0, (gdouble) FB_MAX_FONT_HEIGHT, 1.0);
	gtk_label_set_mnemonic_widget(GTK_LABEL(lbl), nfw.sb_ascent);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(nfw.sb_ascent), 6.0);
	gtk_entry_set_activates_default(GTK_ENTRY(nfw.sb_ascent), TRUE);
	gtk_widget_set_sensitive(nfw.sb_ascent, FALSE);
	gtk_table_attach(GTK_TABLE(tbl), nfw.sb_ascent, 1, 2, 0, 1,
			 GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(nfw.sb_ascent);

	lbl = gtk_label_new_with_mnemonic("_Descent:");
	gtk_misc_set_alignment(GTK_MISC(lbl), LABEL_X_ALIGN, 0.5);
	gtk_table_attach(GTK_TABLE(tbl), lbl, 0, 1, 1, 2,
			 GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(lbl);

	nfw.sb_descent = gtk_spin_button_new_with_range
	  (0.0, (gdouble) FB_MAX_FONT_HEIGHT, 1.0);
	gtk_label_set_mnemonic_widget(GTK_LABEL(lbl), nfw.sb_descent);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(nfw.sb_descent), 1.0);
	gtk_entry_set_activates_default(GTK_ENTRY(nfw.sb_descent), TRUE);
	gtk_widget_set_sensitive(nfw.sb_descent, FALSE);
	gtk_table_attach(GTK_TABLE(tbl), nfw.sb_descent, 1, 2, 1, 2,
			 GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(nfw.sb_descent);

	lbl = gtk_label_new_with_mnemonic("Minimum _row spacing:");
	gtk_misc_set_alignment(GTK_MISC(lbl), LABEL_X_ALIGN, 0.5);
	gtk_table_attach(GTK_TABLE(tbl), lbl, 0, 1, 2, 3,
			 GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(lbl);

	nfw.sb_rowspacing = gtk_spin_button_new_with_range
	  (0.0, (gdouble) FB_MAX_FONT_HEIGHT + 1, 1.0);
	gtk_label_set_mnemonic_widget(GTK_LABEL(lbl), nfw.sb_rowspacing);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(nfw.sb_rowspacing), 6.0);
	gtk_entry_set_activates_default(GTK_ENTRY(nfw.sb_rowspacing), TRUE);
	gtk_widget_set_sensitive(nfw.sb_rowspacing, FALSE);
	gtk_table_attach(GTK_TABLE(tbl), nfw.sb_rowspacing, 1, 2, 2, 3,
			 GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(nfw.sb_rowspacing);
      }
      gtk_container_add(GTK_CONTAINER(align), tbl);
      gtk_widget_show(tbl);
    }
    gtk_box_pack_start(GTK_BOX(vbox), align, TRUE, TRUE, 0);
    gtk_widget_show(align);
  }
  gtk_widget_show(vbox);

  frame = new_gnome_style_frame("Font Style", vbox);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox), frame, TRUE, TRUE, 0);

  g_signal_connect(rb_custom, "toggled",
		   G_CALLBACK(custom_mode_changed), &nfw);

  adj_a = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(nfw.sb_ascent));
  g_signal_connect(adj_a, "value-changed",
		   G_CALLBACK(ascent_descent_changed), &nfw);

  adj_d = gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(nfw.sb_descent));
  g_signal_connect(adj_d, "value-changed",
		   G_CALLBACK(ascent_descent_changed), &nfw);

  if (gtk_dialog_run(GTK_DIALOG(dlg)) == GTK_RESPONSE_ACCEPT) {
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(rb_small)))
      new_font(ai, ENCODING_TI83P_SMALL, 8, 7, 6, 5, 0);
    else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(rb_large)))
      new_font(ai, ENCODING_TI83P_LARGE, 8, 8, 8, 6, 6);
    else {
      va = gtk_adjustment_get_value(adj_a);
      vd = gtk_adjustment_get_value(adj_d);
      vr = gtk_adjustment_get_value
	(gtk_spin_button_get_adjustment(GTK_SPIN_BUTTON(nfw.sb_rowspacing)));
      new_font(ai, ENCODING_UNKNOWN, 8, va + vd, vr, va - 1, 0);
    }
  }
  gtk_widget_destroy(dlg);
}

static int parse_header(char* line, char** valptr)
{
  char* p;

  if ((p = strchr(line, '#')))
    *p = 0;

  p = line;
  while (*p && *p != ':' && *p != '\n' && *p != '\r')
    p++;

  if (*p != ':' || p == line)
    return 0;

  *p = 0;
  p++;
  while (*p == ' ' || *p == '\t')
    p++;
  *valptr = p;
  while (*p && *p != '\n' && *p != '\r')
    p++;
  *p = 0;

  return 1;
}

static void open_font(appinfo* ai, const gchar* fname)
{
  editinfo* ei;
  FILE* f;
  char buf[256];
  int charset, cell_width, cell_height, min_row_height, baseline, default_width;
  char *headers, *p, *value, *name;
  gunichar v;
  unsigned int x;
  fbfontchar newbm;
  gunichar c;
  int i, n;

  f = g_fopen(fname, "rb");
  if (!f) {
    display_io_error(ai->mainwin, fname, TRUE, errno);
    return;
  }

  headers = NULL;

  charset = ENCODING_UNKNOWN;
  cell_width = 8;
  cell_height = 0;
  min_row_height = 0;
  baseline = 0;
  default_width = 0;

  while (fgets(buf, sizeof(buf), f)) {
    if (buf[0] == 0 || buf[0] == '\n' || buf[0] == '\r')
      break;

    if (headers) {
      p = g_strconcat(headers, buf, NULL);
      g_free(headers);
      headers = p;
    }
    else
      headers = g_strdup(buf);

    if (parse_header(buf, &value)) {
      if (!strcmp(buf, "Default-Encoding")) {
	if (!strcmp(value, "TI83P-Large"))
	  charset = ENCODING_TI83P_LARGE;
	else if (!strcmp(value, "TI83P-Small"))
	  charset = ENCODING_TI83P_SMALL;
	else
	  charset = ENCODING_UNKNOWN;
      }
      else if (!strcmp(buf, "Cell-Height"))
	cell_height = strtol(value, &value, 10);
      else if (!strcmp(buf, "Min-Row-Height"))
	min_row_height = strtol(value, &value, 10);
      else if (!strcmp(buf, "Baseline"))
	baseline = strtol(value, &value, 10);
      else if (!strcmp(buf, "Default-Width"))
	default_width = strtol(value, &value, 10);
    }
  }

  if (!cell_height) {
    name = filename_to_display(fname);
    messagebox01(ai->mainwin, GTK_MESSAGE_ERROR, "Invalid font file",
		 "The file %s is not a valid Flashbook font file.", name);
    g_free(name);
    g_free(headers);
    fclose(f);
    return;
  }

  ei = new_editor(ai, charset, cell_width, cell_height, min_row_height,
		  baseline, default_width);
  ei->headers = headers;
  ei->filename = g_strdup(fname);

  name = filename_to_display(fname);
  gtk_frame_set_label(GTK_FRAME(gtk_bin_get_child(GTK_BIN(ei->box))), name);
  g_free(name);

  while (fgets(buf, sizeof(buf), f)) {
    if ((p = strchr(buf, '#')))
      *p = 0;
    v = strtol(buf, &p, 16);

    if (p == buf || v == 0)
      continue;

    newbm.value = v;
    newbm.custom = 1;
    newbm.allowover = 0;
    while (*p && (*p < '0' || *p > '9')) {
      if (*p == 'd')
	newbm.custom = 0;
      if (*p == 'o')
	newbm.allowover = 1;
      p++;
    }

    newbm.width = strtol(p, &p, 10);

    while (*p == ' ')
      p++;

    for (i = 0; i < ei->cell_height; i++) {
      if (0 < sscanf(p, "%02X%n", &x, &n)) {
	newbm.bitmap[i] = x;
	p += n;
      }
    }
    editor_add_character(ei, &newbm);
  }

  fclose(f);

  c = charmap_get_char(ai->charmap);
  editor_select_character(ei, GINT_TO_POINTER(c));
  ai->edits = g_list_append(ai->edits, ei);
  gtk_box_pack_start(GTK_BOX(ai->editbox), ei->box, FALSE, FALSE, 0);
  select_editor(ei);
  gtk_widget_grab_focus(ei->da_edit);
}

struct savefontinfo {
  editinfo* ei;
  FILE* f;
};

static gboolean print_editor_bitmap(gpointer key G_GNUC_UNUSED,
				    gpointer value,
				    gpointer data)
{
  struct savefontinfo* sfi = data;
  fbfontchar* bm = value;
  int i;

  fprintf(sfi->f, "%5X", (unsigned int) bm->value);

  if (!bm->custom || bm->allowover) {
    fputc(' ', sfi->f);
    if (!bm->custom)
      fputc('d', sfi->f);
    if (bm->allowover)
      fputc('o', sfi->f);
  }

  fprintf(sfi->f, " %d ", bm->width);

  for (i = 0; i < sfi->ei->cell_height; i++)
    fprintf(sfi->f, "%02X", bm->bitmap[i]);

  fputc('\n', sfi->f);

  return FALSE;
}

static gboolean save_font(editinfo* ei, const gchar* fname)
{
  struct savefontinfo sfi;
  char* name;

  editor_commit(ei);

  sfi.ei = ei;
  sfi.f = g_fopen(fname, "wt");
  if (!sfi.f) {
    display_io_error(ei->app->mainwin, fname, FALSE, errno);
    return FALSE;
  }

  if (ei->headers)
    fputs(ei->headers, sfi.f);
  else {
    fprintf(sfi.f, "Flashbook-Version: %s\n", PACKAGE_VERSION);

    fprintf(sfi.f, "Default-Encoding: %s\n",
	    (ei->charset == ENCODING_TI83P_SMALL
	     ? "TI83P-Small"
	     : (ei->charset == ENCODING_TI83P_LARGE
		? "TI83P-Large"
		: "Custom")));
    fprintf(sfi.f, "Cell-Width: %d\n", ei->cell_width);
    fprintf(sfi.f, "Cell-Height: %d\n", ei->cell_height);
    fprintf(sfi.f, "Min-Row-Height: %d\n", ei->min_row_height);
    fprintf(sfi.f, "Baseline: %d\n", ei->baseline);
    fprintf(sfi.f, "Default-Width: %d\n", ei->default_width);
  }
  fputs("\n", sfi.f);

  g_tree_foreach(ei->bitmap_tree, print_editor_bitmap, &sfi);

  if (ferror(sfi.f)) {
    fclose(sfi.f);
    display_io_error(ei->app->mainwin, fname, FALSE, errno);
    return FALSE;
  }
  if (fclose(sfi.f)) {
    display_io_error(ei->app->mainwin, fname, FALSE, errno);
    return FALSE;
  }

  if (fname != ei->filename) {
    g_free(ei->filename);
    ei->filename = g_strdup(fname);

    name = filename_to_display(fname);
    gtk_frame_set_label(GTK_FRAME(gtk_bin_get_child(GTK_BIN(ei->box))), name);
    g_free(name);
  }

  editor_set_current_modified(ei, FALSE);
  ei->modified = FALSE;
  return TRUE;
}

static void action_open(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  appinfo* ai = data;
  gchar* fname;

  fname = run_file_chooser("Open font file",
			   GTK_WINDOW(ai->mainwin),
			   GTK_FILE_CHOOSER_ACTION_OPEN,
			   NULL, NULL, NULL,
			   "Flashbook Font Files|*.fbf|");

  if (fname) {
    open_font(ai, fname);
    g_free(fname);
  }
}

static gboolean do_save(editinfo* ei, gboolean saveas)
{
  gchar* fname;

  if (ei->filename && !saveas)
    return save_font(ei, ei->filename);

  fname = run_file_chooser("Save font file",
			   GTK_WINDOW(ei->app->mainwin),
			   GTK_FILE_CHOOSER_ACTION_SAVE,
			   ei->filename,
			   "untitled.fbf",
			   NULL,
			   "Flashbook Font File|*.fbf|");
  if (!fname)
    return FALSE;

  if (!save_font(ei, fname)) {
    g_free(fname);
    return FALSE;
  }

  g_free(fname);
  return TRUE;
}

static gboolean prompt_save(editinfo* ei)
{
  GtkWidget* dlg;
  char* p;

  if (!ei->modified)
    return TRUE;

  if (ei->filename)
    p = filename_to_display(ei->filename);
  else
    p = NULL;

  dlg = gtk_message_dialog_new(GTK_WINDOW(ei->app->mainwin), GTK_DIALOG_MODAL,
			       GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE,
			       "Save changes to %s before closing?",
			       (p ? p : "untitled font"));
  g_free(p);

  gtk_message_dialog_format_secondary_markup
    (GTK_MESSAGE_DIALOG(dlg),
     "Your changes will be lost if you don't save them.");

  gtk_dialog_add_buttons(GTK_DIALOG(dlg),
			 "Close _Without Saving", GTK_RESPONSE_REJECT,
			 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			 GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
			 NULL);
  gtk_dialog_set_alternative_button_order(GTK_DIALOG(dlg),
					  GTK_RESPONSE_ACCEPT,
					  GTK_RESPONSE_REJECT,
					  GTK_RESPONSE_CANCEL,
					  -1);
  gtk_dialog_set_default_response(GTK_DIALOG(dlg), GTK_RESPONSE_ACCEPT);

  switch (gtk_dialog_run(GTK_DIALOG(dlg))) {
  case GTK_RESPONSE_ACCEPT:
    gtk_widget_destroy(dlg);
    return do_save(ei, FALSE);

  case GTK_RESPONSE_REJECT:
    gtk_widget_destroy(dlg);
    return TRUE;

  default:
    gtk_widget_destroy(dlg);
    return FALSE;
  }
}

static void action_save(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  appinfo* ai = data;
  if (ai->active_edit)
    do_save(ai->active_edit, FALSE);
}

static void action_save_as(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  appinfo* ai = data;
  if (ai->active_edit)
    do_save(ai->active_edit, TRUE);
}

static void action_close(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  appinfo* ai = data;

  if (ai->active_edit) {
    if (!prompt_save(ai->active_edit))
      return;

    free_editor(ai->active_edit);
  }
}

static void action_quit(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  appinfo* ai = data;
  editinfo* ei;
  GList *p;

  for (p = ai->edits; p; p = p->next) {
    ei = p->data;
    if (!prompt_save(ei))
      return;
  }

  gtk_widget_destroy(ai->mainwin);
}

static void action_choose_font(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  appinfo* ai = data;
  GtkWidget* fontsel;
  gchar* fontname;

  fontsel = gtk_font_selection_dialog_new("Select Reference Font");

  fontname = charmap_get_font(ai->charmap);
  gtk_font_selection_dialog_set_font_name(GTK_FONT_SELECTION_DIALOG(fontsel),
					  fontname);
  g_free(fontname);

  gtk_window_set_transient_for(GTK_WINDOW(fontsel), GTK_WINDOW(ai->mainwin));

  if (gtk_dialog_run(GTK_DIALOG(fontsel)) == GTK_RESPONSE_OK) {
    fontname = gtk_font_selection_dialog_get_font_name
      (GTK_FONT_SELECTION_DIALOG(fontsel));
    charmap_set_font(ai->charmap, fontname);
    g_free(fontname);
  }
  gtk_widget_destroy(fontsel);
}

static void run_sample_text_dialog(appinfo* ai)
{
  GtkWidget *dlg, *frame, *hbox, *hbox2, *lbl, *ent_left, *ent_right;

  dlg = gtk_dialog_new_with_buttons("Sample Text", GTK_WINDOW(ai->mainwin),
				    GTK_DIALOG_MODAL,
				    GTK_STOCK_CANCEL,
				    GTK_RESPONSE_CANCEL,
				    GTK_STOCK_OK,
				    GTK_RESPONSE_ACCEPT,
				    NULL);
  gtk_dialog_set_alternative_button_order(GTK_DIALOG(dlg),
					  GTK_RESPONSE_ACCEPT,
					  GTK_RESPONSE_CANCEL,
					  -1);
  gtk_dialog_set_default_response(GTK_DIALOG(dlg),
				  GTK_RESPONSE_ACCEPT);


  hbox = gtk_hbox_new(FALSE, 24);
  {
    hbox2 = gtk_hbox_new(FALSE, 12);
    {
      lbl = gtk_label_new_with_mnemonic("_Before:");
      gtk_misc_set_alignment(GTK_MISC(lbl), LABEL_X_ALIGN, 0.5);
      gtk_box_pack_start(GTK_BOX(hbox2), lbl, FALSE, FALSE, 0);
      gtk_widget_show(lbl);

      ent_left = gtk_entry_new();
      if (ai->preview_left_text)
	gtk_entry_set_text(GTK_ENTRY(ent_left), ai->preview_left_text);
      gtk_entry_set_width_chars(GTK_ENTRY(ent_left), 10);
      gtk_entry_set_activates_default(GTK_ENTRY(ent_left), TRUE);
      gtk_box_pack_start(GTK_BOX(hbox2), ent_left, TRUE, TRUE, 0);
      gtk_label_set_mnemonic_widget(GTK_LABEL(lbl), ent_left);
      gtk_widget_show(ent_left);
    }
    gtk_box_pack_start(GTK_BOX(hbox), hbox2, TRUE, TRUE, 0);
    gtk_widget_show(hbox2);

    hbox2 = gtk_hbox_new(FALSE, 12);
    {
      lbl = gtk_label_new_with_mnemonic("_After:");
      gtk_misc_set_alignment(GTK_MISC(lbl), LABEL_X_ALIGN, 0.5);
      gtk_box_pack_start(GTK_BOX(hbox2), lbl, FALSE, FALSE, 0);
      gtk_widget_show(lbl);

      ent_right = gtk_entry_new();
      if (ai->preview_right_text)
	gtk_entry_set_text(GTK_ENTRY(ent_right), ai->preview_right_text);
      gtk_entry_set_width_chars(GTK_ENTRY(ent_right), 10);
      gtk_entry_set_activates_default(GTK_ENTRY(ent_right), TRUE);
      gtk_box_pack_start(GTK_BOX(hbox2), ent_right, TRUE, TRUE, 0);
      gtk_label_set_mnemonic_widget(GTK_LABEL(lbl), ent_right);
      gtk_widget_show(ent_right);
    }
    gtk_box_pack_start(GTK_BOX(hbox), hbox2, TRUE, TRUE, 0);
    gtk_widget_show(hbox2);
  }
  gtk_widget_show(hbox);

  frame = new_gnome_style_frame("Preview Sample Text", hbox);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox), frame, TRUE, TRUE, 0);

  if (gtk_dialog_run(GTK_DIALOG(dlg)) == GTK_RESPONSE_ACCEPT) {
    g_free(ai->preview_left_text);
    ai->preview_left_text = g_strdup(gtk_entry_get_text(GTK_ENTRY(ent_left)));
    g_free(ai->preview_right_text);
    ai->preview_right_text = g_strdup(gtk_entry_get_text(GTK_ENTRY(ent_right)));
    g_list_foreach(ai->edits, editor_update_preview, NULL);
  }
  gtk_widget_destroy(dlg);
}

static void action_set_sample_text(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  appinfo* ai = data;
  run_sample_text_dialog(ai);
}

static void action_prev(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  appinfo* ai = data;
  gunichar c;

  c = charmap_get_char(ai->charmap);
  if (c > 0)
    charmap_set_char(ai->charmap, c - 1);

  if (ai->active_edit && ai->active_edit->da_edit)
    gtk_widget_grab_focus(ai->active_edit->da_edit);
}

static void action_next(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  appinfo* ai = data;
  gunichar c;

  c = charmap_get_char(ai->charmap);
  if (c < 0x10ffff)
    charmap_set_char(ai->charmap, c + 1);

  if (ai->active_edit && ai->active_edit->da_edit)
    gtk_widget_grab_focus(ai->active_edit->da_edit);
}

static gboolean is_missing(appinfo* ai, gunichar c)
{
  GUnicodeType t = gucharmap_unichar_type(c);
  GList* l;
  editinfo* ei;

  /* skip invalid characters and categories we don't care about */
  if (t == G_UNICODE_CONTROL || t == G_UNICODE_FORMAT
      || t == G_UNICODE_UNASSIGNED || t == G_UNICODE_SURROGATE
      || t == G_UNICODE_LINE_SEPARATOR
      || t == G_UNICODE_PARAGRAPH_SEPARATOR)
    return FALSE;

  for (l = ai->edits; l; l = l->next) {
    ei = l->data;
    if (c != ei->current_bitmap->value
	&& !g_tree_lookup(ei->bitmap_tree, GINT_TO_POINTER(c)))
      return TRUE;
  }

  return FALSE;
}

static void action_prev_missing(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  appinfo* ai = data;
  gunichar c;

  c = charmap_get_char(ai->charmap);

  do {
    if (c == 0)
      return;
    c--;
  } while (!is_missing(ai, c));

  charmap_set_char(ai->charmap, c);

  if (ai->active_edit && ai->active_edit->da_edit)
    gtk_widget_grab_focus(ai->active_edit->da_edit);
}

static void action_next_missing(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  appinfo* ai = data;
  gunichar c;

  c = charmap_get_char(ai->charmap);

  do {
    if (c >= 0x10ffff)
      return;
    c++;
  } while (!is_missing(ai, c));

  charmap_set_char(ai->charmap, c);

  if (ai->active_edit && ai->active_edit->da_edit)
    gtk_widget_grab_focus(ai->active_edit->da_edit);
}

static void action_about(GtkAction* action G_GNUC_UNUSED,
			 gpointer data G_GNUC_UNUSED)
{
  appinfo* ai = data;
  show_about_dialog(GTK_WINDOW(ai->mainwin));
}

static gboolean window_delete(GtkWidget* widget G_GNUC_UNUSED,
			      GdkEvent* event G_GNUC_UNUSED,
			      gpointer data)
{
  appinfo* ai = data;
  editinfo* ei;
  GList* p;

  for (p = ai->edits; p; p = p->next) {
    ei = p->data;
    if (!prompt_save(ei))
      return TRUE;
  }

  return FALSE;
}

static void window_destroy(GtkWidget* widget G_GNUC_UNUSED,
			   gpointer data)
{
  appinfo* ai = data;
  editinfo* ei;
  GList *p, *q;

  p = ai->edits;
  while (p) {
    q = p->next;
    ei = p->data;
    free_editor(ei);
    p = q;
  }

#ifdef USE_UNIFIED_MDI
  ai->mainwin = NULL;
  fontedit_quit();
#else
  gtk_main_quit();
#endif
}

static const GtkActionEntry menuactions[] =
  {{ "file-menu", 0, "_File", 0, 0, 0 },

   { "new",     GTK_STOCK_NEW,     "_New Font...",          0,
     "Create a new font",
     G_CALLBACK(action_new) },

   { "open",    GTK_STOCK_OPEN,    "_Open Font...",         0,
     "Open an existing font",
     G_CALLBACK(action_open) },

   { "save",    GTK_STOCK_SAVE,    "_Save Font",            0,
     "Save the current font",
     G_CALLBACK(action_save) },

   { "save-as", GTK_STOCK_SAVE_AS, "Save _As...",    "<Shift><Control>S",
     "Save the current font to a new file",
     G_CALLBACK(action_save_as) },

   { "close",   GTK_STOCK_CLOSE,   "_Close Font",           0,
     "Close the current font",
     G_CALLBACK(action_close) },

   { "quit",    GTK_STOCK_QUIT,    0,                       0,
     "Quit Flashbook Font Editor",
     G_CALLBACK(action_quit) },

   { "view-menu", 0, "_View", 0, 0, 0 },

   { "choose-font", GTK_STOCK_SELECT_FONT, "_Reference Font...",   0,
     "Set the reference font displayed in the character picker",
     G_CALLBACK(action_choose_font) },

   { "set-sample-text", 0,         "_Sample Text...",       0,
     "Set the sample text displayed in the character preview",
     G_CALLBACK(action_set_sample_text) },

   { "go-menu", 0, "_Go", 0, 0, 0 },

   { "prev-char", GTK_STOCK_GO_BACK, "_Previous Character", "P",
     "Jump to the previous character",
     G_CALLBACK(action_prev) },

   { "next-char", GTK_STOCK_GO_FORWARD, "_Next Character",  "N",
     "Jump to the next character",
     G_CALLBACK(action_next) },

   { "prev-missing", GTK_STOCK_GOTO_FIRST, "P_revious Missing Character", "B",
     "Jump to the previous character not yet defined",
     G_CALLBACK(action_prev_missing) },

   { "next-missing", GTK_STOCK_GOTO_LAST, "N_ext Missing Character",  "F",
     "Jump to the next character not yet defined",
     G_CALLBACK(action_next_missing) },

   { "help-menu", 0, "_Help", 0, 0, 0 },

   { "about",   GTK_STOCK_ABOUT,   0,                       0,
     "About Flashbook Font Editor",
     G_CALLBACK(action_about) }};


#ifndef USE_UNIFIED_MDI

static const char uidesc[] =
  "<menubar name=\"menu-bar\">"
  " <menu action=\"file-menu\">"
  "  <menuitem action=\"new\"/>"
  "  <menuitem action=\"open\"/>"
  "  <menuitem action=\"save\"/>"
  "  <menuitem action=\"save-as\"/>"
  "  <separator/>"
  "  <menuitem action=\"close\"/>"
  "  <menuitem action=\"quit\"/>"
  " </menu>"
  " <menu action=\"view-menu\">"
  "  <menuitem action=\"choose-font\"/>"
  "  <menuitem action=\"set-sample-text\"/>"
  " </menu>"
  " <menu action=\"go-menu\">"
  "  <menuitem action=\"prev-char\"/>"
  "  <menuitem action=\"next-char\"/>"
  "  <separator/>"
  "  <menuitem action=\"prev-missing\"/>"
  "  <menuitem action=\"next-missing\"/>"
  " </menu>"
  " <menu action=\"help-menu\">"
  "  <menuitem action=\"about\" />"
  " </menu>"
  "</menubar>";

int main(int argc, char** argv)
{
  GtkWidget *vbox, *menubar, *sw, *vp;
  GtkActionGroup* actgrp;
  GtkUIManager* uimgr;
  GtkAccelGroup* accelgrp;
  GError* err = NULL;
  appinfo ai;
  int i;
  char* fname;

  fix_main_args(&argc, &argv);

  gtk_init(&argc, &argv);
  g_set_application_name("Flashbook Font Editor");

  if ((fname = fb_find_common_file(NULL, "fbfontedit.png"))) {
    gtk_window_set_default_icon_from_file(fname, NULL);
    fbfree(fname);
  }

  load_menurc_file("fontedit");

  memset(&ai, 0, sizeof(appinfo));

  ai.preview_left_text = g_strdup("Ab");
  ai.preview_right_text = g_strdup("cd");

  ai.mainwin = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_default_size(GTK_WINDOW(ai.mainwin), 600, 400);
  {
    vbox = gtk_vbox_new(FALSE, 0);
    {
      actgrp = gtk_action_group_new("Menu");
      gtk_action_group_add_actions(actgrp, menuactions,
				   G_N_ELEMENTS(menuactions), &ai);

      uimgr = gtk_ui_manager_new();
      gtk_ui_manager_insert_action_group(uimgr, actgrp, 0);
      accelgrp = gtk_ui_manager_get_accel_group(uimgr);
      gtk_window_add_accel_group(GTK_WINDOW(ai.mainwin), accelgrp);
      if (!gtk_ui_manager_add_ui_from_string(uimgr, uidesc, -1, &err))
	g_error("Failed to create menus: %s", err->message);
      menubar = gtk_ui_manager_get_widget(uimgr, "/menu-bar");
      gtk_box_pack_start(GTK_BOX(vbox), menubar, FALSE, FALSE, 0);

      ai.charmap = charmap_new();
      gtk_box_pack_start(GTK_BOX(vbox), ai.charmap, TRUE, TRUE, 0);
      gtk_widget_show(ai.charmap);

      sw = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
				     GTK_POLICY_AUTOMATIC,
				     GTK_POLICY_NEVER);
      {
	vp = gtk_viewport_new(NULL, NULL);
	gtk_viewport_set_shadow_type(GTK_VIEWPORT(vp), GTK_SHADOW_NONE);
	gtk_container_set_resize_mode(GTK_CONTAINER(vp), GTK_RESIZE_PARENT);
	{
	  ai.editbox = gtk_hbox_new(FALSE, 10);
	  gtk_container_add(GTK_CONTAINER(vp), ai.editbox);
	  gtk_widget_show(ai.editbox);
	}
	gtk_container_add(GTK_CONTAINER(sw), vp);
	gtk_widget_show(vp);

	ai.edithadj = gtk_viewport_get_hadjustment(GTK_VIEWPORT(vp));
      }
      gtk_box_pack_start(GTK_BOX(vbox), sw, FALSE, FALSE, 0);
      gtk_widget_show(sw);
    }
    gtk_container_add(GTK_CONTAINER(ai.mainwin), vbox);
    gtk_widget_show(vbox);
  }
  gtk_widget_show(ai.mainwin);

  for (i = 1; i < argc; i++)
    open_font(&ai, argv[i]);

  charmap_connect_active_char(ai.charmap, &set_active_char, &ai);
  g_signal_connect(ai.mainwin, "delete-event", G_CALLBACK(window_delete), &ai);
  g_signal_connect(ai.mainwin, "destroy", G_CALLBACK(window_destroy), &ai);

  gtk_main();

  save_menurc_file("fontedit");

  return 0;
}

#else  /* USE_UNIFIED_MDI */

static const char uidesc[] =
  "<toolbar name=\"fontedit-toolbar\">"
  " <toolitem action=\"new\"/>"
  " <toolitem action=\"open\"/>"
  " <separator/>"
  " <toolitem action=\"choose-font\"/>"
  " <separator/>"
  " <toolitem action=\"prev-char\"/>"
  " <toolitem action=\"next-char\"/>"
  " <toolitem action=\"prev-missing\"/>"
  " <toolitem action=\"next-missing\"/>"
  "</toolbar>";

static appinfo* mdi_app;

void fontedit_init(gboolean (*focusin)(), gboolean (*focusout)())
{
  GtkWidget *vbox, *toolbar, *sw, *vp;
  GtkActionGroup* actgrp;
  GtkUIManager* uimgr;
  GtkAccelGroup* accelgrp;
  GError* err = NULL;

  if (mdi_app)
    return;

  mdi_app = g_new0(appinfo, 1);

  mdi_app->mainwin = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(mdi_app->mainwin), "Font Editor");
  gtk_window_set_default_size(GTK_WINDOW(mdi_app->mainwin), 600, 400);
#ifdef HAVE_LIBIGE_MAC_INTEGRATION
  ige_mac_menu_connect_window_key_handler(GTK_WINDOW(mdi_app->mainwin));
#endif
  {
    vbox = gtk_vbox_new(FALSE, 0);
    {
      actgrp = gtk_action_group_new("Menu");
      gtk_action_group_add_actions(actgrp, menuactions,
				   G_N_ELEMENTS(menuactions), mdi_app);

      uimgr = gtk_ui_manager_new();
      gtk_ui_manager_insert_action_group(uimgr, actgrp, 0);
      accelgrp = gtk_ui_manager_get_accel_group(uimgr);
      gtk_window_add_accel_group(GTK_WINDOW(mdi_app->mainwin), accelgrp);
      if (!gtk_ui_manager_add_ui_from_string(uimgr, uidesc, -1, &err))
	g_error("Failed to create menus: %s", err->message);
      toolbar = gtk_ui_manager_get_widget(uimgr, "/fontedit-toolbar");
      gtk_toolbar_set_style(GTK_TOOLBAR(toolbar), GTK_TOOLBAR_ICONS);
      gtk_box_pack_start(GTK_BOX(vbox), toolbar, FALSE, FALSE, 0);

      mdi_app->charmap = charmap_new();
      gtk_box_pack_start(GTK_BOX(vbox), mdi_app->charmap, TRUE, TRUE, 0);
      gtk_widget_show(mdi_app->charmap);

      sw = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
				     GTK_POLICY_AUTOMATIC,
				     GTK_POLICY_NEVER);
      {
	vp = gtk_viewport_new(NULL, NULL);
	gtk_viewport_set_shadow_type(GTK_VIEWPORT(vp), GTK_SHADOW_NONE);
	gtk_container_set_resize_mode(GTK_CONTAINER(vp), GTK_RESIZE_PARENT);
	{
	  mdi_app->editbox = gtk_hbox_new(FALSE, 10);
	  gtk_container_add(GTK_CONTAINER(vp), mdi_app->editbox);
	  gtk_widget_show(mdi_app->editbox);
	}
	gtk_container_add(GTK_CONTAINER(sw), vp);
	gtk_widget_show(vp);

	mdi_app->edithadj = gtk_viewport_get_hadjustment(GTK_VIEWPORT(vp));
      }
      gtk_box_pack_start(GTK_BOX(vbox), sw, FALSE, FALSE, 0);
      gtk_widget_show(sw);
    }
    gtk_container_add(GTK_CONTAINER(mdi_app->mainwin), vbox);
    gtk_widget_show(vbox);
  }

  charmap_connect_active_char(mdi_app->charmap, &set_active_char, mdi_app);
  g_signal_connect(mdi_app->mainwin, "delete-event", G_CALLBACK(window_delete), mdi_app);
  g_signal_connect(mdi_app->mainwin, "destroy", G_CALLBACK(window_destroy), mdi_app);

  if (focusin)
    g_signal_connect(mdi_app->mainwin, "focus-in-event",
		     G_CALLBACK(focusin), NULL);
  if (focusout)
    g_signal_connect(mdi_app->mainwin, "focus-out-event",
		     G_CALLBACK(focusout), NULL);
}

void fontedit_new()
{
  if (mdi_app) {
    gtk_widget_show(mdi_app->mainwin); 
    action_new(NULL, mdi_app);
  }
}

void fontedit_open(const char* fname)
{
  if (mdi_app) {
    gtk_widget_show(mdi_app->mainwin);
    open_font(mdi_app, fname);
  }
}

void fontedit_save()
{
  if (mdi_app) {
    gtk_widget_show(mdi_app->mainwin);
    action_save(NULL, mdi_app);
  }
}

void fontedit_save_as()
{
  if (mdi_app) {
    gtk_widget_show(mdi_app->mainwin);
    action_save_as(NULL, mdi_app);
  }
}

void fontedit_close()
{
  if (mdi_app) {
    gtk_widget_show(mdi_app->mainwin);
    action_close(NULL, mdi_app);
  }
}

gboolean fontedit_prompt_save()
{
  editinfo* ei;
  GList *p;

  if (mdi_app) {
    for (p = mdi_app->edits; p; p = p->next) {
      ei = p->data;
      if (!prompt_save(ei))
	return FALSE;
    }
  }

  return TRUE;
}

void fontedit_quit()
{
  if (mdi_app) {
    if (mdi_app->mainwin)
      gtk_widget_destroy(mdi_app->mainwin);
    g_free(mdi_app);
    mdi_app = NULL;
  }
}

#endif
