/*
 * Flashbook GUI
 * Copyright (c) 2008-2010 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <fb.h>
#include <gtk/gtk.h>
#include "fbtextbuffer.h"
#include "creator.h"
#include "msgbox.h"
#include "fbgui.h"


/**************** Input character set detection ****************/

#define DETECT_SIZE 4096

/* Determine whether given string is a valid charset name known to
   GLib. */
static gboolean is_known_charset(const gchar* s)
{
  GIConv ictest;

  if (!s || !s[0])
    return FALSE;

  ictest = g_iconv_open("UTF-8", s);
  if (ictest != (GIConv) -1) {
    g_iconv_close(ictest);
    return TRUE;
  }
  else
    return FALSE;
}

/* Check for a string that could indicate the encoding.  PG texts, for
   instance, often include this information at the start of the
   file. */
static gchar* check_encoding_str(const gchar* p)
{
  gboolean allownum = FALSE;
  const gchar* q;
  gchar* result;

  if (!g_ascii_strncasecmp(p, "coding", 6))
    p += 6;
  else if (!g_ascii_strncasecmp(p, "charset", 7))
    p += 7;
  else if (!g_ascii_strncasecmp(p, "codepage", 8)) {
    allownum = TRUE;
    p += 8;
  }
  else
    return NULL;

  while (*p == ' ' || *p == '\t')
    p++;
  if (*p == ':' || *p == '=')
    p++;
  while (*p == ' ' || *p == '\t')
    p++;

  if (!g_ascii_isalpha(*p) && (!allownum || !g_ascii_isdigit(*p)))
    return NULL;

  q = p + 1;
  while (g_ascii_isalnum(*q) || *q == '-' || *q == '_' || *q == '.')
    q++;

  if (q == p + 1)
    return NULL;

  if (g_ascii_isdigit(*p)) {
    result = g_new(char, q - p + 3);
    result[0] = 'c';
    result[1] = 'p';
    strncpy(result + 2, p, q - p);
    result[q - p + 2] = 0;
  }
  else
    result = g_strndup(p, q - p);

  if (is_known_charset(result))
    return result;
  else {
    g_free(result);
    return NULL;
  }
}

/* Search first and last 4k of the file for the patterns above. */
static gchar* get_declared_encoding(const gchar* input, gsize length)
{
  gchar* p;
  gsize i;

  for (i = 0; i < length && i < DETECT_SIZE; i++) {
    if ((p = check_encoding_str(&input[i])))
      return p;
  }

  for (i = length - 10; i > 0 && i > length - DETECT_SIZE; i--) {
    if ((p = check_encoding_str(&input[i])))
      return p;
  }

  return NULL;
}

/* Attempt to automatically detect the text encoding. */
gchar* detect_text_encoding(const gchar* input, gsize length)
{
  gboolean could_be_ascii, could_be_utf8;
  gsize zero_odd, zero_even;
  gsize i;
  gchar* p;
  const gchar* kp;

  if (!(length % 2) && length > 4) {
    if (input[0] == '\xfe' && input[1] == '\xff')
      return g_strdup("UTF-16");
    else if (input[0] == '\xff' && input[1] == '\xfe') {
      if (!(length % 4) && input[2] == 0 && input[3] == 0)
	return g_strdup("UTF-32");
      else
	return g_strdup("UTF-16");
    }
    else if (!(length % 4) && input[0] == 0 && input[1] == 0
	     && input[2] == '\xfe' && input[3] == '\xff')
      return g_strdup("UTF-32");
  }

  could_be_ascii = TRUE;
  zero_odd = zero_even = 0;
  for (i = 0; i < length; i++) {
    if ((guchar) input[i] & 0x80)
      could_be_ascii = FALSE;
    else if (input[i] == 0) {
      could_be_ascii = FALSE;
      if (i & 1)
	zero_odd++;
      else
	zero_even++;
    }
  }

  if (!(length % 2) && (zero_odd > (length / 100 + 1)
			|| zero_even > (length / 100 + 1))) {
    if (zero_odd > zero_even)
      return g_strdup("UTF-16LE");
    else
      return g_strdup("UTF-16BE");
  }

  if (could_be_ascii)
    could_be_utf8 = TRUE;
  else
    could_be_utf8 = g_utf8_validate(input, length, NULL);

  p = get_declared_encoding(input, length);

  if (p) {
    if (!g_ascii_strcasecmp(p, "ascii")) {
      if (could_be_ascii)
	return p;
    }
    else if (!g_ascii_strcasecmp(p, "utf-8")
	     || !g_ascii_strcasecmp(p, "utf8")) {
      if (could_be_utf8)
	return p;
    }
    else
      return p;

    g_free(p);
  }

  if (could_be_ascii)
    return g_strdup("ASCII");
  else if (could_be_utf8)
    return g_strdup("UTF-8");
  else if (g_get_charset(&kp) && !could_be_utf8)
    return g_strdup("Windows-1252");
  else
    return g_strdup(kp);
}


/**************** Text unwrapping ****************/

struct linestat {
  gchar* end;
  gint width;			/* Monospaced width of line (obviously
				   this is approximate as the exact
				   width is font-dependent) */
  gint indent;			/* Monospaced width of indentation
				   (initial spaces and tabs) */
  gint first_word;		/* Monospaced width of first word
				   (including punctuation) */
  gint first_nonpunct;		/* Unicode type of first non-space,
				   non-punctuation character */
  gint nupper;			/* Number of uppercase and titlecase
				   letters */
  gint nlower;			/* Number of lowercase letters */
  gint nother;			/* Number of other letters, marks, and
				   unknown characters */
  gint eoltype;			/* UNKNOWN, LINE_BREAK, or
				   PARAGRAPH_BREAK */
  guint tabular : 1;		/* if line contains tabs or long
				   strings of spaces medially */
  guint endhyph : 1;		/* if line ends with a hyphen or dash
				   preceded by letter/number/symbol */
};

struct indstat {
  gsize count;
  gsize nupper;
  gdouble fupper;
};

enum { UNKNOWN, LINE_BREAK, PARAGRAPH_BREAK };

#define IGNORE 1

/* Fix up line endings and other cruft in the imported text.  (The
   text must be valid UTF-8.) */
void format_imported_text(gchar* buf, gboolean unwrap)
{
  gsize nlines, nlines_a;
  gsize i, j;
  struct linestat* lines;
  gchar* p;
  gchar* q;
  gunichar c;
  gint utype, utype_prev;
  gint width, indent, first_word, first_nonpunct;
  gint blank, in_first, nspaces, tabular, endhyph;
  gint maxindent;
  gsize chunkstart;
  gsize nupper, nlower, nother;
  gdouble fupper;
  struct indstat* indlevels;

  /* Check for stray byte order mark */

  if (buf[0] == '\xef' && buf[1] == '\xbb' && buf[2] == '\xbf') {
    buf[0] = buf[1] = buf[2] = IGNORE;
    p = buf+3;
  }
  else
    p = buf;

  /* Read through buffer and collect information about each line */

  nlines = 0;
  nlines_a = 64;
  lines = g_new(struct linestat, nlines_a);
  width = indent = first_word = first_nonpunct = 0;
  nupper = nlower = nother = 0;
  blank = 1;
  in_first = 0;
  nspaces = tabular = endhyph = 0;
  utype = G_UNICODE_CONTROL;

  for (; *p; p = g_utf8_next_char(p)) {
    c = g_utf8_get_char(p);
    utype_prev = utype;
    utype = g_unichar_type(c);

    if (c == '\n' || c == '\r'
	|| utype == G_UNICODE_LINE_SEPARATOR
	|| utype == G_UNICODE_PARAGRAPH_SEPARATOR) {
      if (c == '\r' && p[1] == '\n') {
	*p = IGNORE;
	p++;
      }

      lines[nlines].end = p;
      lines[nlines].width = width;
      lines[nlines].indent = indent;
      lines[nlines].first_word = first_word;
      lines[nlines].first_nonpunct = first_nonpunct;
      lines[nlines].nupper = nupper;
      lines[nlines].nlower = nlower;
      lines[nlines].nother = nother;
      lines[nlines].tabular = tabular;
      lines[nlines].endhyph = endhyph;

      if (c == '\n' || c == '\r')
	lines[nlines].eoltype = UNKNOWN;
      else
	lines[nlines].eoltype = PARAGRAPH_BREAK;

      nlines++;
      if (nlines >= nlines_a) {
	nlines_a = nlines * 2;
	lines = g_renew(struct linestat, lines, nlines_a);
      }

      width = indent = first_word = first_nonpunct = 0;
      nupper = nlower = nother = 0;
      blank = 1;
      in_first = 0;
      nspaces = tabular = endhyph = 0;
    }
    else {
      if (c == '\t') {
	/* change tab to space -- tabs aren't supported (yet) */
	*p = ' ';
	width = (width & ~7) + 8;
	utype = G_UNICODE_SPACE_SEPARATOR;
	if (!blank)
	  tabular = 1;
      }
      else if (utype == G_UNICODE_CONTROL) {
	/* other control characters aren't supported either */
	for (q = p; q < g_utf8_next_char(p); q++)
	  *q = IGNORE;
      }
      else if (utype != G_UNICODE_NON_SPACING_MARK) {
	if (g_unichar_iswide(c))
	  width += 2;
	else
	  width += 1;
      }

      if (c == ' ' && !blank) {
	nspaces++;
	if (nspaces > 3)
	  tabular = 1;
      }
      else {
	nspaces = 0;
      }

      if (utype == G_UNICODE_DASH_PUNCTUATION
	  && utype_prev != G_UNICODE_CONTROL
	  && utype_prev != G_UNICODE_FORMAT
	  && utype_prev != G_UNICODE_CONNECT_PUNCTUATION
	  && utype_prev != G_UNICODE_DASH_PUNCTUATION
	  && utype_prev != G_UNICODE_CLOSE_PUNCTUATION
	  && utype_prev != G_UNICODE_FINAL_PUNCTUATION
	  && utype_prev != G_UNICODE_INITIAL_PUNCTUATION
	  && utype_prev != G_UNICODE_OTHER_PUNCTUATION
	  && utype_prev != G_UNICODE_OPEN_PUNCTUATION
	  && utype_prev != G_UNICODE_LINE_SEPARATOR
	  && utype_prev != G_UNICODE_PARAGRAPH_SEPARATOR
	  && utype_prev != G_UNICODE_SPACE_SEPARATOR)
	endhyph = 1;
      else
	endhyph = 0;

      if (utype == G_UNICODE_SPACE_SEPARATOR) {
	if (blank)
	  indent = width;
	else if (in_first)
	  in_first = 0;
      }
      else if (utype != G_UNICODE_CONTROL) {
	blank = 0;
	if (utype == G_UNICODE_UNASSIGNED
	    || utype == G_UNICODE_PRIVATE_USE
	    || utype == G_UNICODE_LOWERCASE_LETTER
	    || utype == G_UNICODE_MODIFIER_LETTER
	    || utype == G_UNICODE_OTHER_LETTER
	    || utype == G_UNICODE_TITLECASE_LETTER
	    || utype == G_UNICODE_UPPERCASE_LETTER
	    || utype == G_UNICODE_COMBINING_MARK
	    || utype == G_UNICODE_ENCLOSING_MARK
	    || utype == G_UNICODE_NON_SPACING_MARK) {
	  if (utype == G_UNICODE_LOWERCASE_LETTER)
	    nlower++;
	  else if (utype == G_UNICODE_UPPERCASE_LETTER
		   || utype == G_UNICODE_TITLECASE_LETTER)
	    nupper++;
	  else
	    nother++;

	  if (!first_nonpunct) {
	    first_nonpunct = utype;
	    in_first = 1;
	  }
	}
	else if (utype == G_UNICODE_DECIMAL_NUMBER
		 || utype == G_UNICODE_LETTER_NUMBER
		 || utype == G_UNICODE_OTHER_NUMBER
		 || utype == G_UNICODE_CURRENCY_SYMBOL
		 || utype == G_UNICODE_MODIFIER_SYMBOL
		 || utype == G_UNICODE_MATH_SYMBOL
		 || utype == G_UNICODE_OTHER_SYMBOL) {
	  if (!first_nonpunct) {
	    first_nonpunct = utype;
	    in_first = 1;
	  }
	}

	if (in_first)
	  first_word = width - indent;
      }
    }
  }

  lines[nlines].end = p;
  lines[nlines].width = width;
  lines[nlines].indent = indent;
  lines[nlines].first_word = first_word;
  lines[nlines].first_nonpunct = first_nonpunct;
  lines[nlines].nupper = nupper;
  lines[nlines].nlower = nlower;
  lines[nlines].nother = nother;
  lines[nlines].tabular = tabular;
  lines[nlines].endhyph = endhyph;
  lines[nlines].eoltype = PARAGRAPH_BREAK;
  nlines++;

  if (unwrap) {

    /* Mark all blank lines (and lines containing nothing but
       punctuation or symbols) as paragraph separators.  Also mark
       lines that appear to contain tabular data. */

    for (i = 0; i < nlines; i++) {
      if ((lines[i].nupper == 0
	   && lines[i].nlower == 0
	   && lines[i].nother == 0)
	  || lines[i].width > 150
	  || lines[i].tabular) {
	if (lines[i].eoltype == UNKNOWN)
	  lines[i].eoltype = PARAGRAPH_BREAK;
	if (i > 0 && lines[i-1].eoltype == UNKNOWN)
	  lines[i-1].eoltype = PARAGRAPH_BREAK;
      }
    }


    /* Analyze the "chunks" left after applying the rules above.  We
       attempt to classify each chunk as poetry or prose and apply
       slightly different rules in each case.  Note that in both cases,
       we enforce the rule that the first line of a paragraph may be
       indented specially, but all subsequent lines must be indented the
       same amount. */

    chunkstart = -1;
    for (i = 0; i < nlines; i++) {
      if (lines[i].eoltype == PARAGRAPH_BREAK) {
	if (chunkstart != i - 1) {
	  nupper = 0;
	  nlower = 0;
	  nother = 0;
	  maxindent = 0;

	  for (j = chunkstart + 1; j <= i; j++) {
	    if (lines[j].first_nonpunct == G_UNICODE_UPPERCASE_LETTER
		|| lines[j].first_nonpunct == G_UNICODE_TITLECASE_LETTER)
	      nupper++;
	    else if (lines[j].first_nonpunct == G_UNICODE_LOWERCASE_LETTER)
	      nlower++;
	    else
	      nother++;
	    if (lines[j].indent > maxindent)
	      maxindent = lines[j].indent;
	  }

	  if (nupper > (nlower + nother) * 3 + 1) {
	    /* Assume poetry.  Only join lines when there's a hyphen
	       or an initial lowercase letter. */
	    for (j = chunkstart + 1; j < i; j++) {
	      if (lines[j].eoltype == UNKNOWN) {
		if (lines[j+1].first_nonpunct != G_UNICODE_LOWERCASE_LETTER
		    && !lines[j].endhyph)
		  lines[j].eoltype = PARAGRAPH_BREAK;
		else if (lines[j].width + 2 + lines[j+1].first_word
			 < lines[j+1].width)
		  lines[j].eoltype = PARAGRAPH_BREAK;
		else if (j > 0
			 && lines[j+1].indent != lines[j].indent
			 && lines[j-1].eoltype != PARAGRAPH_BREAK)
		  lines[j].eoltype = PARAGRAPH_BREAK;
		else
		  lines[j].eoltype = LINE_BREAK;
	      }
	    }
	  }
	  else {
	    /* Assume prose.  Guess that a given indentation level is
	       considered the start of a paragraph if lines at that
	       indentation level are statistically more likely to begin
	       with uppercase/titlecase letters. */

	    indlevels = g_new(struct indstat, maxindent + 1);
	    fupper = (gdouble) nupper / (nupper + nlower + nother);

	    for (indent = 0; indent <= maxindent; indent++) {
	      indlevels[indent].nupper = 0;
	      indlevels[indent].count = 0;

	      for (j = chunkstart + 1; j <= i; j++) {
		if (lines[j].indent == indent) {
		  indlevels[indent].count++;
		  if (lines[j].first_nonpunct == G_UNICODE_UPPERCASE_LETTER
		      || lines[j].first_nonpunct == G_UNICODE_TITLECASE_LETTER)
		    indlevels[indent].nupper++;
		}
	      }

	      if (indlevels[indent].count)
		indlevels[indent].fupper = ((gdouble) indlevels[indent].nupper
					    / indlevels[indent].count);
	    }

	    for (j = chunkstart + 1; j < i; j++) {
	      if (lines[j].eoltype == UNKNOWN) {
		if (indlevels[lines[j+1].indent].fupper > (fupper + 0.001))
		  lines[j].eoltype = PARAGRAPH_BREAK;
		else if (lines[j].width + 2 + lines[j+1].first_word
			 < lines[j+1].width)
		  lines[j].eoltype = PARAGRAPH_BREAK;
		else if (j > 0
			 && lines[j+1].indent != lines[j].indent
			 && lines[j-1].eoltype != PARAGRAPH_BREAK)
		  lines[j].eoltype = PARAGRAPH_BREAK;
		else
		  lines[j].eoltype = LINE_BREAK;
	      }
	    }

	    g_free(indlevels);
	  }
	}
	chunkstart = i;
      }
    } /* for (i = 0; i < nlines; i++) */
  } /* if (unwrap) */


  /* Reformat text */

  /* FIXME: add two spaces after a period if suggested by document
     style. */

  for (i = 0; i < nlines; i++) {
    if (*(lines[i].end)) {
      if (lines[i].eoltype == LINE_BREAK) {
	p = lines[i].end;
	while (*p && g_unichar_isspace(g_utf8_get_char(p)))
	  p = g_utf8_next_char(p);
	for (q = lines[i].end + 1; q < p; q++)
	  *q = IGNORE;
	if (lines[i].endhyph)
	  *(lines[i].end) = IGNORE;
	else
	  *(lines[i].end) = ' ';
      }
      else {
	p = g_utf8_next_char(lines[i].end);
	for (q = lines[i].end + 1; q < p; q++)
	  *q = IGNORE;
	*(lines[i].end) = '\n';
      }
    }
  }

  p = q = buf;
  while (*p) {
    if (*p != IGNORE) {
      *q = *p;
      q++;
    }
    p++;
  }
  *q = 0;
}


/**************** Import dialog ****************/

gboolean import_text_file(fbeditwindow* ew, const gchar* fname, gboolean replace)
{
  /* Try up to 3 different names for each character set.  (This is
     insane!) */
  static const struct {
    const char* group;
    const char* label;
    const char* cs1;
    const char* cs2;
    const char* cs3;
  } suggest_charsets[] =
    /* From the Mozilla list.  (Some of the more obscure character
       sets in this list are supported under different names by glibc,
       libiconv, and win_iconv.  Mozilla, of course, has its own
       system for handling character set conversion.) */
      {{ "West European", 0, 0, 0, 0 },
       { 0, "Western (ISO-8859-1)",        "ISO-8859-1",      0, 0 },
       { 0, "Western (ISO-8859-15)",       "ISO-8859-15",     0, 0 },
       { 0, "Western (IBM-850)",           "IBM850",          0, 0 },
       { 0, "Western (MacRoman)",          "Macintosh",       "MacRoman", 0 },
       { 0, "Western (Windows-1252)",      "Windows-1252",    0, 0 },
       { 0, "Celtic (ISO-8859-14)",        "ISO-8859-14",     0, 0 },
       { 0, "Greek (ISO-8859-7)",          "ISO-8859-7",      0, 0 },
       { 0, "Greek (MacGreek)",            "x-mac-greek",     "MacGreek", 0 },
       { 0, "Greek (Windows-1253)",        "Windows-1253",    0, 0 },
       { 0, "Icelandic (MacIcelandic)",    "x-mac-icelandic", "MacIceland", "Mac-IS" },
       { 0, "Nordic (ISO-8859-10)",        "ISO-8859-10",     0, 0 },
       { 0, "South European (ISO-8859-3)", "ISO-8859-3",      0, 0 },

       { "East European", 0, 0, 0, 0 },
       { 0, "Baltic (ISO-8859-4)",             "ISO-8859-4",     0, 0 },
       { 0, "Baltic (ISO-8859-13)",            "ISO-8859-13",    0, 0 },
       { 0, "Baltic (Windows-1257)",           "Windows-1257",   0, 0 },
       { 0, "Central European (IBM-852)",      "IBM852",         0, 0 },
       { 0, "Central European (ISO-8859-2)",   "ISO-8859-2",     0, 0 },
       { 0, "Central European (MacCE)",        "x-mac-ce",       "Mac-CentralEurope", 0 },
       { 0, "Central European (Windows-1250)",   "Windows-1250",   0, 0 },
       { 0, "Croatian (MacCroatian)",            "x-mac-croatian", "MacCroatian", 0 },
       { 0, "Cyrillic (IBM-855)",                "IBM855",         0, 0 },
       { 0, "Cyrillic (ISO-8859-5)",             "ISO-8859-5",     0, 0 },
       { 0, "Cyrillic (ISO-IR-111)",             "ISO-IR-111",     0, 0 },
       { 0, "Cyrillic (KOI8-R)",                 "KOI8-R",         0, 0 },
       { 0, "Cyrillic (MacCyrillic)",            "x-mac-cyrillic", "MacCyrillic", 0 },
       { 0, "Cyrillic (Windows-1251)",           "Windows-1251",   0, 0 },
       { 0, "Cyrillic/Russian (CP-866)",         "IBM866",         0, 0 },
       { 0, "Cyrillic/Ukrainian (KOI8-U)",       "KOI8-U",         0, 0 },
       { 0, "Cyrillic/Ukrainian (MacUkrainian)", "x-mac-ukrainian", "MacUkrainian", "MacUkraine" },
       { 0, "Romanian (ISO-8859-16)",            "ISO-8859-16",    0, 0 },
       { 0, "Romanian (MacRomanian)",            "x-mac-romanian", "MacRomania", "MacRomanian" },

       { "East Asian", 0, 0, 0, 0 },
       { 0, "Chinese Simplified (GB2312)",      "GB2312",      0, 0 },
       { 0, "Chinese Simplified (GBK)",         "GBK",         0, 0 },
       { 0, "Chinese Simplified (GB18030)",     "GB18030",     0, 0 },
       { 0, "Chinese Simplified (HZ)",          "HZ-GB-2312",  0, 0 },
       { 0, "Chinese Simplified (ISO-2022-CN)", "ISO-2022-CN", "x-cp50227", 0 },
       { 0, "Chinese Traditional (Big5)",       "BIG5",        0, 0 },
       { 0, "Chinese Traditional (Big5-HKSCS)", "BIG5-HKSCS",  0, 0 },
       { 0, "Chinese Traditional (EUC-TW)",     "EUC-TW",      0, 0 },
       { 0, "Japanese (EUC-JP)",                "EUC-JP",      0, 0 },
       { 0, "Japanese (ISO-2022-JP)",           "ISO-2022-JP", 0, 0 },
       { 0, "Japanese (Shift_JIS)",             "Shift_JIS",   0, 0 },
       { 0, "Korean (EUC-KR)",                  "EUC-KR",      0, 0 },
       { 0, "Korean (UHC)",                     "UHC",         0, 0 },
       { 0, "Korean (JOHAB)",                   "JOHAB",       0, 0 },
       { 0, "Korean (ISO-2022-KR)",             "ISO-2022-KR", 0, 0 },

       { "SE & SW Asian", 0, 0, 0, 0 },
       { 0, "Armenian (ARMSCII-8)",      "ARMSCII-8",        0, 0 },
       { 0, "Georgian (GEOSTD8)",        "GEOSTD8",          0, 0 },
       { 0, "Thai (TIS-620)",            "TIS-620",          0, 0 },
       { 0, "Thai (ISO-8859-11)",        "ISO-8859-11",      0, 0 },
       { 0, "Thai (Windows-874)",        "Windows-874",      0, 0 },
       { 0, "Turkish (IBM-857)",         "IBM857",           0, 0 },
       { 0, "Turkish (ISO-8859-9)",      "ISO-8859-9",       0, 0 },
       { 0, "Turkish (MacTurkish)",      "x-mac-turkish",    "MacTurkish", 0 },
       { 0, "Turkish (Windows-1254)",    "Windows-1254",     0, 0 },
       { 0, "Vietnamese (TCVN)",         "TCVN-5712",        0, 0 },
       { 0, "Vietnamese (VISCII)",       "VISCII",           0, 0 },
       { 0, "Vietnamese (VPS)",          "VPS",              0, 0 },
       { 0, "Vietnamese (Windows-1258)", "Windows-1258",     0, 0 },
       { 0, "Hindi (MacDevangari)",      "x-mac-devanagari", "MacDevangari", 0 },
       { 0, "Gujarati (MacGujarati)",    "x-mac-gujarati",   "MacGujarati", 0 },
       { 0, "Gurmukhi (MacGurmukhi)",    "x-mac-gurmukhi",   "MacGurmukhi", 0 },

       { "Middle Eastern", 0, 0, 0, 0 },
       { 0, "Arabic (ISO-8859-6)",        "ISO-8859-6",   0, 0 },
       { 0, "Arabic (Windows-1256)",      "Windows-1256", 0, 0 },
       { 0, "Arabic (IBM-864)",           "IBM864",       0, 0 },
       { 0, "Arabic (MacArabic)",         "x-mac-arabic", "MacArabic", 0 },
       { 0, "Farsi (MacFarsi)",           "x-mac-farsi",  "MacFarsi", 0 },
       { 0, "Hebrew (ISO-8859-8-I)",      "ISO-8859-8-I", 0, 0 },
       { 0, "Hebrew (Windows-1255)",      "Windows-1255", 0, 0 },
       { 0, "Hebrew Visual (ISO-8859-8)", "ISO-8859-8",   0, 0 },
       { 0, "Hebrew (IBM-862)",           "IBM862",       0, 0 },
       { 0, "Hebrew (MacHebrew)",         "x-mac-hebrew", "MacHebrew", 0 },

       { "Unicode", 0, 0, 0, 0 },
       { 0, "Unicode (UTF-8)",                "UTF-8",    0, 0 },
       { 0, "Unicode (UTF-16 Little Endian)", "UTF-16LE", 0, 0 },
       { 0, "Unicode (UTF-16 Big Endian)",    "UTF-16BE", 0, 0 },
       { 0, "Unicode (UTF-32 Little Endian)", "UTF-32LE", 0, 0 },
       { 0, "Unicode (UTF-32 Big Endian)",    "UTF-32BE", 0, 0 },
       { 0, "Unicode (UTF-7)",                "UTF-7",    0, 0 }};

  gchar* contents;
  gsize length;
  gchar* encoding;
  gchar* converted;
  GtkWidget *dlg, *button, *vbox, *hbox, *lbl, *cmb_encoding, *ent,
    *cb_unwrap;
  GtkTreeStore *store;
  GtkTreeIter group, option;
  GtkCellRenderer *rend;
  gchar *p, *q;
  gsize i;
  GtkTextIter start, end;
  gchar* norm;
  GError *err = NULL;

  if (!g_file_get_contents(fname, &contents, &length, &err)) {
    p = filename_to_display(fname);
    messagebox11(ew->window, GTK_MESSAGE_ERROR,
		 "Could not open the file %s", p,
		 "An error occurred when trying to read the file:\n%s",
		 err->message);
    g_free(p);
    g_error_free(err);
    return FALSE;
  }

  encoding = detect_text_encoding(contents, length);

  dlg = gtk_dialog_new_with_buttons("Import Text File",
				    GTK_WINDOW(ew->window),
				    GTK_DIALOG_MODAL,
				    GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				    NULL);

  button = gtk_button_new_with_mnemonic("_Import");
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_button_set_image(GTK_BUTTON(button),
		       gtk_image_new_from_stock(GTK_STOCK_OPEN,
						GTK_ICON_SIZE_BUTTON));
  gtk_dialog_add_action_widget(GTK_DIALOG(dlg), button, GTK_RESPONSE_ACCEPT);
  gtk_widget_show(button);

  gtk_dialog_set_alternative_button_order(GTK_DIALOG(dlg),
					  GTK_RESPONSE_ACCEPT,
					  GTK_RESPONSE_CANCEL,
					  -1);

  gtk_dialog_set_default_response(GTK_DIALOG(dlg), GTK_RESPONSE_ACCEPT);

  vbox = gtk_vbox_new(FALSE, 6);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 6);
  {
    p = filename_to_display(fname);
    q = g_markup_printf_escaped("<big><b>Importing text"
				" from %s</b></big>", p);
    lbl = gtk_label_new(q);
    gtk_label_set_use_markup(GTK_LABEL(lbl), TRUE);
    g_free(p);
    g_free(q);
    gtk_box_pack_start(GTK_BOX(vbox), lbl, FALSE, FALSE, 3);
    gtk_widget_show(lbl);

    hbox = gtk_hbox_new(FALSE, 12);
    {
      lbl = gtk_label_new_with_mnemonic("Character _encoding:");
      gtk_box_pack_start(GTK_BOX(hbox), lbl, FALSE, FALSE, 0);
      gtk_widget_show(lbl);

      store = gtk_tree_store_new(2, G_TYPE_STRING, G_TYPE_STRING);

      gtk_tree_store_append(store, &option, NULL);
      p = g_strdup_printf("Automatic (%s)", encoding);
      gtk_tree_store_set(store, &option, 0, p, 1, encoding, -1);
      g_free(p);
      g_free(encoding);

      for (i = 0; i < G_N_ELEMENTS(suggest_charsets); i++) {
	if (suggest_charsets[i].group) {
	  gtk_tree_store_append(store, &group, NULL);
	  gtk_tree_store_set(store, &group,
			     0, suggest_charsets[i].group,
			     1, "", -1);
	}
	else if (is_known_charset(suggest_charsets[i].cs1)) {
	  gtk_tree_store_append(store, &option, &group);
	  gtk_tree_store_set(store, &option,
			     0, suggest_charsets[i].label,
			     1, suggest_charsets[i].cs1, -1);
	}
	else if (is_known_charset(suggest_charsets[i].cs2)) {
	  gtk_tree_store_append(store, &option, &group);
	  gtk_tree_store_set(store, &option,
			     0, suggest_charsets[i].label,
			     1, suggest_charsets[i].cs2, -1);
	}
	else if (is_known_charset(suggest_charsets[i].cs3)) {
	  gtk_tree_store_append(store, &option, &group);
	  gtk_tree_store_set(store, &option,
			     0, suggest_charsets[i].label,
			     1, suggest_charsets[i].cs3, -1);
	}
      }

      cmb_encoding = gtk_combo_box_entry_new_with_model
	(GTK_TREE_MODEL(store),	1);
      gtk_widget_set_name(cmb_encoding, "fb-import-encoding");
      gtk_label_set_mnemonic_widget(GTK_LABEL(lbl), cmb_encoding);
      gtk_combo_box_set_active(GTK_COMBO_BOX(cmb_encoding), 0);

      ent = gtk_bin_get_child(GTK_BIN(cmb_encoding));
      gtk_entry_set_activates_default(GTK_ENTRY(ent), TRUE);

      rend = gtk_cell_renderer_text_new();
      gtk_cell_layout_clear(GTK_CELL_LAYOUT(cmb_encoding));
      gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(cmb_encoding), rend, TRUE);
      gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(cmb_encoding),
				    rend, "text", 0);

      gtk_box_pack_start(GTK_BOX(hbox), cmb_encoding, TRUE, TRUE, 0);
      gtk_widget_show(cmb_encoding);
    }
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    cb_unwrap = gtk_check_button_new_with_mnemonic
      ("_Automatically unwrap paragraphs");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cb_unwrap), TRUE);
    gtk_box_pack_start(GTK_BOX(vbox), cb_unwrap, FALSE, FALSE, 0);
    gtk_widget_show(cb_unwrap);
  }
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox), vbox, FALSE, FALSE, 0);
  gtk_widget_show(vbox);

  while (gtk_dialog_run(GTK_DIALOG(dlg)) == GTK_RESPONSE_ACCEPT) {
    encoding = gtk_combo_box_get_active_text(GTK_COMBO_BOX(cmb_encoding));

    converted = g_convert(contents, length, "UTF-8", encoding,
			  NULL, NULL, &err);

    if (!converted) {
      messagebox11(dlg, GTK_MESSAGE_ERROR,
		   "Unable to convert from encoding '%s'", encoding,
		   "An error occurred during conversion:\n%s",
		   err->message);
      g_free(encoding);
      g_clear_error(&err);
    }
    else {
      g_free(contents);
      g_free(encoding);

      if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cb_unwrap)))
	format_imported_text(converted, TRUE);
      else
	format_imported_text(converted, FALSE);

      if (replace) {
	fb_text_buffer_set_recording(FB_TEXT_BUFFER(ew->textbuf), FALSE);
	gtk_text_buffer_set_text(ew->textbuf, "", -1);
      }

      gtk_text_buffer_begin_user_action(ew->textbuf);

      gtk_text_buffer_get_selection_bounds(ew->textbuf, &start, &end);
      gtk_text_buffer_delete_interactive(ew->textbuf, &start, &end,
					 gtk_text_view_get_editable
					 (GTK_TEXT_VIEW(ew->textedit)));

      for (p = q = converted; *p; p++) {
	if (*p == '\n' || !p[1]) {
	  norm = g_utf8_normalize(q, p + 1 - q, G_NORMALIZE_DEFAULT_COMPOSE);
	  gtk_text_buffer_insert(ew->textbuf, &start, norm, -1);
	  g_free(norm);
	  q = p + 1;
	}
      }

      gtk_text_buffer_end_user_action(ew->textbuf);

      if (replace) {
	fb_text_buffer_set_recording(FB_TEXT_BUFFER(ew->textbuf), TRUE);

	g_free(ew->filename);
	ew->filename = NULL;

	g_free(ew->displayname);
	ew->displayname = g_strdup("Untitled");

	g_free(ew->booktitle);
	ew->booktitle = NULL;

	g_free(ew->appname);
	ew->appname = NULL;
	p = g_path_get_basename(fname);
	q = g_filename_to_utf8(p, -1, NULL, NULL, NULL);
	g_free(p);
	if (q) {
	  if ((p = strrchr(q, '.')))
	    *p = 0;
	  if (strlen(q) > 8)
	    q[8] = 0;
	  ew->appname = g_utf8_normalize(q, -1, G_NORMALIZE_DEFAULT_COMPOSE);
	  g_free(q);
	}

	g_free(ew->fontfile);
	ew->fontfile = g_strdup(FB_DEFAULT_SMALL_FONT);
      }

      g_free(converted);
      gtk_widget_destroy(dlg);
      return TRUE;
    }
  }

  g_free(contents);
  gtk_widget_destroy(dlg);
  return FALSE;
}
