/*
 * Flashbook GUI
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <string.h>
#include <gtk/gtk.h>
#include <fb.h>
#include "creator.h"
#include "fbgui.h"
#include "msgbox.h"

#if GLIB_CHECK_VERSION(2, 14, 0)
# define USE_REGEX
#else
# define GRegex int
#endif

static GtkListStore* search_history;
static GtkListStore* replace_history;
static gboolean last_search_case;
static gboolean last_search_regex;

typedef struct _fbsearchwindow {
  fbeditwindow* ew;
  GtkWidget* dlg;
  GtkWidget* cmb_search;
  GtkWidget* cmb_replace;
  GtkWidget* btn_search;
  GtkWidget* btn_replace;
  GtkWidget* cb_case;
  GtkWidget* cb_regex;
} fbsearchwindow;

enum {
  RESPONSE_SEARCH = 1,
  RESPONSE_REPLACE = 2,
  RESPONSE_REPLACE_ALL = 3
};

#define MAX_SEARCH_HISTORY_SIZE 10


/**************** String normalization ****************/

/* Convert string into a normalized, case-free form */
static char* normalize_string(const char* str, int nbytes)
{
  char *norm, *fold;
  norm = g_utf8_normalize(str, nbytes, G_NORMALIZE_ALL);
  fold = g_utf8_casefold(norm, -1); 
  g_free(norm);
  return fold;
}

/* Compare the normalized forms of two strings */
static int strcmp_norm(const char* str1, const char* str2)
{
  char *s1norm, *s2norm;
  int res;

  s1norm = normalize_string(str1, -1);
  s2norm = normalize_string(str2, -1);
  res = strcmp(s1norm, s2norm);
  g_free(s1norm);
  g_free(s2norm);
  return res;
}

/* Check if the normalized form of str begins with prefix; if so,
   determine how many characters of str are needed to produce prefix
   exactly. */
static gboolean startswith_norm(const char* str, const char* prefix,
				int* nsrcchars)
{
  char* norm;
  const char* ep;
  int nb, nb2, cc;

  norm = normalize_string(str, -1);

  for (nb = 0; prefix[nb]; nb++) {
    if (prefix[nb] != norm[nb]) {
      g_free(norm);
      return FALSE;
    }
  }

  cc = g_utf8_strlen(str, -1);

  if (!norm[nb]) {
    g_free(norm);
    *nsrcchars = cc;
    return TRUE;
  }
  else {
    g_free(norm);

    ep = str + strlen(str);
    do {
      norm = normalize_string(str, ep - str);
      nb2 = strlen(norm);
      g_free(norm);

      if (nb2 == nb) {
	*nsrcchars = cc;
	return TRUE;
      }
      else if (nb2 < nb) {
	break;
      }

      cc--;
      ep = g_utf8_find_prev_char(str, ep);
    } while (ep);

    return FALSE;
  }
}


/**************** Case matching ****************/

enum {
  CASE_NONE = 0,
  CASE_LOWER = 1,
  CASE_TITLE = 2,
  CASE_UPPER = 3
};

/* Get case of a string.  If skipbs is true, then don't consider the
   case of letters that are preceded by backslashes. */
static int getcase(const char* str, gboolean skipbs)
{
  gunichar c;
  GUnicodeType t;
  int wordstart = 1;
  int let = 0, lcw = 0, tcw = 0, ucw = 0, lcws = 0, tcws = 0, ucws = 0;

  while (*str) {
    c = g_utf8_get_char(str);
    str = g_utf8_next_char(str);
    t = g_unichar_type(c);

    if (c == '\\' && skipbs)
      continue;

    switch (t) {
    case G_UNICODE_LOWERCASE_LETTER:
      let = 1;
      if (wordstart)
	lcws = 1;
      else
	lcw = 1;
      wordstart = 0;
      break;

    case G_UNICODE_TITLECASE_LETTER:
      let = 1;
      if (wordstart)
	tcws = 1;
      else
	tcw = 1;
      wordstart = 0;
      break;

    case G_UNICODE_UPPERCASE_LETTER:
      let = 1;
      if (wordstart)
	ucws = 1;
      else
	ucw = 1;
      wordstart = 0;
      break;

    default:
      wordstart = 1;
    }
  }

  if (!let)
    return CASE_NONE;

  if (tcw)
    return CASE_NONE;

  if (lcws && !tcws && !ucws && !ucw)
    return CASE_LOWER;
  if (ucws && !tcws && !lcws && !lcw)
    return CASE_UPPER;

  if (!ucw && tcws)
    return CASE_TITLE;
  if (!ucw && ucws && lcw)
    return CASE_TITLE;

  return CASE_NONE;
}

/* Change string into given case. */
static char* changecase(const char* str, int ncase)
{
  char* result;
  int n, na;
  gunichar c;
  int wordstart = 1;

  switch (ncase) {
  case CASE_LOWER:
    return g_utf8_strdown(str, -1);

  case CASE_UPPER:
    return g_utf8_strup(str, -1);

  case CASE_TITLE:
    na = strlen(str) + 7;
    result = g_new(char, na);
    n = 0;
    do {
      if (n + 6 > na) {
	na = (n + 6) * 2;
	result = g_renew(char, result, na);
      }

      c = g_utf8_get_char(str);

      /* FIXME: this doesn't address non-one-to-one case
	 transformations */

      switch (g_unichar_type(c)) {
      case G_UNICODE_LOWERCASE_LETTER:
	if (wordstart)
	  c = g_unichar_totitle(c);
	wordstart = 0;
	break;

      case G_UNICODE_TITLECASE_LETTER:
	if (!wordstart)
	  c = g_unichar_tolower(c);
	wordstart = 0;
	break;

      case G_UNICODE_UPPERCASE_LETTER:
	if (wordstart)
	  c = g_unichar_totitle(c);
	else
	  c = g_unichar_tolower(c);
	wordstart = 0;
	break;

      default:
	wordstart = 1;
      }

      n += g_unichar_to_utf8(c, result + n);

      str = g_utf8_next_char(str);
    } while (c != 0);
    return result;

  default:
    return g_strdup(str);
  }
}


/**************** Searching ****************/

/* Search forward for a regular expression.  Currently this only
   allows single-line searching. */
static gboolean search_forward_regex(const GtkTextIter* searchstart,
				     const GtkTextIter* searchend,
				     const GRegex* regex,
				     GtkTextIter* matchstart,
				     GtkTextIter* matchend)
{
#ifdef USE_REGEX
  GtkTextIter head, tail;
  GMatchInfo* match = NULL;
  char* bufstr;
  int pos, startpos, endpos;

  head = *searchstart;
  pos = gtk_text_iter_get_line_index(&head);
  gtk_text_iter_set_line_offset(&head, 0);

  while (!searchend || gtk_text_iter_compare(&head, searchend) < 0) {
    tail = head;
    if (!gtk_text_iter_ends_line(&head))
      gtk_text_iter_forward_to_line_end(&head);
    bufstr = gtk_text_iter_get_slice(&tail, &head);

    if (g_regex_match_full(regex, bufstr, -1, pos, 0, &match, NULL)
	&& g_match_info_fetch_pos(match, 0, &startpos, &endpos)) {
      head = tail;
      gtk_text_iter_set_line_index(&tail, startpos);
      gtk_text_iter_set_line_index(&head, endpos);
      g_match_info_free(match);

      if (searchend && gtk_text_iter_compare(&tail, searchend) >= 0) {
	return FALSE;
      }
      else {
	*matchstart = tail;
	*matchend = head;
	return TRUE;
      }
    }
    g_match_info_free(match);

    if (!gtk_text_iter_forward_char(&head))
      break;

    pos = 0;
  }
#endif
  return FALSE;
}

/* Search forward for a string.  (If exact is true, search for that
   exact sequence of characters.  If false, search for any subsequence
   with the same normalized form.) */
static gboolean search_forward_string(const GtkTextIter* searchstart,
				      const GtkTextIter* searchend,
				      const char* searchstr,
				      gboolean exact,
				      GtkTextIter* matchstart,
				      GtkTextIter* matchend)
{
  GtkTextIter head, tail;
  char *searchstr_norm, *bufstr;
  int nch, snch;

  if (exact) {
    return gtk_text_iter_forward_search(searchstart, searchstr, 0,
					matchstart, matchend,
					searchend);
  }
  else {
    if (searchend && gtk_text_iter_compare(searchstart, searchend) > 0)
      return FALSE;

    searchstr_norm = normalize_string(searchstr, -1);

    /* Number of characters in searchstr_norm should be >= the number
       of characters we have to look ahead, since neither
       g_utf8_normalize() nor g_utf8_casefold() returns a smaller
       number of characters than it's given... I think. */

    nch = g_utf8_strlen(searchstr_norm, -1);

    tail = *searchstart;
    head = *searchstart;
    gtk_text_iter_forward_chars(&head, nch);

    while (!searchend || !gtk_text_iter_equal(&tail, searchend)) {
      bufstr = gtk_text_iter_get_slice(&tail, &head);

      if (startswith_norm(bufstr, searchstr_norm, &snch)) {
	head = tail;
	gtk_text_iter_forward_chars(&head, snch);

	*matchend = head;
	*matchstart = tail;
	g_free(bufstr);
	g_free(searchstr_norm);
	return TRUE;
      }

      g_free(bufstr);

      if (!gtk_text_iter_forward_char(&tail))
	break;
      gtk_text_iter_forward_char(&head);
    }

    g_free(searchstr_norm);
    return FALSE;
  }
}

/* Search forward for a string/regex. */
static gboolean search_forward(const GtkTextIter* searchstart,
			       const GtkTextIter* searchend,
			       const char* searchstr,
			       gboolean exact,
			       const GRegex* regex,
			       GtkTextIter* matchstart,
			       GtkTextIter* matchend)
{
  if (regex)
    return search_forward_regex(searchstart, searchend, regex,
				matchstart, matchend);
  else
    return search_forward_string(searchstart, searchend, searchstr,
				 exact, matchstart, matchend);
}


/**************** Replacement ****************/

/* Replace current selection with the given string */
static void replace_selected_text(const GtkTextIter* selstart,
				  const GtkTextIter* selend,
				  const char* repstr)
{
  GtkTextBuffer* buffer = gtk_text_iter_get_buffer(selstart);
  GtkTextIter start = *selstart, end = *selend;

  gtk_text_buffer_begin_user_action(buffer);
  gtk_text_buffer_delete_interactive(buffer, &start, &end, TRUE);
  gtk_text_buffer_insert_interactive(buffer, &start, repstr, -1, TRUE);
  gtk_text_buffer_end_user_action(buffer);
}

/* Replace current selection with the given string, changing case if
   appropriate.  If the original pattern and replacement have the same
   case (both lower, both upper, or both title) then we change the
   actual replacement string's case to match the case of the text
   already in the buffer. */
static void replace_selected_text_with_case(const GtkTextIter* selstart,
					    const GtkTextIter* selend,
					    const char* repstr,
					    const char* origpattern,
					    const char* origrepstr,
					    const char* bufstr,
					    gboolean skipbs)
{
  int case_pat, case_rep, case_buf;
  char* repstr_trans;

  case_pat = getcase(origpattern, skipbs);
  case_rep = getcase(origrepstr, skipbs);
  case_buf = getcase(bufstr, FALSE);

  if (case_pat && case_pat == case_rep && case_pat != case_buf) {
    repstr_trans = changecase(repstr, case_buf);
    replace_selected_text(selstart, selend, repstr_trans);
    g_free(repstr_trans);
  }
  else {
    replace_selected_text(selstart, selend, repstr);
  }
}

/* If current selection matches the given regex, replace it with the
   given replacement. */
static gboolean replace_selection_regex(const GtkTextIter* selstart,
					const GtkTextIter* selend,
					const GRegex* regex,
					const char* pattern,
					gboolean exact,
					const char* repstr)
{
#ifdef USE_REGEX
  GtkTextIter start, end;
  GMatchInfo* match = NULL;
  int startpos, endpos, matchstart, matchend;
  char *bufstr, *repstr_exp;

  if (gtk_text_iter_get_line(selstart) != gtk_text_iter_get_line(selend))
    return FALSE;

  startpos = gtk_text_iter_get_line_index(selstart);
  endpos = gtk_text_iter_get_line_index(selend);

  start = end = *selstart;
  gtk_text_iter_set_line_offset(&start, 0);
  if (!gtk_text_iter_ends_line(&end))
    gtk_text_iter_forward_to_line_end(&end);
  bufstr = gtk_text_iter_get_slice(&start, &end);

  if (!g_regex_match_full(regex, bufstr, -1, startpos,
			  G_REGEX_MATCH_ANCHORED, &match, NULL)
      || !g_match_info_fetch_pos(match, 0, &matchstart, &matchend)
      || startpos != matchstart || endpos != matchend) {
    g_match_info_free(match);
    g_free(bufstr);
    return FALSE;
  }

  repstr_exp = g_match_info_expand_references(match, repstr, NULL);
  if (!repstr_exp) {
    g_match_info_free(match);
    g_free(bufstr);
    return FALSE;
  }

  if (exact) {
    replace_selected_text(selstart, selend, repstr_exp);
  }
  else {
    char* bufstr_sub = g_match_info_fetch(match, 0);
    replace_selected_text_with_case(selstart, selend, repstr_exp,
				    pattern, repstr, bufstr_sub, TRUE);
    g_free(bufstr_sub);
  }

  g_match_info_free(match);
  g_free(bufstr);
  g_free(repstr_exp);
  return TRUE;
#else
  return FALSE;
#endif
}

/* If current selection matches the given string, replace it with the
   given replacement. */
static gboolean replace_selection_string(const GtkTextIter* selstart,
					 const GtkTextIter* selend,
					 const char* searchstr,
					 gboolean exact,
					 const char* repstr)
{
  char* bufstr = gtk_text_iter_get_slice(selstart, selend);

  if (!strcmp(bufstr, searchstr)
      || (!exact && !strcmp_norm(searchstr, bufstr))) {

    if (exact)
      replace_selected_text(selstart, selend, repstr);
    else
      replace_selected_text_with_case(selstart, selend, repstr,
				      searchstr, repstr, bufstr, FALSE);
    g_free(bufstr);
    return TRUE;
  }
  else {
    g_free(bufstr);
    return FALSE;
  }
}

/* If current selection matches, replace it. */
static gboolean replace_selection(const GtkTextIter* selstart,
				  const GtkTextIter* selend,
				  const char* searchstr, gboolean exact,
				  const GRegex* regex,
				  const char* repstr)
{
  if (regex)
    return replace_selection_regex(selstart, selend, regex,
				   searchstr, exact, repstr);
  else
    return replace_selection_string(selstart, selend,
				    searchstr, exact, repstr);
}


/**************** Search + Replace dialog ****************/

static gboolean do_search(fbeditwindow* ew, const char* str,
			  const GRegex* regex, gboolean exact,
			  gboolean wrap, gboolean showerr)
{
  GtkTextIter bstart, sstart, send, mstart, mend;
  GtkTextMark* mark;

  gtk_text_buffer_get_start_iter(ew->textbuf, &bstart);
  gtk_text_buffer_get_selection_bounds(ew->textbuf, &sstart, &send);

  if (search_forward(&send, NULL, str, exact, regex, &mstart, &mend)
      || (wrap && search_forward(&bstart, &sstart, str, exact, regex,
				 &mstart, &mend))) {
    gtk_text_buffer_select_range(ew->textbuf, &mstart, &mend);
    mark = gtk_text_buffer_get_insert(ew->textbuf);
    gtk_text_view_scroll_mark_onscreen(GTK_TEXT_VIEW(ew->textedit), mark);
    return TRUE;
  }
  else if (showerr) {
    messagebox01(ew->dlg_search, GTK_MESSAGE_INFO,
		 "Not found", "The string \"%s\" was not found.", str);
  }

  return FALSE;
}

static gboolean do_replace(fbeditwindow* ew, const char* searchstr,
			   GRegex* regex, gboolean exact,
			   const char* repstr)
{
  GtkTextIter start, end;

  if (!gtk_text_view_get_editable(GTK_TEXT_VIEW(ew->textedit)))
    return FALSE;

  gtk_text_buffer_get_selection_bounds(ew->textbuf, &start, &end);
  return replace_selection(&start, &end, searchstr, exact, regex, repstr);
}

static void search_dialog_activate_search(GtkEntry* ent G_GNUC_UNUSED,
				      gpointer data)
{
  fbsearchwindow* sw = data;
  gtk_widget_activate(sw->btn_search);
}

static void search_dialog_activate_replace(GtkEntry* ent G_GNUC_UNUSED,
					 gpointer data)
{
  fbsearchwindow* sw = data;
  gtk_widget_activate(sw->btn_replace);
}

static void history_add(GtkListStore* history, char* newent)
{
  GtkTreeIter iter;
  char* str;

  if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(history), &iter)) {
    do {
      gtk_tree_model_get(GTK_TREE_MODEL(history), &iter,
			 0, &str, -1);
      if (!strcmp(str, newent)) {
	gtk_list_store_remove(history, &iter);
	g_free(str);
	break;
      }
      g_free(str);
    } while (gtk_tree_model_iter_next(GTK_TREE_MODEL(history), &iter));
  }

  if (gtk_tree_model_iter_nth_child(GTK_TREE_MODEL(history), &iter,
				    NULL, MAX_SEARCH_HISTORY_SIZE - 1))
    gtk_list_store_remove(history, &iter);

  gtk_list_store_prepend(history, &iter);
  gtk_list_store_set(history, &iter, 0, newent, -1);
}

static void search_dialog_case_toggled(GtkToggleButton* tb,
				     gpointer data G_GNUC_UNUSED)
{
  last_search_case = gtk_toggle_button_get_active(tb);
}

static void search_dialog_regex_toggled(GtkToggleButton* tb,
				     gpointer data G_GNUC_UNUSED)
{
  last_search_regex = gtk_toggle_button_get_active(tb);
}

static void search_dialog_response(GtkDialog* dlg, gint response,
				 gpointer data)
{
  fbsearchwindow* sw = data;
  char *searchstr, *repstr;
  gboolean usecase, useregex;
  int count;
  GRegex* regex = NULL;

  searchstr = gtk_combo_box_get_active_text(GTK_COMBO_BOX(sw->cmb_search));
  if (searchstr[0])
    history_add(search_history, searchstr);
  else {
    g_free(searchstr);
    searchstr = NULL;
  }

  repstr = gtk_combo_box_get_active_text(GTK_COMBO_BOX(sw->cmb_replace));
  history_add(replace_history, repstr);

  if (response == GTK_RESPONSE_CANCEL) {
    gtk_widget_destroy(GTK_WIDGET(dlg));
    return;
  }
  else if (response != RESPONSE_SEARCH && response != RESPONSE_REPLACE
	   && response != RESPONSE_REPLACE_ALL) {
    return;
  }

  if (!searchstr)
    return;

  usecase = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(sw->cb_case));
  useregex = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(sw->cb_regex));

#ifdef USE_REGEX
  if (useregex) {
    GError* err = NULL;

    regex = g_regex_new(searchstr,
			((usecase ? 0 : G_REGEX_CASELESS)
			 | G_REGEX_OPTIMIZE),
			G_REGEX_MATCH_NOTEMPTY,
			&err);
    if (!regex) {
      messagebox01(sw->dlg, GTK_MESSAGE_ERROR,
		   "Invalid regular expression", "%s", err->message);
      g_error_free(err);
      return;
    }

    if (response != RESPONSE_SEARCH) {
      if (!g_regex_check_replacement(repstr, NULL, &err)) {
	messagebox01(sw->dlg, GTK_MESSAGE_ERROR,
		     "Invalid replacement string", "%s", err->message);
	g_error_free(err);
	g_regex_unref(regex);
	return;
      }
    }
  }
#endif

  switch (response) {
  case RESPONSE_SEARCH:
    do_search(sw->ew, searchstr, regex, usecase, TRUE, TRUE);
    break;

  case RESPONSE_REPLACE:
    do_replace(sw->ew, searchstr, regex, usecase, repstr);
    do_search(sw->ew, searchstr, regex, usecase, TRUE, TRUE);
    break;

  case RESPONSE_REPLACE_ALL:
    gtk_text_buffer_begin_user_action(sw->ew->textbuf);
    count = 0;
    do {
      if (do_replace(sw->ew, searchstr, regex, usecase, repstr))
	count++;
    } while (do_search(sw->ew, searchstr, regex, usecase, FALSE, FALSE));
    gtk_text_buffer_end_user_action(sw->ew->textbuf);

    messagebox01(sw->dlg, GTK_MESSAGE_INFO, "Finished searching",
		 (count == 1
		  ? "Replaced 1 occurrence."
		  : "Replaced %d occurrences."), count);
    break;
  }

#ifdef USE_REGEX
  if (regex)
    g_regex_unref(regex);
#endif
}

static void search_dialog_destroy(GtkWidget* w G_GNUC_UNUSED,
				gpointer data)
{
  fbsearchwindow* sw = data;
  sw->ew->dlg_search = NULL;
  g_free(sw);
}

void start_search_dialog(fbeditwindow* ew)
{
  fbsearchwindow* sw;
  GtkWidget *hbox, *tbl, *lbl, *btn_replace_all, *ent_search, *ent_replace;
  GtkTreeIter iter;
  char* str;

  if (!search_history)
    search_history = gtk_list_store_new(1, G_TYPE_STRING);
  if (!replace_history)
    replace_history = gtk_list_store_new(1, G_TYPE_STRING);

  sw = g_new0(fbsearchwindow, 1);
  sw->ew = ew;

  ew->dlg_search = sw->dlg = gtk_dialog_new_with_buttons
    ("Find and Replace", GTK_WINDOW(ew->window),
     GTK_DIALOG_DESTROY_WITH_PARENT,
     GTK_STOCK_CLOSE, GTK_RESPONSE_CANCEL, NULL);

  btn_replace_all = gtk_button_new_with_mnemonic("Replace _All");
  gtk_dialog_add_action_widget(GTK_DIALOG(sw->dlg), btn_replace_all,
			       RESPONSE_REPLACE_ALL);
  GTK_WIDGET_SET_FLAGS(btn_replace_all, GTK_CAN_DEFAULT);
  gtk_widget_show(btn_replace_all);

  sw->btn_replace = gtk_button_new_with_mnemonic("_Replace");
  gtk_button_set_image(GTK_BUTTON(sw->btn_replace),
		       gtk_image_new_from_stock(GTK_STOCK_FIND_AND_REPLACE,
						GTK_ICON_SIZE_BUTTON));
  gtk_dialog_add_action_widget(GTK_DIALOG(sw->dlg), sw->btn_replace,
			       RESPONSE_REPLACE);
  GTK_WIDGET_SET_FLAGS(sw->btn_replace, GTK_CAN_DEFAULT);
  gtk_widget_show(sw->btn_replace);

  sw->btn_search = gtk_button_new_from_stock(GTK_STOCK_FIND);
  gtk_dialog_add_action_widget(GTK_DIALOG(sw->dlg), sw->btn_search,
			       RESPONSE_SEARCH);
  GTK_WIDGET_SET_FLAGS(sw->btn_search, GTK_CAN_DEFAULT);
  gtk_widget_show(sw->btn_search);

  gtk_dialog_set_alternative_button_order(GTK_DIALOG(sw->dlg),
					  RESPONSE_SEARCH,
					  RESPONSE_REPLACE,
					  RESPONSE_REPLACE_ALL,
					  GTK_RESPONSE_CANCEL, -1);

  tbl = gtk_table_new(3, 2, FALSE);
  gtk_container_set_border_width(GTK_CONTAINER(tbl), 6);
  gtk_table_set_row_spacings(GTK_TABLE(tbl), 6);
  gtk_table_set_col_spacings(GTK_TABLE(tbl), 12);
  {
    lbl = gtk_label_new_with_mnemonic("_Search for:");
    gtk_misc_set_alignment(GTK_MISC(lbl), LABEL_X_ALIGN, 0.5);
    gtk_table_attach(GTK_TABLE(tbl), lbl, 0, 1, 0, 1,
		     GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show(lbl);

    sw->cmb_search = gtk_combo_box_entry_new_with_model
      (GTK_TREE_MODEL(search_history), 0);
    gtk_label_set_mnemonic_widget(GTK_LABEL(lbl), sw->cmb_search);
    gtk_table_attach(GTK_TABLE(tbl), sw->cmb_search, 1, 2, 0, 1,
		     GTK_EXPAND | GTK_FILL, 0, 0, 0);
    gtk_widget_show(sw->cmb_search);

    lbl = gtk_label_new_with_mnemonic("Replace _with:");
    gtk_misc_set_alignment(GTK_MISC(lbl), LABEL_X_ALIGN, 0.5);
    gtk_table_attach(GTK_TABLE(tbl), lbl, 0, 1, 1, 2,
		     GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show(lbl);

    sw->cmb_replace = gtk_combo_box_entry_new_with_model
      (GTK_TREE_MODEL(replace_history), 0);
    gtk_label_set_mnemonic_widget(GTK_LABEL(lbl), sw->cmb_replace);
    gtk_table_attach(GTK_TABLE(tbl), sw->cmb_replace, 1, 2, 1, 2,
		     GTK_EXPAND | GTK_FILL, 0, 0, 0);
    gtk_widget_show(sw->cmb_replace);

    hbox = gtk_hbox_new(FALSE, 12);
    {
      sw->cb_case = gtk_check_button_new_with_mnemonic("_Match case");
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sw->cb_case),
				   last_search_case);
      gtk_box_pack_start(GTK_BOX(hbox), sw->cb_case, FALSE, FALSE, 0);
      gtk_widget_show(sw->cb_case);

      sw->cb_regex = gtk_check_button_new_with_mnemonic
	("Regular e_xpression");
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sw->cb_regex),
				   last_search_regex);
      gtk_box_pack_start(GTK_BOX(hbox), sw->cb_regex, FALSE, FALSE, 0);
#ifdef USE_REGEX
      gtk_widget_show(sw->cb_regex);
#endif
    }
    gtk_table_attach(GTK_TABLE(tbl), hbox, 1, 2, 2, 3,
		     (GTK_EXPAND | GTK_FILL), 0, 12, 0);
    gtk_widget_show(hbox);
  }
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(sw->dlg)->vbox), tbl,
		     FALSE, FALSE, 0);
  gtk_widget_show(tbl);

  ent_search = gtk_bin_get_child(GTK_BIN(sw->cmb_search));
  ent_replace = gtk_bin_get_child(GTK_BIN(sw->cmb_replace));

  if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(search_history),
				    &iter)) {
    gtk_tree_model_get(GTK_TREE_MODEL(search_history), &iter,
		       0, &str, -1);
    gtk_entry_set_text(GTK_ENTRY(ent_search), str);
    g_free(str);
  }

  if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(replace_history),
				    &iter)) {
    gtk_tree_model_get(GTK_TREE_MODEL(replace_history), &iter,
		       0, &str, -1);
    gtk_entry_set_text(GTK_ENTRY(ent_replace), str);
    g_free(str);
  }

  g_signal_connect(ent_search, "activate",
		   G_CALLBACK(search_dialog_activate_search), sw);
  g_signal_connect(ent_replace, "activate",
		   G_CALLBACK(search_dialog_activate_replace), sw);

  g_signal_connect(sw->cb_case, "toggled",
		   G_CALLBACK(search_dialog_case_toggled), sw);
  g_signal_connect(sw->cb_regex, "toggled",
		   G_CALLBACK(search_dialog_regex_toggled), sw);

  g_signal_connect(sw->dlg, "response",
		   G_CALLBACK(search_dialog_response), sw);
  g_signal_connect(sw->dlg, "destroy",
		   G_CALLBACK(search_dialog_destroy), sw);

  gtk_widget_show(sw->dlg);
}


