/*
 * libflashbook - Library for creating eBooks for the TI-83/84 Plus
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <miniasm.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "fb.h"

struct _asfile {
  FILE* fp;
};

asfile* miniasm_open(const char* filename, void* data FBATTR_UNUSED)
{
  fbbook* fb = data;
  asfile* f = fbnew(asfile, 1);

  if (!(f->fp = fb_open_common_file(fb, filename, "rt"))) {
    fbfree(f);
    return NULL;
  }

  return f;
}

void miniasm_close(asfile* filep, void* data FBATTR_UNUSED)
{
  if (filep->fp != stdin)
    fclose(filep->fp);
  fbfree(filep);
}

char* miniasm_gets(asfile* filep, void* data FBATTR_UNUSED)
{
  int len = 0, len_a = 256;
  int c;
  char* str;

  if (feof(filep->fp) || ferror(filep->fp)) {
    return NULL;
  }

  str = fbnew(char, len_a);

  do {
    c = fgetc(filep->fp);

    if (c != EOF) {
      if (len >= len_a) {
	len_a *= 2;
	str = fbrenew(char, str, len_a);
      }
      str[len] = c;
      len++;
    }
    else
      break;
  } while (c != '\n' && c != '\r');

  if (c == '\r') {
    c = fgetc(filep->fp);
    if (c == '\n') {
      if (len >= len_a) {
	len_a *= 2;
	str = fbrenew(char, str, len_a);
      }
      str[len] = c;
      len++;
    }
    else
      ungetc(c, filep->fp);
  }

  if (len) {
    if (len >= len_a) {
      len_a = len + 1;
      str = fbrenew(char, str, len_a);
    }
    str[len] = 0;
    return str;
  }

  fbfree(str);
  return NULL;
}

void miniasm_gets_free(char* str)
{
  fbfree(str);
}

int miniasm_rewind(asfile* filep, void* data FBATTR_UNUSED)
{
  return fseek(filep->fp, 0L, SEEK_SET);
}

int miniasm_get_bin_size(const char* filename, long int* lengthp, void* data)
{
  struct stat st;
  fbbook* fb = data;

  if (!strcmp(filename, "fb:appname")) {
    *lengthp = fb->appname->length;
  }
  else if (!strcmp(filename, "fb:title")) {
    *lengthp = fb->title->length;
  }
  else if (!strcmp(filename, "fb:markname")) {
    *lengthp = 8;
  }
  else {
    if (stat(filename, &st))
      return 1;

    *lengthp = st.st_size;
  }
  return 0;
}

int miniasm_read_bin(const char* filename, unsigned char* buf, long int n,
		     void* data)
{
  FILE* f;
  fbbook* fb = data;
  int i;

  if (!strcmp(filename, "fb:appname")) {
    for (i = 0; i < n; i++)
      buf[i] = fb->appname->converted[i];
    return 0;
  }
  else if (!strcmp(filename, "fb:title")) {
    for (i = 0; i < n; i++)
      buf[i] = fb->title->converted[i];
    return 0;
  }
  else if (!strcmp(filename, "fb:markname")) {
    for (i = 0; i < 4; i++) {
      if (i < (int) fb->appname->length)
	buf[i] = fb->appname->converted[i];
      else
	buf[i] = '-';
    }
    for (i = 0; i < 4; i++) {
      buf[i + 4] = fb->markid[i];
    }
    return 0;
  }
  else {
    f = fb_open_common_file(fb, filename, "rb");
    if (!f)
      return 1;

    if (n == (long) fread(buf, 1, n, f)) {
      fclose(f);
      return 0;
    }
    else {
      fclose(f);
      return 1;
    }
  }
}

void miniasm_write_message(const char* string, void* data)
{
  fbbook* fb = data;
  fb_error(fb, 0, "miniasm: %s", string);
}

void miniasm_write_listing(int linenum, const char* text,
			   long int lineaddr, long int linecount,
			   const unsigned char* linebytes, void* data)
{
  fbbook* fb = data;
  int i;
  const char* p;

  if (!fb->lstfile)
    return;

  fprintf(fb->lstfile, "%5d %08X", linenum, (unsigned int) lineaddr);

  for (i = 0; i < linecount && i < 5; i++)
    fprintf(fb->lstfile, " %02X", (unsigned int) linebytes[i]);
  for (; i < 5; i++)
    fputs("   ", fb->lstfile);
  fputs("   ", fb->lstfile);

  for (p = text; *p; p++)
    if (*p != '\n' && *p != '\r')
      fputc(*p, fb->lstfile);

  for (; i < linecount; i++) {
    if (!(i % 5))
      fputs("\n              ", fb->lstfile);
    fprintf(fb->lstfile, " %02X", (unsigned int) linebytes[i]);
  }

  fputc('\n', fb->lstfile);
}
