/*
 * libflashbook - Library for creating eBooks for the TI-83/84 Plus
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#ifdef HAVE_ALLOCA
# ifndef __GNUC__
#  if HAVE_ALLOCA_H
#   include <alloca.h>
#  else
#   ifdef _AIX
#pragma alloca
#   else
#    ifndef alloca
char *alloca();
#    endif
#   endif
#  endif
# endif
# define TMP_ALLOC(n) alloca(n)
# define TMP_FREE(p) /* no-op */
#else
# define TMP_ALLOC(n) fbrealloc(0, (n))
# define TMP_FREE(p) fbfree((p))
#endif

#include <stdio.h>
#include <stdlib.h>
#include "fb.h"
#include "compress.h"

#ifdef HAVE_STDINT_H
# include <stdint.h>
#else
# ifdef HAVE_INTTYPES_H
#  include <inttypes.h>
# endif
#endif

struct _fbcompresscode {
  uint64_t v;         /* bit string */
  int l;              /* length */
};

#define MIN_ITERATIONS 2
#define MAX_ITERATIONS 100

/**************** LZ77 ****************/

static int lzlen(const fbchar* data, fbsize len, fbsize pos1, fbsize pos2)
{
  int count;

  if (pos1 == pos2)
    return 0;

  count = 0;
  while (pos1 < len && pos2 < len) {
    if (data[pos1] != data[pos2])
      return count;

    pos1++;
    pos2++;
    count++;
  }
  return count;
}

#define MODULUS 16381

static void lz77_pack(fbcompressctx* comp, fbchar* cdata, fbsize* clenp,
                      const fbchar* data, fbsize len)
{
  /* Chained hash table of buffer positions */
  fbsize shashtab[MODULUS];
  fbsize* shashtabnext;
  fbsize i;
  fbsize pos;
  int hash, ll, bestll;
  fbsize bestlpos = 0;
  uint64_t lzvalue;
  int j;
  int bitlen;
  int lzbitlen = comp->codes[LZ_MARKER].l + LZ_PARAM_BITS;

  for (i = 0; i < MODULUS; i++)
    shashtab[i] = (fbsize) -1;

  shashtabnext = TMP_ALLOC(len * sizeof(fbsize));

  *clenp = 0;
  for (i = 0; i < len; i++) {
    bestll = 0;
    if (i <= len - MIN_LZ_LENGTH) {
      hash = 0;
      for (j = 0; j < MIN_LZ_LENGTH; j++)
        hash = (hash * 61 + data[i + j]) % MODULUS;

      for (pos = shashtab[hash];
           pos != (fbsize) -1;
           pos = shashtabnext[pos]) {

        if (pos >= i || i - pos < MIN_LZ_OFFSET)
          continue;
        if (i - pos > MAX_LZ_OFFSET)
          break;

        ll = lzlen(data, len, i, pos);
        if (ll > MAX_LZ_LENGTH)
          ll = MAX_LZ_LENGTH;
        if (ll > bestll) {
          bestll = ll;
          bestlpos = pos;
        }
      }

      shashtabnext[i] = shashtab[hash];
      shashtab[hash] = i;
    }

    if (bestll >= MIN_LZ_LENGTH && comp->have_codes) {
      bitlen = 0;
      for (j = 0; j < bestll; j++) {
        bitlen += comp->codes[data[i + j]].l;
        if (bitlen > lzbitlen)
          break;
      }
      if (lzbitlen > bitlen)
        bestll = 0;
    }

    if (bestll >= MIN_LZ_LENGTH) {
      lzvalue = LZ_VALUE(i - bestlpos, bestll);
      cdata[*clenp] = LZ_MARKER;
      (*clenp)++;
      for (j = 0; j < LZ_PARAM_BYTES; j++) {
        cdata[*clenp] = lzvalue & 0xff;
        lzvalue >>= 8;
        (*clenp)++;
      }

      i += bestll - 1;
    }
    else {
      cdata[*clenp] = data[i];
      (*clenp)++;
    }
  }

  /* add EOF marker */
  cdata[*clenp] = LZ_MARKER;
  (*clenp)++;
  for (j = 0; j < LZ_PARAM_BYTES; j++) {
    cdata[*clenp] = 0;
    (*clenp)++;
  }

  TMP_FREE(shashtabnext);
}

/**************** Huffman ****************/

static int cmp_roots(const void* a, const void* b)
{
  const fbcompresstree* const* ta = a;
  const fbcompresstree* const* tb = b;
  return ((*tb)->count - (*ta)->count);
}

static void tree_insert(fbcompressctx* comp, fbcompresstree* t)
{
  int i, j;

  if (t->count >= comp->roots[0]->count)
    i = 0;
  else if (t->count < comp->roots[comp->nroots-1]->count)
    i = comp->nroots;
  else {
    i = comp->nroots-1;
    j = 0;
    while (j + 1 < i) {
      if (t->count < comp->roots[(i+j)/2]->count)
        j = (i+j)/2;
      else
        i = (i+j)/2;
    }
  }

  for (j = comp->nroots; j > i; j--)
    comp->roots[j] = comp->roots[j-1];
  comp->roots[j] = t;
  comp->nroots++;
}

static fbcompresstree* new_tree(fbchar value, fbcompresstree* left,
                                fbcompresstree* right, unsigned long count)
{
  fbcompresstree* t;

  t = fbnew(fbcompresstree, 1);
  t->value = value;
  t->left = left;
  t->right = right;
  t->count = count;
  return t;
}

static void free_tree(fbcompresstree* t)
{
  if (!t)
    return;
  if (t->left)
    free_tree(t->left);
  if (t->right)
    free_tree(t->right);
  fbfree(t);
}

static void huffman_init(fbcompressctx* comp)
{
  int c;
  for (c = 0; c < 256; c++)
    comp->roots[c] = new_tree(c, NULL, NULL, 0);
  comp->nroots = 256;
}

static void huffman_reset(fbcompressctx* comp)
{
  free_tree(comp->roots[0]);
}

static void huffman_count(fbcompressctx* comp, const fbchar* data, fbsize len)
{
  while (len > 0) {
    if (*data == LZ_MARKER) {
      comp->roots[LZ_MARKER]->count++;
      data += LZ_PARAM_BYTES + 1;
      len -= LZ_PARAM_BYTES + 1;
    }
    else {
      comp->roots[*data]->count++;
      data++;
      len--;
    }
  }
}

static void huffman_build_codes(fbcompressctx* comp, fbcompresstree* t,
                                uint64_t v, int l)
{
  if (t->left) {
    huffman_build_codes(comp, t->left, v << 1, l + 1);
    huffman_build_codes(comp, t->right, (v << 1) | 1, l + 1);
  }
  else {
    comp->codes[t->value].v = v;
    comp->codes[t->value].l = l;
  }
}

static void huffman_build_tree(fbcompressctx* comp)
{
  fbcompresstree *t;

  qsort(comp->roots, comp->nroots, sizeof(fbcompresstree*), &cmp_roots);
  while (comp->nroots > 2 && comp->roots[comp->nroots-1]->count == 0) {
    comp->nroots--;
    free_tree(comp->roots[comp->nroots]);
  }

  while (comp->nroots > 1) {
    t = new_tree(0, comp->roots[comp->nroots-1],
                 comp->roots[comp->nroots-2],
                 (comp->roots[comp->nroots-1]->count
                  + comp->roots[comp->nroots-2]->count));
    comp->nroots -= 2;
    tree_insert(comp, t);
  }

  huffman_build_codes(comp, comp->roots[0], (uint64_t) 0, 0);
  comp->have_codes = 1;
}

static fbsize get_tree_bitlen(fbcompressctx* comp, const fbcompresstree* t,
                              int prebits)
{
  if (t->left)
    return (get_tree_bitlen(comp, t->left, prebits + 1)
            + get_tree_bitlen(comp, t->right, prebits + 1));
  else if (t->value == LZ_MARKER)
    return (prebits + LZ_PARAM_BITS) * t->count;
  else
    return prebits * t->count;
}


static void huffman_output_bit(int bit, unsigned char* bitbuf, int* l,
                               fboutbuf* b)
{
  if (*l == 8) {
    fb_outbuf_append_byte(b, *bitbuf);
    *l = 0;
  }
  (*l)++;
  *bitbuf = (*bitbuf << 1) | bit;
}

static void huffman_flush_bits(unsigned char* bitbuf, int* l, fboutbuf* b)
{
  int i;
  for (i = 0; i < 7; i++)
    huffman_output_bit(0, bitbuf, l, b);
}

static void huffman_output_tree(fbcompresstree* t, unsigned char* bitbuf,
                                int* l, fboutbuf* b)
{
  int i;
  if (t->left) {
    huffman_output_bit(1, bitbuf, l, b);
    huffman_output_tree(t->left, bitbuf, l, b);
    huffman_output_tree(t->right, bitbuf, l, b);
  }
  else {
    huffman_output_bit(0, bitbuf, l, b);
    for (i = 0; i < 8; i++)
      huffman_output_bit(1 & (t->value >> (7 - i)), bitbuf, l, b);
  }
}

static void huffman_pack(fbcompressctx* comp, const fbchar* data,
                         fbsize len, fboutbuf* b)
{
  int i;
  int nbits = 0;
  unsigned char bitbuf = 0;
  int l;
  uint64_t v;

  while (len > 0) {

    l = comp->codes[*data].l;
    v = comp->codes[*data].v;

    if (*data == LZ_MARKER) {
      l += LZ_PARAM_BITS;
      v <<= LZ_PARAM_BITS;
      for (i = 0; i < LZ_PARAM_BITS; i += 8) {
        data++;
        len--;
        v |= (*data << i);
      }

      if (!(v & LZ_OFFSET_MASK)) {
        /* omit length bits from EOF marker */
        v >>= LZ_LENGTH_BITS;
        l -= LZ_LENGTH_BITS;
      }
    }

    data++;
    len--;

    if (l <= 8 - nbits) {
      bitbuf <<= l;
      bitbuf |= v & 0xff;
      nbits += l;
    }
    else {
      bitbuf <<= 8 - nbits;
      l -= 8 - nbits;
      bitbuf |= (v >> l) & 0xff;
      fb_outbuf_append_byte(b, bitbuf);
      while (l > 8) {
        l -= 8;
        bitbuf = (v >> l) & 0xff;
        fb_outbuf_append_byte(b, bitbuf);
      }
      nbits = l;
      bitbuf = v & 0xff;
    }

    if (nbits == 8) {
      fb_outbuf_append_byte(b, bitbuf);
      nbits = 0;
    }
  }

  if (nbits) {
    bitbuf <<= 8 - nbits;
    fb_outbuf_append_byte(b, bitbuf);
  }
}



/**************** Exported ****************/

fbcompressctx* fb_compressctx_new(fbsize blocksize)
{
  fbcompressctx* comp = fbnew(fbcompressctx, 1);
  int i;

  i = 0;
  while (((fbsize) 1 << i) < blocksize)
    i++;

  comp->lz_offset_bits = i - 1;
  comp->lz_length_bits = 3;
  comp->lz_length_bias = 5;

  comp->roots[0] = 0;
  comp->nroots = 0;
  comp->codes = fbnew(fbcompresscode, 256);
  comp->have_codes = 0;
  comp->last_bitlen = 0;
  comp->iterations = 0;

  /* cdata will never need to be longer than the input data (plus a
     little extra for the EOF marker) so long as
     (1 + LZ_PARAM_BYTES) <= LZ_LENGTH_BIAS */
  comp->cdata = fbnew(fbchar, blocksize + 1 + LZ_PARAM_BYTES);

  return comp;
}

void fb_compressctx_free(fbcompressctx* comp)
{
  huffman_reset(comp);
  fbfree(comp->codes);
  fbfree(comp->cdata);
  fbfree(comp);
}

void fb_compress_begin(fbcompressctx* comp)
{
  huffman_reset(comp);
  huffman_init(comp);
  comp->last_bitlen = comp->current_bitlen;
}

void fb_compress_try_data(fbcompressctx* comp, const fbchar* data, fbsize size)
{
  fbsize clen;

  lz77_pack(comp, comp->cdata, &clen, data, size);
  huffman_count(comp, comp->cdata, clen);
}

void fb_compress_finish(fbcompressctx* comp)
{
  huffman_build_tree(comp);
  comp->current_bitlen = get_tree_bitlen(comp, comp->roots[0], 0);
  comp->iterations++;
}

int fb_compress_again(fbcompressctx* comp)
{
  if (comp->iterations < MIN_ITERATIONS)
    return 1;
  if (comp->iterations > MAX_ITERATIONS)
    return 0;
  if (comp->current_bitlen < comp->last_bitlen)
    return 1;
  return 0;
}

void fb_compress_write_global_data(fboutbuf* buffer, fbcompressctx* comp)
{
  unsigned char bitbuf = 0;
  int l = 0;

  fb_outbuf_append_byte(buffer, LZ_OFFSET_BITS);
  fb_outbuf_append_byte(buffer, LZ_LENGTH_BITS);
  fb_outbuf_append_byte(buffer, LZ_LENGTH_BIAS);

  huffman_output_tree(comp->roots[0]->left, &bitbuf, &l, buffer);
  huffman_output_tree(comp->roots[0]->right, &bitbuf, &l, buffer);
  huffman_flush_bits(&bitbuf, &l, buffer);
}

void fb_compress_write_data(fboutbuf* buffer, fbcompressctx* comp,
                            const fbchar* data, fbsize size)
{
  fbsize clen;
  lz77_pack(comp, comp->cdata, &clen, data, size);
  huffman_pack(comp, comp->cdata, clen, buffer);
}
