/*
 * libflashbook - Library for creating eBooks for the TI-83/84 Plus
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include "fb.h"

/**************** Memory management ****************/

/* General memory allocation function.

   This is used to implement the fbnew, fbrenew, and fbfree macros
   used throughout the Flashbook library.  It works similarly to
   GLib's g_realloc() in that it Does The Right Thing when p and/or sz
   are zero, and aborts if we run out of memory.  Note that unlike GNU
   realloc, this does not allocate anything when sz = 0.
*/
void* fbrealloc(void* p, unsigned long sz)
{
  if (sz) {
    if (p)
      p = realloc(p, sz);
    else
      p = malloc(sz);
    if (!p) {
      fprintf(stderr, "Out of memory (need %lu bytes)\n", sz);
      abort();
    }
  }
  else {
    if (p)
      free(p);
    p = 0;
  }
  return p;
}

/* Create a new, empty fbchunk structure. */
fbchunk* fb_chunk_new()
{
  fbchunk* c = fbnew(fbchunk, 1);
  c->utf8 = 0;
  c->utf8size = 0;
  c->utf8size_a = 0;
  c->nutf8chars = 0;
  c->converted = 0;
  c->length = 0;
  c->length_a = 0;
  c->linestart = 0;
  c->lineend = 0;
  c->nlines = 0;
  fb_chartable_clear(&c->charcounts);
  return c;
}

/* Free an fbchunk structure. */
void fb_chunk_free(fbchunk* chunk)
{
  if (!chunk)
    return;

  fbfree(chunk->utf8);
  fbfree(chunk->converted);
  fbfree(chunk->linestart);
  fbfree(chunk->lineend);
  fbfree(chunk);
}

/* Create a new, empty fbbook structure. */
fbbook* fb_book_new()
{
  fbbook* fb = fbnew(fbbook, 1);
  int i;
  fb->title = fb_chunk_new();
  fb->appname = fb_chunk_new();
  fb->body = fb_chunk_new();
  fb->sections = 0;
  fb->nsections = 0;
  fb->nsections_a = 0;
  fb->blockstart = 0;
  fb->blockend = 0;
  fb->nblocks = 0;
  fb->blocksize_limit = FB_DEFAULT_BLOCK_SIZE;
  fb->blocksize_actual = 0;
  fb->packedbitlen = 0;
  fb->fontfile = 0;
  fb_chartable_clear(&fb->encoding_table);
  for (i = 0; i < 256; i++) {
    fb->encoding[i] = 0;
    fb->charcount[i] = 0;
    fb->fontchars[i].width = 0;
    fb->fontchars[i].value = 0;
    fb->fontchars[i].set = 0;
    fb->fontchars[i].custom = 0;
  }
  fb->rowheight = 6;
  fb->hyphenate = 0;
  fb->hyphlang = 0;
  fb->searchdir = 0;
  fb->asmfile = 0;
  fb->lstfile = 0;
  fb->errfunc = &fb_error_default;
  fb->userdata = 0;
  return fb;
}

/* Free an fbbook structure. */
void fb_book_free(fbbook* fb)
{
  fbsize i;
  fb_chunk_free(fb->title);
  fb_chunk_free(fb->appname);
  fb_chunk_free(fb->body);
  for (i = 0; i < fb->nsections; i++)
    fb_chunk_free(fb->sections[i].title);
  fbfree(fb->fontfile);
  fbfree(fb->sections);
  fbfree(fb->blockstart);
  fbfree(fb->blockend);
  fbfree(fb->hyphlang);
  fbfree(fb->searchdir);
  fbfree(fb->asmfile);
  fbfree(fb);
}

/* Clear an fbchunk. */
void fb_chunk_clear(fbchunk* chunk)
{
  fbfree(chunk->utf8);
  fbfree(chunk->converted);
  fbfree(chunk->linestart);
  fbfree(chunk->lineend);
  chunk->utf8 = 0;
  chunk->utf8size = 0;
  chunk->utf8size_a = 0;
  chunk->nutf8chars = 0;
  chunk->converted = 0;
  chunk->length = 0;
  chunk->length_a = 0;
  chunk->linestart = 0;
  chunk->lineend = 0;
  chunk->nlines = 0;
  fb_chartable_clear(&chunk->charcounts);
}

/* Add a new section to the end of the book. */
fbsection* fb_book_append_section(fbbook* fb)
{
  fbsection* sect;

  fb->nsections++;
  if (fb->nsections >= fb->nsections_a) {
    fb->nsections_a = fb->nsections * 2;
    fb->sections = fbrenew(fbsection, fb->sections,
                           fb->nsections_a);
  }

  sect = &fb->sections[fb->nsections - 1];
  sect->title = fb_chunk_new();
  sect->pos_utf8 = fb->body->nutf8chars;
  sect->pos_conv = fb->body->length;

  return sect;
}

/* Create an output buffer */
fboutbuf* fb_outbuf_new()
{
  fboutbuf* buf = fbnew(fboutbuf, 1);
  buf->data = 0;
  buf->length = 0;
  buf->length_a = 0;
  return buf;
}

/* Free an output buffer */
void fb_outbuf_free(fboutbuf* buf)
{
  fbfree(buf->data);
  fbfree(buf);
}

/* Add a byte to the end of an output buffer */
void fb_outbuf_append_byte(fboutbuf* buf, unsigned char d)
{
  if (buf->length >= buf->length_a) {
    buf->length_a += 1024;
    buf->data = fbrenew(unsigned char, buf->data, buf->length_a);
  }
  buf->data[buf->length++] = d;
}

/* Preallocate space for writing data */
void fb_outbuf_prealloc(fboutbuf* buf, size_t size)
{
  if (buf->length + size >= buf->length_a) {
    buf->length_a = buf->length + size + 1024;
    buf->data = fbrenew(unsigned char, buf->data, buf->length_a);
  }
}
