/*
 * libflashbook - Library for creating eBooks for the TI-83/84 Plus
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include "fb.h"

static inline int put_xml_char(FILE* f, fbunichar c)
{
  if (c == '<')
    return (fputs("&lt;", f) < 0);
  else if (c == '>')
    return (fputs("&gt;", f) < 0);
  else if (c == '&')
    return (fputs("&amp;", f) < 0);
  else if (c == '"')
    return (fputs("&quot;", f) < 0);
  else if (c == '\'')
    return (fputs("&apos;", f) < 0);
  else if (c < 0x80)
    return (fputc(c, f) == EOF);
  else if (c < 0x800) {
    return (fputc(0xc0 | (c >> 6), f) == EOF
            || fputc(0x80 | (c & 0x3f), f) == EOF);
  }
  else if (c < 0x10000) {
    return (fputc(0xe0 | (c >> 12), f) == EOF
            || fputc(0x80 | ((c >> 6) & 0x3f), f) == EOF
            || fputc(0x80 | (c & 0x3f), f) == EOF);
  }
  else if (c < 0x200000) {
    return (fputc(0xf0 | (c >> 18), f) == EOF
            || fputc(0x80 | ((c >> 12) & 0x3f), f) == EOF
            || fputc(0x80 | ((c >> 6) & 0x3f), f) == EOF
            || fputc(0x80 | (c & 0x3f), f) == EOF);
  }
  else {
    return (fprintf(f, "&#x%x;", c) < 0);
  }
}

static int write_chunk(const fbbook* fb, const fbchunk* chunk,
                       FILE* f, int system)
{
  fbsize i;

  if (chunk->utf8) {
    for (i = 0; i < chunk->utf8size; i++) {
      if (i == chunk->utf8size - 1
          && (chunk->utf8[i] == '\n' || chunk->utf8[i] == '\r'))
        break;
      else if (i == chunk->utf8size - 2
               && chunk->utf8[i] == '\r'
               && chunk->utf8[i + 1] == '\n')
        break;

      if (chunk->utf8[i] < 0x80) {
        if (put_xml_char(f, chunk->utf8[i]))
          return 1;
      }
      else {
        if (fputc(chunk->utf8[i], f) == EOF)
          return 1;
      }
    }
  }
  else if (system) {
    for (i = 0; i < chunk->length; i++) {
      if (i == chunk->length - 1
          && chunk->converted[i] == fb->newlinechar)
        break;

      if (put_xml_char(f, fb_book_get_system_char(fb, chunk->converted[i])))
        return 1;
    }
  }
  else {
    for (i = 0; i < chunk->length; i++) {
      if (i == chunk->length - 1
          && chunk->converted[i] == fb->newlinechar)
        break;

      if (put_xml_char(f, fb->encoding[chunk->converted[i]]))
        return 1;
    }
  }

  return 0;
}

int fb_book_write_xml(fbbook* fb, FILE* f)
{
  fbsize i, j, n;
  char *e;
  unsigned char *p;
  fbunichar c;
  int from_utf8;

  if (0 > fputs("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                "<flashbook>\n<title>", f)
      || write_chunk(fb, fb->title, f, 0)
      || 0 > fputs("</title>\n<appname>", f)
      || write_chunk(fb, fb->appname, f, 1)
      || 0 > fputs("</appname>\n", f))
    return ERR_FILE_IO;

  if (fb->blocksize_limit != FB_DEFAULT_BLOCK_SIZE)
    fprintf(f, "<blocksize>%lu</blocksize>\n", fb->blocksize_limit);
  if (fb->fontfile && fb->fontfile[0]) {
    e = fb_escape_filename(fb->fontfile, -1);
    fprintf(f, "<font>%s</font>\n", e);
    fbfree(e);
  }
  if (fb->hyphenate || fb->hyphlang) {
    fprintf(f, "<hyphenation enabled=\"%c\"", fb->hyphenate ? '1' : '0');
    if (fb->hyphlang)
      fprintf(f, " lang=\"%s\"", fb->hyphlang);
    fprintf(f, "/>\n");
  }

  if (0 > fputs("<body>", f))
    return ERR_FILE_IO;

  p = fb->body->utf8;
  if (p) {
    from_utf8 = 1;
    n = fb->body->nutf8chars;
  }
  else {
    from_utf8 = 0;
    n = fb->body->length;
  }

  for (i = j = 0; i < n; i++) {
    while (j < fb->nsections
	   && i >= (from_utf8
		    ? fb->sections[j].pos_utf8
		    : fb->sections[j].pos_conv)) {
      if (0 > fputs("<section", f))
        return ERR_FILE_IO;

      if (fb->sections[j].title->length > 0) {
        if (0 > fputs(" title=\"", f)
            || write_chunk(fb, fb->sections[j].title, f, 0)
            || fputc('"', f) == EOF)
          return ERR_FILE_IO;
      }
      if (0 > fputs("/>", f))
        return ERR_FILE_IO;

      j++;
    }

    if (from_utf8)
      c = read_utf8(&p);
    else
      c = fb->encoding[fb->body->converted[i]];

    if (from_utf8 || c != 0xAD)
      if (put_xml_char(f, c))
	return ERR_FILE_IO;
  }

  if (0 > fputs("</body>\n</flashbook>\n", f))
    return ERR_FILE_IO;

  return 0;
}
