/*
 * libflashbook - Library for creating eBooks for the TI-83/84 Plus
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <expat.h>
#include "fb.h"

struct parseinfo {
  XML_Parser parser;
  const char* filename;
  fbbook* fb;
  int ntags;
  fbchunk* ochunk1;
  fbchunk* ochunk2;
  fbchunk* tmpchunk;
};

#define BUF_SIZE 1024

#define IS_SPACE(c) ((c) == ' ' || (c) == '\t'          \
                     || (c) == '\n' || (c) == '\v'      \
                     || (c) == '\f' || (c) == '\r')


static void handle_start_tag(void* data, const XML_Char* name,
                             const XML_Char** atts)
{
  struct parseinfo* pi = data;
  fbsection* sect;

  if (pi->ntags == 0) {
    if (strcmp(name, "flashbook")) {
      fb_error(pi->fb, ERR_UNKNOWN_XML_ROOT,
               "Unrecognized root element %s in %s", name,
               pi->filename);
      XML_StopParser(pi->parser, 0);
      return;
    }
  }
  else if (pi->ntags == 1) {
    if (!strcmp(name, "title"))
      pi->ochunk1 = pi->fb->title;
    else if (!strcmp(name, "appname"))
      pi->ochunk1 = pi->fb->appname;
    else if (!strcmp(name, "body"))
      pi->ochunk1 = pi->fb->body;
    else if (!strcmp(name, "font"))
      pi->tmpchunk = fb_chunk_new();
    else if (!strcmp(name, "hyphenation")) {
      while (atts && atts[0]) {
	if (!strcmp(atts[0], "enabled")) {
	  pi->fb->hyphenate = (atts[1][0] == '0' ? 0 : 1);
	}
	else if (!strcmp(atts[0], "lang")) {
	  fbfree(pi->fb->hyphlang);
	  pi->fb->hyphlang = fbnew(char, strlen(atts[1]) + 1);
	  strcpy(pi->fb->hyphlang, atts[1]);
	}
	atts += 2;
      }
    }
    else if (!strcmp(name, "blocksize"))
      pi->tmpchunk = fb_chunk_new();
  }
  else if (pi->ntags == 2 && pi->ochunk1 == pi->fb->body) {
    if (!strcmp(name, "section")) {
      sect = fb_book_append_section(pi->fb);
      pi->ochunk2 = sect->title;

      while (atts && atts[0]) {
        if (!strcmp(atts[0], "title"))
          fb_chunk_append_text_utf8(pi->fb, sect->title,
                                    atts[1], strlen(atts[1]));
        atts += 2;
      }
    }
  }

  pi->ntags++;
}

static void handle_end_tag(void* data, const XML_Char* name)
{
  struct parseinfo* pi = data;
  char* p;
  long int bs = 0;

  pi->ntags--;

  if (pi->tmpchunk && !strcmp(name, "blocksize")) {
    fb_chunk_append_text_utf8(pi->fb, pi->tmpchunk, "$", 1);
    pi->tmpchunk->utf8[pi->tmpchunk->utf8size - 1] = 0;
    bs = strtol((char*) pi->tmpchunk->utf8, &p, 10);

    if (p == (char*) pi->tmpchunk->utf8 || *p) {
      fb_error(pi->fb, ERR_INVALID_NUMERIC_VALUE,
               "%s:%ld.%ld: Invalid numeric value '%s'", pi->filename,
               (long) XML_GetCurrentLineNumber(pi->parser),
               (long) XML_GetCurrentColumnNumber(pi->parser),
               pi->tmpchunk->utf8);
      XML_StopParser(pi->parser, 0);
    }
    else {
      pi->fb->blocksize_limit = bs;
    }
    fb_chunk_free(pi->tmpchunk);
    pi->tmpchunk = NULL;
  }
  else if (pi->tmpchunk && !strcmp(name, "font")) {
    fbfree(pi->fb->fontfile);
    pi->fb->fontfile = fb_unescape_filename((char*) pi->tmpchunk->utf8,
                                            pi->tmpchunk->utf8size);
    fb_chunk_free(pi->tmpchunk);
    pi->tmpchunk = NULL;
  }

  if (pi->ntags == 2)
    pi->ochunk2 = NULL;
  else if (pi->ntags == 1)
    pi->ochunk1 = NULL;
}

static void handle_chars(void* data, const XML_Char* s, int len)
{
  struct parseinfo* pi = data;

  if (pi->ochunk1)
    fb_chunk_append_text_utf8(pi->fb, pi->ochunk1, s, len);
  if (pi->ochunk2)
    fb_chunk_append_text_utf8(pi->fb, pi->ochunk2, s, len);
  if (pi->tmpchunk)
    fb_chunk_append_text_utf8(NULL, pi->tmpchunk, s, len);
}

int fb_book_read_xml(fbbook* fb, FILE* f, const char* filename)
{
  struct parseinfo pi;
  char* buf;
  size_t n;
  const XML_LChar* msg;
  int err;

  pi.parser = XML_ParserCreate(NULL);
  pi.filename = filename;
  pi.fb = fb;
  pi.ntags = 0;
  pi.ochunk1 = pi.ochunk2 = pi.tmpchunk = NULL;

  XML_SetElementHandler(pi.parser, &handle_start_tag, &handle_end_tag);
  XML_SetCharacterDataHandler(pi.parser, &handle_chars);
  XML_SetUserData(pi.parser, &pi);

  while (1) {
    buf = XML_GetBuffer(pi.parser, BUF_SIZE);
    n = fread(buf, 1, BUF_SIZE, f);

    if (!XML_ParseBuffer(pi.parser, n, n ? 0 : 1)) {
      err = XML_GetErrorCode(pi.parser);
      if (err != XML_ERROR_ABORTED) {
        msg = XML_ErrorString(XML_GetErrorCode(pi.parser));
        fb_error(fb, ERR_XML_PARSER, "%s:%ld.%ld: %s", filename,
                 (long) XML_GetCurrentLineNumber(pi.parser),
                 (long) XML_GetCurrentColumnNumber(pi.parser),
                 msg);
        XML_ParserFree(pi.parser);
      }
      return ERR_XML_PARSER;
    }
    if (n == 0) {
      XML_ParserFree(pi.parser);
      return 0;
    }
  }
}
