/*
 * libflashbook - Library for creating eBooks for the TI-83/84 Plus
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include "fb.h"

/**************** Block Splitting ****************/

static void get_line_length(fbbook* fb, fbchunk* chunk,
			    fbsize line_start, fbsize* line_end,
			    fbsize* next_line_start,
			    const int* widthtab, int maxcc,
			    fbchar zwsp, fbchar hyphen, fbhyphctx* hyph)
{
  fbsize i;
  int xcoord, breakx, cc, breakcc, foundbreak;
  fbsize prebreakpos, postbreakpos, hyphpos;

  xcoord = cc = foundbreak = 0;
  prebreakpos = postbreakpos = line_start;
  breakx = breakcc = 0;

  for (i = line_start; i < chunk->length; i++) {
    if (chunk->converted[i] == fb->newlinechar)
      break;

    xcoord += widthtab[chunk->converted[i]];
    cc++;

    if (chunk->converted[i] == fb->spacechar) {
      prebreakpos = i;
      postbreakpos = i + 1;
      breakx = xcoord;
      breakcc = cc;
      foundbreak = 1;
    }
    else if (xcoord <= 96 && cc <= maxcc
	     && (chunk->converted[i] == hyphen
		 || chunk->converted[i] == zwsp)) {
      prebreakpos = i + 1;
      postbreakpos = i + 1;
      breakx = xcoord;
      breakcc = cc;
      foundbreak = 1;
    }

    if (xcoord > 96 || cc > maxcc) {
      if (hyph && (hyphpos = fb_hyphenate(hyph, chunk, postbreakpos,
					  maxcc - breakcc, 96 - breakx))) {
	fb_chunk_insert_text_ti(fb, chunk, hyphpos, &hyph->softhyphen, 1);
	*line_end = *next_line_start = hyphpos + 1;
	return;
      }

      if (!foundbreak)
	prebreakpos = postbreakpos = i;

      if (xcoord <= 96) {
	/* force line breaking */
	if (chunk->converted[prebreakpos] == fb->spacechar)
	  chunk->converted[prebreakpos] = fb->newlinechar;
	else {
	  fb_chunk_insert_text_ti(fb, chunk, postbreakpos,
				  &fb->newlinechar, 1);
	  postbreakpos++;
	}
      }

      *line_end = prebreakpos;
      *next_line_start = postbreakpos;
      return;
    }
  }

  *line_end = i;
  *next_line_start = i + 1;
}

int fb_chunk_split_lines(fbbook* fb, fbchunk* chunk)
{
  fbsize nlines_a;
  fbsize line_start, line_end;
  int width[256];
  int linespp, maxcc;
  int i;
  fbchar hyphen = 0, zwsp = 0;
  fbhyphctx* hyph;

  for (i = 1; i < 256; i++)
    width[i] = fb->fontchars[i].width;

  for (i = 1; i < 256; i++) {
    if (fb->encoding[i] == 0x2010)
      hyphen = i;
    if (fb->encoding[i] == 0x200B)
      zwsp = i;
  }

  if (fb->hyphenate && fb->hyphlang && chunk == fb->body)
    hyph = fb_hyphctx_new(fb);
  else
    hyph = NULL;

  /* In order to ensure that each block is at least one page long, we
     must impose a limit on the number of characters per line.  (The
     limit in a typical minimum-compression, small-font eBook will be
     76 characters, far more than we would ever expect to see in
     normal text, but not impossible.) */

  linespp = (64 + fb->rowheight - fb->fontheight) / fb->rowheight;
  maxcc = (fb->blocksize_limit / linespp) - 1;
  if (maxcc > 255)
    maxcc = 255;

  chunk->nlines = 0;
  nlines_a = 16;
  chunk->linestart = fbrenew(fbsize, chunk->linestart, nlines_a);
  chunk->lineend = fbrenew(fbsize, chunk->lineend, nlines_a);

  line_start = 0;
  while (line_start < chunk->length) {
    chunk->nlines++;
    chunk->linestart[chunk->nlines-1] = line_start;
    chunk->lineend[chunk->nlines-1] = chunk->length;

    get_line_length(fb, chunk, line_start, &line_end, &line_start,
		    width, maxcc, zwsp, hyphen, hyph);

    chunk->lineend[chunk->nlines-1] = line_end;

    if (chunk->nlines >= nlines_a) {
      nlines_a = chunk->nlines * 2;
      chunk->linestart = fbrenew(fbsize, chunk->linestart, nlines_a);
      chunk->lineend = fbrenew(fbsize, chunk->lineend, nlines_a);
    }
  }

  if (hyph)
    fb_hyphctx_free(hyph);
  else if (fb->hyphenate && fb->hyphlang && chunk == fb->body)
    return ERR_HYPHEN_INIT;

  return 0;
}

void fb_book_split_blocks(fbbook* fb)
{
  fbsize nblocks_a;
  fbsize pos;
  fbsize i;

  fb->nblocks = 1;
  nblocks_a = 16;
  fb->blockstart = fbrenew(fbsize, fb->blockstart, nblocks_a);
  fb->blockend = fbrenew(fbsize, fb->blockend, nblocks_a);

  fb->blockstart[0] = 0;
  i = 0;
  while (i < fb->body->nlines) {
    if (fb->blockstart[fb->nblocks-1] + fb->blocksize_limit
        >= fb->body->lineend[i])
      i++;
    else if (fb->blockstart[fb->nblocks-1] < fb->body->linestart[i]) {
      fb->blockend[fb->nblocks-1] = fb->body->lineend[i-1];
      fb->blockstart[fb->nblocks] = fb->body->linestart[i];
      fb->nblocks++;
    }
    else {
      /* force splitting line (this should never happen with
         non-pathological texts and block sizes) */
      pos = fb->blockstart[fb->nblocks-1] + fb->blocksize_limit;
      fb->blockend[fb->nblocks-1] = pos;
      fb->blockstart[fb->nblocks] = pos;
      fb->nblocks++;
    }

    if (fb->nblocks >= nblocks_a) {
      nblocks_a = fb->nblocks * 2;
      fb->blockstart = fbrenew(fbsize, fb->blockstart, nblocks_a);
      fb->blockend = fbrenew(fbsize, fb->blockend, nblocks_a);
    }
  }

  if (fb->nblocks == 1) {
    /* A lot of the reader code assumes that there are at least two
       blocks.  Make the second one empty. */
    fb->nblocks++;
    fb->blockend[fb->nblocks - 2] = fb->body->length;
    fb->blockstart[fb->nblocks - 1] = fb->body->length;
  }

  fb->blockend[fb->nblocks - 1] = fb->body->length;
}

void fb_book_get_block_addr(fbbook* fb, fbsize pos, fbsize* offs, fbsize* block)
{
  fbsize start, end;

  if (pos == 0) {
    *offs = *block = 0;
    return;
  }

  if (pos >= fb->blockstart[fb->nblocks - 1]) {
    *offs = pos - fb->blockstart[fb->nblocks - 1];
    *block = fb->nblocks - 1;
    return;
  }

  start = 0;
  end = fb->nblocks - 1;

  while (start + 1 < end) {
    if (pos < fb->blockstart[(start + end) / 2])
      end = (start + end) / 2;
    else
      start = (start + end) / 2;
  }

  *block = start;
  *offs = pos - fb->blockstart[start];
}


/* generate a 24-bit CRC */
static unsigned long gen_crc(unsigned long crc, unsigned int value, int bits)
{
  while (bits > 0) {
    bits--;
    if (crc & 0x800000)
      crc = (crc << 1) ^ 0x864cfb;
    else
      crc = (crc << 1);
    crc ^= ((value >> bits) & 1);
  }    
  return crc;
}

void fb_book_generate_mark_id(fbbook* fb)
{
  unsigned long id = 0xb704ce;
  fbsize i, j;
  unsigned char hashchars[64] =
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789[-";

  /* Compute CRC of absolute line positions */

  for (i = j = 0; i < fb->body->nlines; i++) {
    while (j + 1 < fb->nblocks && fb->body->linestart[i] > fb->blockend[j])
      j++;
    id = gen_crc(id, j, 16);
    id = gen_crc(id, fb->body->linestart[i] - fb->blockstart[j], 16);
  }

  /* Generate bookmark file name */

  for (i = 0; i < 4; i++) {
    fb->markid[i] = hashchars[id & 0x3f];
    id >>= 6;
  }
}
