/*
 * fbbuild - command-line utility to build eBooks
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <rabbitsign.h>
#include <fb.h>

#if !defined(strcasecmp) && !defined(HAVE_STRCASECMP) && defined(HAVE_STRICMP)
# define strcasecmp stricmp
# define strncasecmp strnicmp
#endif

#define IS_ASCII_ALNUM(cc) (((cc) >= '0' && (cc) <= '9')        \
                            || ((cc) >= 'A' && (cc) <= 'Z')     \
                            || ((cc) >= 'a' && (cc) <= 'z'))

static void version(const char* progname)
{
  printf("%s (%s)\n"
	 "Copyright (C) 2008 Benjamin Moody\n"
	 "This program is free software.  There is NO WARRANTY of any kind.\n"
	 "Report bugs to %s.\n",
	 progname, PACKAGE_STRING, PACKAGE_BUGREPORT);
}

static void usage(const char* progname)
{
  printf("Usage: %s [options] ebook-file\n", progname);
  fputs("Basic options:\n"
	" -o, --output=APP-FILE:        specify output (.8xk) file\n"
	" -f, --font=FONT-FILE:         specify font (.fbf) file\n"
	" -n, --app-name=NAME:          specify on-calculator app name\n"
	" -c, --compress=N:             set compression level (N = 1 to 5)\n"
	" -h, --hyphenate=LANG:         hyphenate using rules for LANG\n"
	"\n"
	"Advanced options:\n"
	" -S, --asm-source=ASM-FILE:    specify assembly source file\n"
	" -L, --asm-listing=LIST-FILE:  write assembly listing to given file\n"
	" -b, --block-size=N:           set compression block size\n",
	stdout);
}

#define CHECKOPT(ss, ll, vv)						\
  if ((argv[j][1] == ss && argv[j][2] == 0)				\
      || !strcasecmp(argv[j], "--" ll)) {				\
    if (vv) {								\
      fprintf(stderr, "%s: multiple values for %s specified\n",		\
	      argv[0], argv[j]);					\
      return 1;								\
    }									\
    else if (++j < argc) {						\
      (vv) = argv[j];							\
      continue;								\
    }									\
    else {								\
      fprintf(stderr, "%s: %s requires an argument\n",			\
	      argv[0], argv[j]);					\
      return 1;								\
    }									\
  }									\
  else if (argv[j][1] == ss) {						\
    if (vv) {								\
      fprintf(stderr, "%s: multiple values for -%c specified\n",	\
	      argv[0], ss);						\
      return 1;								\
    }									\
    else {								\
      (vv) = &argv[j][2];						\
      continue;								\
    }									\
  }									\
  else if (!strncasecmp(argv[j], "--" ll "=", strlen(ll) + 3)) {	\
    if (vv) {								\
      fprintf(stderr, "%s: multiple values for --%s specified\n",	\
	      argv[0], ll);						\
      return 1;								\
    }									\
    else {								\
      (vv) = &argv[j][strlen(ll) + 3];					\
      continue;								\
    }									\
  }

int main(int argc, char** argv)
{
  const char* infilename = NULL;
  const char* outfilename = NULL;
  const char* fontfilename = NULL;
  const char* asmfilename = NULL;
  const char* lstfilename = NULL;
  const char* hyphlang = NULL;
  const char* appname = NULL;
  const char* blocksize = NULL;
  const char* complevel = NULL;
  fbbook* fb;
  fboutbuf* buffer;
  RSProgram* prgm;
  RSKey* key;
  FILE* ebkfile;
  FILE* appfile;
  char* fname = NULL;
  fbsize i;
  int j;

  for (j = 1; j < argc; j++) {
    if (argv[j][0] != '-' || argv[j][1] == 0) {
      if (infilename) {
        fprintf(stderr, "%s: multiple input files specified\n", argv[0]);
        usage(argv[0]);
        return 1;
      }
      infilename = argv[j];
      continue;
    }

    CHECKOPT('o', "output", outfilename);
    CHECKOPT('f', "font", fontfilename);
    CHECKOPT('S', "asm-source", asmfilename);
    CHECKOPT('L', "asm-listing", lstfilename);
    CHECKOPT('c', "compress", complevel);
    CHECKOPT('n', "app-name", appname);
    CHECKOPT('b', "block-size", blocksize);
    CHECKOPT('h', "hyphenate", hyphlang);

    if (!strcasecmp(argv[j], "--help")) {
      usage(argv[0]);
      return 0;
    }
    else if (!strcasecmp(argv[j], "--version")) {
      version(argv[0]);
      return 0;
    }
    else {
      fprintf(stderr, "%s: unknown option %s\n", argv[0], argv[j]);
      usage(argv[0]);
      return 1;
    }
  }
  if (!infilename) {
    fprintf(stderr, "%s: no input file specified\n", argv[0]);
    usage(argv[0]);
    return 1;
  }

  rs_set_progname(argv[0]);
  rs_set_verbose(0);

  if (strcmp(infilename, "-"))
    ebkfile = fopen(infilename, "rt");
  else
    ebkfile = stdin;
  if (!ebkfile) {
    perror(infilename);
    return 2;
  }

  fb = fb_book_new();
  fb->searchdir = fb_get_dirname(infilename);

  if (fb_book_read_xml(fb, ebkfile, infilename)) {
    fb_book_free(fb);
    if (ebkfile != stdin)
      fclose(ebkfile);
    return 2;
  }

  if (ebkfile != stdin)
    fclose(ebkfile);


  if (appname) {
    fb_chunk_clear(fb->appname);
    fb_chunk_append_text_utf8(fb, fb->appname, appname, strlen(appname));
  }

  if (blocksize) {
    fb->blocksize_limit = strtol(blocksize, NULL, 10);
  }
  else if (complevel) {
    switch (strtol(complevel, NULL, 10)) {
    case 1:
      fb->blocksize_limit = 765;
      break;
    case 2:
      fb->blocksize_limit = 1023;
      break;
    case 3:
      fb->blocksize_limit = 2048;
      break;
    case 4:
      fb->blocksize_limit = 4096;
      break;
    case 5:
      fb->blocksize_limit = 8192;
      break;
    default:
      fprintf(stderr, "%s: invalid compression level %s\n",
	      argv[0], complevel);
      fb_book_free(fb);
      return 1;
    }
  }

  if (fontfilename) {
    fb->fontfile = fbnew(char, 1 + strlen(fontfilename));
    strcpy(fb->fontfile, fontfilename);
  }

  if (hyphlang) {
    fb->hyphenate = 1;
    fb->hyphlang = fbnew(char, 1 + strlen(hyphlang));
    strcpy(fb->hyphlang, hyphlang);
  }

  if (asmfilename) {
    fb->asmfile = fbnew(char, 1 + strlen(asmfilename));
    strcpy(fb->asmfile, asmfilename);
  }
  else
    asmfilename = FB_DEFAULT_SOURCE_FILE;

  if (lstfilename) {
    fb->lstfile = fopen(lstfilename, "wt");
    if (!fb->lstfile) {
      perror(lstfilename);
      fb_book_free(fb);
      return 4;
    }
    fprintf(fb->lstfile, "fbbuild (%s)\n", PACKAGE_STRING);
    fprintf(fb->lstfile, "Building reader for eBook: %s\n", infilename);
    fprintf(fb->lstfile, "Source File: %s\n\n", asmfilename);
  }

  buffer = fb_outbuf_new();
  if (fb_book_write_app(buffer, fb)) {
    if (fb->lstfile)
      fclose(fb->lstfile);
    fb_book_free(fb);
    fb_outbuf_free(buffer);
    return 3;
  }

  if (fb->lstfile)
    fclose(fb->lstfile);

  key = rs_key_new();
  if (!key) {
    fb_book_free(fb);
    fb_outbuf_free(buffer);
    return 3;
  }

  prgm = rs_program_new_with_data(RS_CALC_TI83P, RS_DATA_APP,
				  buffer->data, buffer->length,
				  buffer->length_a);
  if (!prgm) {
    fb_book_free(fb);
    fb_outbuf_free(buffer);
    rs_key_free(key);
    return 3;
  }
  else {
    buffer->data = NULL;
    fb_outbuf_free(buffer);
  }

  if (rs_key_find_for_id(key, rs_program_get_key_id(prgm), 0)
      || rs_repair_ti8x_app(prgm, (RS_FIX_PAGE_COUNT | RS_ZEALOUSLY_PAD_APP))
      || rs_sign_ti8x_app(prgm, key, 0)) {
    rs_key_free(key);
    rs_program_free(prgm);
    fb_book_free(fb);
    return 3;
  }

  rs_key_free(key);

  if (!outfilename) {
    fname = fbnew(char, 5 + fb->appname->length);
    for (i = 0; i < fb->appname->length; i++) {
      fname[i] = fb->appname->converted[i];
      if (!IS_ASCII_ALNUM(fname[i]))
        fname[i] = '_';
    }
    strcpy(&fname[i], ".8xk");
    outfilename = fname;
  }

  if (strcmp(outfilename, "-"))
    appfile = fopen(outfilename, "wb");
  else
    appfile = stdout;

  fbfree(fname);

  if (!appfile) {
    perror(outfilename);
    rs_program_free(prgm);
    fb_book_free(fb);
    return 4;
  }

  if (rs_write_ti8x_file(prgm, appfile, 0, 0, 0, 0)) {
    if (appfile != stdout)
      fclose(appfile);
    rs_program_free(prgm);
    fb_book_free(fb);
    return 4;
  }

  if (appfile != stdout)
    fclose(appfile);
  rs_program_free(prgm);
  fb_book_free(fb);
  return 0;
}
