/*
 * fbunpack - command-line utility to unpack eBooks
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <rabbitsign.h>
#include <fb.h>

#define IS_ASCII_ALNUM(cc) (((cc) >= '0' && (cc) <= '9')        \
                            || ((cc) >= 'A' && (cc) <= 'Z')     \
                            || ((cc) >= 'a' && (cc) <= 'z'))

static void version(const char* progname)
{
  printf("%s (%s)\n"
	 "Copyright (C) 2008 Benjamin Moody\n"
	 "This program is free software.  There is NO WARRANTY of any kind.\n"
	 "Report bugs to %s.\n",
	 progname, PACKAGE_STRING, PACKAGE_BUGREPORT);
}

static void usage(const char* progname)
{
  printf("Usage: %s app-file [-o ebook-file]\n", progname);
}

int main(int argc, char** argv)
{
  const char* infilename = NULL;
  const char* outfilename = NULL;
  fbbook* fb;
  RSProgram* prgm;
  FILE* appfile;
  FILE* ebkfile;
  char* fname = NULL;
  fbsize i;
  int n;

  for (n = 1; n < argc; n++) {
    if (argv[n][0] != '-' || argv[n][1] == 0) {
      if (infilename) {
        fprintf(stderr, "%s: multiple input files specified\n", argv[0]);
        usage(argv[0]);
        return 1;
      }
      infilename = argv[n];
    }
    else {
      if (argv[n][1] == 'o') {
        if (outfilename) {
          fprintf(stderr, "%s: multiple output files specified\n", argv[0]);
          usage(argv[0]);
          return 1;
        }
        if (argv[n][2])
          outfilename = &argv[n][2];
        else if (++n < argc)
          outfilename = argv[n];
        else {
          fprintf(stderr, "%s: -o requires an argument\n", argv[0]);
          return 1;
        }
      }
      else if (!strcmp(argv[n], "--help")) {
        usage(argv[0]);
        return 0;
      }
      else if (!strcmp(argv[n], "--version")) {
        version(argv[0]);
        return 0;
      }
      else {
        fprintf(stderr, "%s: unknown option %s\n", argv[0], argv[n]);
        usage(argv[0]);
        return 1;
      }
    }
  }
  if (!infilename) {
    fprintf(stderr, "%s: no input file specified\n", argv[0]);
    usage(argv[0]);
    return 1;
  }

  rs_set_progname(argv[0]);
  rs_set_verbose(0);

  if (strcmp(infilename, "-"))
    appfile = fopen(infilename, "rb");
  else
    appfile = stdin;
  if (!appfile) {
    perror(infilename);
    return 2;
  }

  prgm = rs_program_new();
  if (!prgm)
    return 2;

  if (rs_read_program_file(prgm, appfile,
			   argc > 1 ? argv[1] : "standard input", 0)) {
    if (appfile != stdin)
      fclose(appfile);
    rs_program_free(prgm);
    return 2;
  }

  fb = fb_book_new();
  if (fb_book_read_app(fb, prgm->data, prgm->length, 0x4000)) {
    fb_book_free(fb);
    rs_program_free(prgm);
    return 1;
  }

  rs_program_free(prgm);

  if (!outfilename) {
    fname = fbnew(char, 5 + fb->appname->length);
    for (i = 0; i < fb->appname->length; i++) {
      fname[i] = fb->appname->converted[i];
      if (!IS_ASCII_ALNUM(fname[i]))
        fname[i] = '_';
    }
    strcpy(&fname[i], ".ebk");
    outfilename = fname;
  }

  if (strcmp(outfilename, "-"))
    ebkfile = fopen(outfilename, "wt");
  else
    ebkfile = stdout;
  if (!ebkfile) {
    perror(outfilename);
    fbfree(fname);
    fb_book_free(fb);
    return 4;
  }
  fbfree(fname);

  n = fb_book_write_xml(fb, ebkfile);

  if (ebkfile != stdout)
    fclose(ebkfile);
  fb_book_free(fb);
  return (n ? 4 : 0);
}
